
	MODULE dp_alloc


!!
!! We will maintain a linked list of "registered" global arrays for
!! each combination of dimensionality / data type combination.
!! This is necessary in order to handle global arrays whose dimensions
!! are not evenly divisible by the the dimensions of the processor
!! array.  The linked list node corresponding to the global array
!! will describe how the global array is mapped to the 
!! processor array.
!!
!! When necessary, a dp_routine that receives an array will search
!! the appropriate linked list and will use the layout description
!! to do perform its work.  If an array is used that has not been
!! registered, it is assumed to be distributed in equal-sized blocks
!! on each processor.
!! 



	use DP_INTERNAL

	implicit none

	private


#if 1



public dp_get_local_array_block
public dp_initialize_array_lists

public dp_register_array_1D_L
public dp_register_array_2D_L
public dp_register_array_3D_L
public dp_register_array_4D_L

#if 0
public dp_register_array_1D_L1
public dp_register_array_1D_L2
public dp_register_array_2D_L1
public dp_register_array_2D_L2
public dp_register_array_3D_L1
public dp_register_array_3D_L2
public dp_register_array_4D_L1
public dp_register_array_4D_L2
#endif

public dp_register_array_1D_R
public dp_find_arr_desc_1D_R 
public dp_alloc_array_A_1D_R
public dp_alloc_array_B_1D_R
public dp_register_array_2D_R
public dp_find_arr_desc_2D_R 
public dp_alloc_array_A_2D_R
public dp_alloc_array_B_2D_R
public dp_register_array_3D_R
public dp_find_arr_desc_3D_R 
public dp_alloc_array_A_3D_R
public dp_alloc_array_B_3D_R
public dp_register_array_4D_R
public dp_find_arr_desc_4D_R 
public dp_alloc_array_A_4D_R
public dp_alloc_array_B_4D_R

#if 0
public dp_register_array_1D_R4
public dp_find_arr_desc_1D_R4 
public dp_alloc_array_A_1D_R4
public dp_alloc_array_B_1D_R4
public dp_register_array_2D_R4
public dp_find_arr_desc_2D_R4 
public dp_alloc_array_A_2D_R4
public dp_alloc_array_B_2D_R4
public dp_register_array_3D_R4
public dp_find_arr_desc_3D_R4 
public dp_alloc_array_A_3D_R4
public dp_alloc_array_B_3D_R4
public dp_register_array_4D_R4
public dp_find_arr_desc_4D_R4 
public dp_alloc_array_A_4D_R4
public dp_alloc_array_B_4D_R4
public dp_register_array_1D_R8
public dp_find_arr_desc_1D_R8 
public dp_alloc_array_A_1D_R8
public dp_alloc_array_B_1D_R8
public dp_register_array_2D_R8
public dp_find_arr_desc_2D_R8 
public dp_alloc_array_A_2D_R8
public dp_alloc_array_B_2D_R8
public dp_register_array_3D_R8
public dp_find_arr_desc_3D_R8 
public dp_alloc_array_A_3D_R8
public dp_alloc_array_B_3D_R8
public dp_register_array_4D_R8
public dp_find_arr_desc_4D_R8 
public dp_alloc_array_A_4D_R8
public dp_alloc_array_B_4D_R8
public dp_register_array_1D_R16
public dp_find_arr_desc_1D_R16 
public dp_alloc_array_A_1D_R16
public dp_alloc_array_B_1D_R16
public dp_register_array_2D_R16
public dp_find_arr_desc_2D_R16 
public dp_alloc_array_A_2D_R16
public dp_alloc_array_B_2D_R16
public dp_register_array_3D_R16
public dp_find_arr_desc_3D_R16 
public dp_alloc_array_A_3D_R16
public dp_alloc_array_B_3D_R16
public dp_register_array_4D_R16
public dp_find_arr_desc_4D_R16 
public dp_alloc_array_A_4D_R16
public dp_alloc_array_B_4D_R16
#endif


public dp_register_array_1D_D
public dp_find_arr_desc_1D_D 
public dp_alloc_array_A_1D_D
public dp_alloc_array_B_1D_D
public dp_register_array_2D_D
public dp_find_arr_desc_2D_D 
public dp_alloc_array_A_2D_D
public dp_alloc_array_B_2D_D
public dp_register_array_3D_D
public dp_find_arr_desc_3D_D 
public dp_alloc_array_A_3D_D
public dp_alloc_array_B_3D_D
public dp_register_array_4D_D
public dp_find_arr_desc_4D_D 
public dp_alloc_array_A_4D_D
public dp_alloc_array_B_4D_D
public dp_register_array_1D_C
public dp_find_arr_desc_1D_C 
public dp_alloc_array_A_1D_C
public dp_alloc_array_B_1D_C
public dp_register_array_2D_C
public dp_find_arr_desc_2D_C 
public dp_alloc_array_A_2D_C
public dp_alloc_array_B_2D_C
public dp_register_array_3D_C
public dp_find_arr_desc_3D_C 
public dp_alloc_array_A_3D_C
public dp_alloc_array_B_3D_C
public dp_register_array_4D_C
public dp_find_arr_desc_4D_C 
public dp_alloc_array_A_4D_C
public dp_alloc_array_B_4D_C

#if 0
public dp_register_array_1D_C8
public dp_find_arr_desc_1D_C8 
public dp_alloc_array_A_1D_C8
public dp_alloc_array_B_1D_C8
public dp_register_array_2D_C8
public dp_find_arr_desc_2D_C8 
public dp_alloc_array_A_2D_C8
public dp_alloc_array_B_2D_C8
public dp_register_array_3D_C8
public dp_find_arr_desc_3D_C8 
public dp_alloc_array_A_3D_C8
public dp_alloc_array_B_3D_C8
public dp_register_array_4D_C8
public dp_find_arr_desc_4D_C8 
public dp_alloc_array_A_4D_C8
public dp_alloc_array_B_4D_C8
public dp_register_array_1D_C16
public dp_find_arr_desc_1D_C16 
public dp_alloc_array_A_1D_C16
public dp_alloc_array_B_1D_C16
public dp_register_array_2D_C16
public dp_find_arr_desc_2D_C16 
public dp_alloc_array_A_2D_C16
public dp_alloc_array_B_2D_C16
public dp_register_array_3D_C16
public dp_find_arr_desc_3D_C16 
public dp_alloc_array_A_3D_C16
public dp_alloc_array_B_3D_C16
public dp_register_array_4D_C16
public dp_find_arr_desc_4D_C16 
public dp_alloc_array_A_4D_C16
public dp_alloc_array_B_4D_C16
public dp_register_array_1D_C32
public dp_find_arr_desc_1D_C32 
public dp_alloc_array_A_1D_C32
public dp_alloc_array_B_1D_C32
public dp_register_array_2D_C32
public dp_find_arr_desc_2D_C32 
public dp_alloc_array_A_2D_C32
public dp_alloc_array_B_2D_C32
public dp_register_array_3D_C32
public dp_find_arr_desc_3D_C32 
public dp_alloc_array_A_3D_C32
public dp_alloc_array_B_3D_C32
public dp_register_array_4D_C32
public dp_find_arr_desc_4D_C32 
public dp_alloc_array_A_4D_C32
public dp_alloc_array_B_4D_C32
#endif


public dp_register_array_1D_I
public dp_find_arr_desc_1D_I 
public dp_alloc_array_A_1D_I
public dp_alloc_array_B_1D_I
public dp_register_array_2D_I
public dp_find_arr_desc_2D_I 
public dp_alloc_array_A_2D_I
public dp_alloc_array_B_2D_I
public dp_register_array_3D_I
public dp_find_arr_desc_3D_I 
public dp_alloc_array_A_3D_I
public dp_alloc_array_B_3D_I
public dp_register_array_4D_I
public dp_find_arr_desc_4D_I 
public dp_alloc_array_A_4D_I
public dp_alloc_array_B_4D_I

#if 0
public dp_register_array_1D_I1
public dp_find_arr_desc_1D_I1 
public dp_alloc_array_A_1D_I1
public dp_alloc_array_B_1D_I1
public dp_register_array_2D_I1
public dp_find_arr_desc_2D_I1 
public dp_alloc_array_A_2D_I1
public dp_alloc_array_B_2D_I1
public dp_register_array_3D_I1
public dp_find_arr_desc_3D_I1 
public dp_alloc_array_A_3D_I1
public dp_alloc_array_B_3D_I1
public dp_register_array_4D_I1
public dp_find_arr_desc_4D_I1 
public dp_alloc_array_A_4D_I1
public dp_alloc_array_B_4D_I1
public dp_register_array_1D_I2
public dp_find_arr_desc_1D_I2 
public dp_alloc_array_A_1D_I2
public dp_alloc_array_B_1D_I2
public dp_register_array_2D_I2
public dp_find_arr_desc_2D_I2 
public dp_alloc_array_A_2D_I2
public dp_alloc_array_B_2D_I2
public dp_register_array_3D_I2
public dp_find_arr_desc_3D_I2 
public dp_alloc_array_A_3D_I2
public dp_alloc_array_B_3D_I2
public dp_register_array_4D_I2
public dp_find_arr_desc_4D_I2 
public dp_alloc_array_A_4D_I2
public dp_alloc_array_B_4D_I2
public dp_register_array_1D_I4
public dp_find_arr_desc_1D_I4 
public dp_alloc_array_A_1D_I4
public dp_alloc_array_B_1D_I4
public dp_register_array_2D_I4
public dp_find_arr_desc_2D_I4 
public dp_alloc_array_A_2D_I4
public dp_alloc_array_B_2D_I4
public dp_register_array_3D_I4
public dp_find_arr_desc_3D_I4 
public dp_alloc_array_A_3D_I4
public dp_alloc_array_B_3D_I4
public dp_register_array_4D_I4
public dp_find_arr_desc_4D_I4 
public dp_alloc_array_A_4D_I4
public dp_alloc_array_B_4D_I4

#endif







public dp_print_array_desc_1D_R 
public dp_print_array_desc_2D_R 
public dp_print_array_desc_3D_R 
public dp_print_array_desc_4D_R 

#if 0
public dp_print_array_desc_1D_R4 
public dp_print_array_desc_2D_R4 
public dp_print_array_desc_3D_R4 
public dp_print_array_desc_4D_R4 
public dp_print_array_desc_1D_R8 
public dp_print_array_desc_2D_R8 
public dp_print_array_desc_3D_R8 
public dp_print_array_desc_4D_R8 
public dp_print_array_desc_1D_R16 
public dp_print_array_desc_2D_R16 
public dp_print_array_desc_3D_R16 
public dp_print_array_desc_4D_R16 
#endif


public dp_print_array_desc_1D_D 
public dp_print_array_desc_2D_D 
public dp_print_array_desc_3D_D 
public dp_print_array_desc_4D_D 
public dp_print_array_desc_1D_C 
public dp_print_array_desc_2D_C 
public dp_print_array_desc_3D_C 
public dp_print_array_desc_4D_C 

#if 0
public dp_print_array_desc_1D_C8 
public dp_print_array_desc_2D_C8 
public dp_print_array_desc_3D_C8 
public dp_print_array_desc_4D_C8 
public dp_print_array_desc_1D_C16 
public dp_print_array_desc_2D_C16 
public dp_print_array_desc_3D_C16 
public dp_print_array_desc_4D_C16 
public dp_print_array_desc_1D_C32 
public dp_print_array_desc_2D_C32 
public dp_print_array_desc_3D_C32 
public dp_print_array_desc_4D_C32 
#endif

public dp_print_array_desc_1D_I 
public dp_print_array_desc_2D_I 
public dp_print_array_desc_3D_I 
public dp_print_array_desc_4D_I 

#if 0
public dp_print_array_desc_1D_I1 
public dp_print_array_desc_2D_I1 
public dp_print_array_desc_3D_I1 
public dp_print_array_desc_4D_I1 
public dp_print_array_desc_1D_I2 
public dp_print_array_desc_2D_I2 
public dp_print_array_desc_3D_I2 
public dp_print_array_desc_4D_I2 
public dp_print_array_desc_1D_I4 
public dp_print_array_desc_2D_I4 
public dp_print_array_desc_3D_I4 
public dp_print_array_desc_4D_I4 
#endif




public dp_get_arr_dist_1D_R
public dp_get_arr_dist_2D_R
public dp_get_arr_dist_3D_R
public dp_get_arr_dist_4D_R

#if 0
public dp_get_arr_dist_1D_R4
public dp_get_arr_dist_2D_R4
public dp_get_arr_dist_3D_R4
public dp_get_arr_dist_4D_R4
public dp_get_arr_dist_1D_R8
public dp_get_arr_dist_2D_R8
public dp_get_arr_dist_3D_R8
public dp_get_arr_dist_4D_R8
public dp_get_arr_dist_1D_R16
public dp_get_arr_dist_2D_R16
public dp_get_arr_dist_3D_R16
public dp_get_arr_dist_4D_R16
#endif

public dp_get_arr_dist_1D_D
public dp_get_arr_dist_2D_D
public dp_get_arr_dist_3D_D
public dp_get_arr_dist_4D_D
public dp_get_arr_dist_1D_C
public dp_get_arr_dist_2D_C
public dp_get_arr_dist_3D_C
public dp_get_arr_dist_4D_C

#if 0
public dp_get_arr_dist_1D_C8
public dp_get_arr_dist_2D_C8
public dp_get_arr_dist_3D_C8
public dp_get_arr_dist_4D_C8
public dp_get_arr_dist_1D_C16
public dp_get_arr_dist_2D_C16
public dp_get_arr_dist_3D_C16
public dp_get_arr_dist_4D_C16
public dp_get_arr_dist_1D_C32
public dp_get_arr_dist_2D_C32
public dp_get_arr_dist_3D_C32
public dp_get_arr_dist_4D_C32
#endif

public dp_get_arr_dist_1D_I
public dp_get_arr_dist_2D_I
public dp_get_arr_dist_3D_I
public dp_get_arr_dist_4D_I

#if 0
public dp_get_arr_dist_1D_I1
public dp_get_arr_dist_2D_I1
public dp_get_arr_dist_3D_I1
public dp_get_arr_dist_4D_I1
public dp_get_arr_dist_1D_I2
public dp_get_arr_dist_2D_I2
public dp_get_arr_dist_3D_I2
public dp_get_arr_dist_4D_I2
public dp_get_arr_dist_1D_I4
public dp_get_arr_dist_2D_I4
public dp_get_arr_dist_3D_I4
public dp_get_arr_dist_4D_I4
#endif

public dp_get_arr_dist_4D_L
public dp_get_arr_dist_3D_L
public dp_get_arr_dist_2D_L
public dp_get_arr_dist_1D_L


#if 0
public dp_get_arr_dist_4D_L1
public dp_get_arr_dist_3D_L1
public dp_get_arr_dist_2D_L1
public dp_get_arr_dist_1D_L1
public dp_get_arr_dist_4D_L2
public dp_get_arr_dist_3D_L2
public dp_get_arr_dist_2D_L2
public dp_get_arr_dist_1D_L2
public dp_get_arr_dist_4D_L4
public dp_get_arr_dist_3D_L4
public dp_get_arr_dist_2D_L4
public dp_get_arr_dist_1d_L4
#endif


#endif


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Here is the data type for the structure that describes
!! It describes how a global array is distributed over the processor array.
!!

type array_distribution

	logical all_even			!! Is array distributed
						!! evenly over all processors?

	logical evenly_divided(DP_MAX_DIMS)	!! Is the array distributed 
						!! evenly across each dimen?

	integer global_size (DP_MAX_DIMS)	!! The global size for
						!! each dimension.

	integer local_bounds (2,DP_MAX_DIMS)	!! The upper and lower bounds 
						!! of the local array block

	integer local_size (DP_MAX_DIMS)	!! The size of the
						!! local array block

	real local_fractional_size(DP_MAX_DIMS)	!! The "ideal" size of
						!! each local block


	integer, pointer :: blk_starts_0_based (:,:)
						!! For each dimension and for
						!! each processor position  
						!! along that dimension, this
						!! array indicates the starting
						!! index of the local block.

	integer, pointer :: blk_sizes (:,:)
						!! For each dimension and for
						!! each processor position
						!! along that dimension, this
						!! array indicates the size
						!! of the local block.

end type array_distribution


public array_distribution


#if 1


!!
!! The following include file contains all of the type definitions and
!! global variables needed for the linked lists.
!!
include 'dp_alloc_spec.h'
#endif



CONTAINS


#if 1

!!
!! The following include file contains all of the routines for
!! allocating, registering, and searching for global arrays.
!!
include 'dp_alloc_exec.h'


SUBROUTINE dp_initialize_array_lists ()

NULLIFY (Arr_List_1D_R)
NULLIFY (Arr_List_2D_R)
NULLIFY (Arr_List_3D_R)
NULLIFY (Arr_List_4D_R)

#if 0
NULLIFY (Arr_List_1D_R4)
NULLIFY (Arr_List_2D_R4)
NULLIFY (Arr_List_3D_R4)
NULLIFY (Arr_List_4D_R4)
NULLIFY (Arr_List_1D_R8)
NULLIFY (Arr_List_2D_R8)
NULLIFY (Arr_List_3D_R8)
NULLIFY (Arr_List_4D_R8)
NULLIFY (Arr_List_1D_R16)
NULLIFY (Arr_List_2D_R16)
NULLIFY (Arr_List_3D_R16)
NULLIFY (Arr_List_4D_R16)
#endif

NULLIFY (Arr_List_1D_D)
NULLIFY (Arr_List_2D_D)
NULLIFY (Arr_List_3D_D)
NULLIFY (Arr_List_4D_D)
NULLIFY (Arr_List_1D_C)
NULLIFY (Arr_List_2D_C)
NULLIFY (Arr_List_3D_C)
NULLIFY (Arr_List_4D_C)

#if 0
NULLIFY (Arr_List_1D_C8)
NULLIFY (Arr_List_2D_C8)
NULLIFY (Arr_List_3D_C8)
NULLIFY (Arr_List_4D_C8)
NULLIFY (Arr_List_1D_C16)
NULLIFY (Arr_List_2D_C16)
NULLIFY (Arr_List_3D_C16)
NULLIFY (Arr_List_4D_C16)
NULLIFY (Arr_List_1D_C32)
NULLIFY (Arr_List_2D_C32)
NULLIFY (Arr_List_3D_C32)
NULLIFY (Arr_List_4D_C32)
#endif

NULLIFY (Arr_List_1D_I)
NULLIFY (Arr_List_2D_I)
NULLIFY (Arr_List_3D_I)
NULLIFY (Arr_List_4D_I)

#if 0
NULLIFY (Arr_List_1D_I1)
NULLIFY (Arr_List_2D_I1)
NULLIFY (Arr_List_3D_I1)
NULLIFY (Arr_List_4D_I1)
NULLIFY (Arr_List_1D_I2)
NULLIFY (Arr_List_2D_I2)
NULLIFY (Arr_List_3D_I2)
NULLIFY (Arr_List_4D_I2)
NULLIFY (Arr_List_1D_I4)
NULLIFY (Arr_List_2D_I4)
NULLIFY (Arr_List_3D_I4)
NULLIFY (Arr_List_4D_I4)
#endif

NULLIFY (Arr_List_1D_L)
NULLIFY (Arr_List_2D_L)
NULLIFY (Arr_List_3D_L)
NULLIFY (Arr_List_4D_L)

#if 0
NULLIFY (Arr_List_1D_L1)
NULLIFY (Arr_List_2D_L1)
NULLIFY (Arr_List_3D_L1)
NULLIFY (Arr_List_4D_L1)
NULLIFY (Arr_List_1D_L2)
NULLIFY (Arr_List_2D_L2)
NULLIFY (Arr_List_3D_L2)
NULLIFY (Arr_List_4D_L2)
NULLIFY (Arr_List_1D_L4)
NULLIFY (Arr_List_2D_L4)
NULLIFY (Arr_List_3D_L4)
NULLIFY (Arr_List_4D_L4)
#endif


end SUBROUTINE dp_initialize_array_lists


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!  Calculate a local array block start, and end in a given dimension
!!  for a given processor position along that dimension
!! 
!! 
!!  public dp_get_local_array_block
!!
SUBROUTINE dp_get_local_array_block (dim, proc_index, &
				     global_start, global_end, &
				     local_start, local_end, ierr)
integer, intent(IN) :: dim
integer, intent(IN) :: proc_index
integer, intent(IN) :: global_start
integer, intent(IN) :: global_end

integer, intent(OUT) :: local_start
integer, intent(OUT) :: local_end
integer, intent(OUT) :: ierr

real :: fractional_size
integer :: global_size
integer :: local_size


print *, "dp_get_local A"

global_size = 1 + global_end - global_start

if (dim .le. DP_MAX_DIMS) then

print *, "dp_get_local B"

	if (MOD (global_size, dp_proc_arr_size(dim)) .eq. 0) then
print *, "dp_get_local C"
		local_size = global_size / dp_proc_arr_size(dim)
		local_start = global_start + proc_index*local_size
        	local_end = local_start + local_size - 1
print *, "dp_get_local D"
	else

print *, "dp_get_local E"
		fractional_size = global_size / REAL (dp_proc_arr_size(dim))
		local_start = global_start + 0.5 + proc_index*fractional_size
		local_end =   &
		   global_start + 0.5 + (proc_index+1)*fractional_size - 1
print *, "dp_get_local F"
	endif

print *, "dp_get_local G"
else
print *, "dp_get_local H"
	local_start = global_start
	local_end = global_end
print *, "dp_get_local I"
endif


print *, "dp_get_local J"

ierr = DP_SUCCESS
return

end SUBROUTINE dp_get_local_array_block




#endif

	end MODULE dp_alloc







