
!	public DP_WRITE_ARRAY_1D_R 
!	public DP_WRITE_ARRAY_2D_R 
!	public DP_WRITE_ARRAY_3D_R 
!	public DP_WRITE_ARRAY_4D_R 
!
!	public DP_READ_ARRAY_1D_R 
!	public DP_READ_ARRAY_2D_R 
!	public DP_READ_ARRAY_3D_R 
!	public DP_READ_ARRAY_4D_R 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of REAL array IO routines
!!!






subroutine DP_SKIP_ELEMENTS_R (unit, text, skip, iostat)
integer unit
logical text
integer skip
integer iostat

	REAL d
	integer i
	


	if (text)  then
		do i = 1, skip
			read (unit, *, iostat=iostat) d
		enddo
	else

		do i = 1, skip
			read (unit, '(A)', iostat=iostat, advance='NO') d
		enddo

	endif



end subroutine dp_skip_elements_R




subroutine DP_WRITE_ARRAY_1D_R (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
REAL arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims = 1
	dims(1) = size(arr,1)

	call dp_write_array_4d_wrapper_R &
		(unit, filename, text, hdr, arr, dims, 1, status, iostat)

end subroutine DP_WRITE_ARRAY_1D_R


subroutine DP_WRITE_ARRAY_2D_R (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
REAL arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_R &
		(unit, filename, text, hdr, arr, dims, 2, status, iostat)

end subroutine DP_WRITE_ARRAY_2D_R


subroutine DP_WRITE_ARRAY_3D_R (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
REAL arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_R &
		(unit, filename, text, hdr, arr, dims, 3, status, iostat)

end subroutine DP_WRITE_ARRAY_3D_R






subroutine DP_WRITE_ARRAY_4D_WRAPPER_R (unit, filename, text, hdr, arr, dims, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
integer dims(DP_MAX_DIMS)
REAL arr(0:dims(1)-1, 0:dims(2)-1, 0:dims(3)-1, 0:dims(4)-1)
integer num_true_dims
integer iostat

    call DP_WRITE_ARRAY_4D_INNER_R &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_WRAPPER_R




subroutine DP_WRITE_ARRAY_4D_R &
	(unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
REAL arr(0:,0:,0:,0:)
integer iostat

	call DP_WRITE_ARRAY_4D_INNER_R &
		(unit, filename, text, hdr, arr, 4, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_R


subroutine DP_WRITE_ARRAY_4D_INNER_R &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
REAL arr(0:,0:,0:,0:)	! dim dependency
integer num_true_dims
integer iostat


	integer ierr
	logical is_first_rank

	integer*4 stat(MPI_STATUS_SIZE)
	integer next_coords(DP_MAX_DIMS)
	integer next_rank
	integer first_coords(DP_MAX_DIMS)
	integer first_rank

	integer i, j, k, l, jj
	integer iproc, jproc, kproc, lproc
	integer dummy



	REAL , allocatable :: local_col(:)
	character*(128) fmt_str


! MUST COORDINATE PROCESSES

	is_first_rank = .true.
	do i = 1, DP_MAX_DIMS
		is_first_rank = is_first_rank .and. (dp_coords(i).eq.0)
	enddo

	if (text) then
		call make_fmt_R (fmt_str)
	endif

	if (dp_num_procs .eq. 1) then

		call dp_open (unit, 'READWRITE', filename, text,  &
			'REWIND', status, iostat)

		if (hdr) then
			call dp_write_hdr &
				(unit, text, shape(arr), num_true_dims, iostat)
		endif

		if (text) then
			write (unit, fmt=fmt_str) arr
		else
			write (fmt_str, '(A, I12, A)') &
				'(', size(arr), 'A)'
			write (unit, fmt=fmt_str, advance='NO') arr
		endif
		call dp_close (unit)

	else ! we have multiple processes

		! The first rank is responsible for all I/O it receives
		! all data to write from other ranks via MPI_RECV
		if (is_first_rank) then
			call dp_open (unit, 'READWRITE', filename, text,  &
				'REWIND', status, iostat)
			if (hdr) then
				call dp_write_hdr (unit, text, shape(arr), &
					num_true_dims, iostat)
			endif

			allocate (local_col(0:size(arr,1)-1))

			!
			! Loop through processors and j & k dimensions
			! within array.  Write out j,kth column from
			! each processor in turn.
			! Note that this could be made more efficient by
			! figuring out when more than a
			! single column is contiguous, and thus can be
			! handled as a unit.
			!
		  DO lproc = 0, dp_proc_arr_size(4) - 1
		    next_coords(4) = lproc

		    DO l = 0, size (arr, 4) - 1

			DO kproc = 0, dp_proc_arr_size(3)-1
			  next_coords(3) = kproc

			  DO k = 0, size (arr,3)-1

			    DO jproc = 0, dp_proc_arr_size(2)-1
			      next_coords(2) = jproc

			      DO j = 0, size (arr,2)-1

			        DO iproc = 0, dp_proc_arr_size(1)-1
			            next_coords(1) = iproc

		    	            call MPI_CART_RANK (dp_comm_cart, &
							next_coords,	&
							next_rank,	&
							ierr)

			            ! get arr (:,j,k,l) from 
				    !     proc at (iproc, jproc, kproc, lproc)

		    	            if (next_rank .eq. dp_rank) then
					if (text) then
					   write (unit, fmt=fmt_str)arr(:,j,k,l)
					else
					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					       advance='NO') arr(:,j,k,l)

					endif
				    else
					!send msg to next_rank
					!requesting column j
					!recv col j into local_col

					call MPI_SENDRECV ( &
						  j,			&
						  1,			&
						  MPI_INTEGER,		&
						  next_rank,		&
						  DP_REQUEST_ARR_COL,	&
						  local_col,		&
						  size(arr,1),		&
						  MPI_REAL,		&
						  next_rank,		&
						  DP_ARR_COL,		&
						  dp_comm_cart,		&
						  stat,			&
						  ierr)

					if (text) then
					   write (unit, fmt=fmt_str) local_col
					else

					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					      advance='NO') local_col

					endif
				    endif

			        ENDDO
			      ENDDO
			    ENDDO
			  ENDDO
			ENDDO

		    ENDDO

		  ENDDO

			deallocate (local_col)
			close (unit)
		  

		else  ! we are not the first rank
			first_coords = 0
			call MPI_CART_RANK (dp_comm_cart, first_coords,	&
							first_rank,	&
							ierr)
		    do l = 0, size(arr,4)-1
			do k = 0, size(arr, 3)-1
			    do j = 0, size(arr,2)-1

				! recv msg requesting column
				call MPI_RECV ( &
				  jj,			      &
				  1,			      &
				  MPI_INTEGER,		      &
				  first_rank,		      &
				  DP_REQUEST_ARR_COL, dp_comm_cart, &
				  status, ierr)

				! send arr(:,j,k,l)
				call MPI_SEND (		      &
				  arr(:,j,k,l),		      &
				  size(arr,1),		      &
				  MPI_REAL,	      	      &
				  first_rank,	 	      &
				  DP_ARR_COL,		      &
				  dp_comm_cart,		      &
				  ierr)

			    ENDDO !! j
							
			ENDDO  !! k

		    ENDDO  !! l

		endif
	endif
			

end subroutine DP_WRITE_ARRAY_4D_INNER_R



subroutine DP_READ_ARRAY_1D_R (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
REAL arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_R &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_1D_R



subroutine DP_READ_ARRAY_2D_R (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
REAL arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_R &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_2D_R


subroutine DP_READ_ARRAY_3D_R (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
REAL arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_R &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_3D_R



subroutine DP_READ_ARRAY_4D_WRAPPER_R &
	(unit, filename, text, hdr, dims, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
integer dims(DP_MAX_DIMS)
REAL arr(0:dims(1)-1,0:dims(2)-1,0:dims(3)-1,0:dims(4)-1)
integer iostat

    call DP_READ_ARRAY_4D_R (unit, filename, text, hdr, arr, iostat)

end subroutine DP_READ_ARRAY_4D_WRAPPER_R


subroutine DP_READ_ARRAY_4D_R (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr			! read a header?
REAL arr(0:,0:,0:,0:)
integer iostat

! EACH PROCESS can do this for itself although this approach
! might be slowed down by contention for disk operation.  Then
! again it my be speeded up by good disk caching.

	!! To revise this for higher dimensionality, need only change
	!! DP_MAX_DIMS and the declaration and indexing of arr

	integer*4 status(MPI_STATUS_SIZE)
	integer skip, i, j, k, l, n

	character*(128) fmt_str
	character*(3) adv

	integer part_shape (DP_MAX_DIMS)
	integer full_shape (DP_MAX_DIMS)
	integer full_start_coords (DP_MAX_DIMS)
	integer skip_mult
	
	integer dim_skips (DP_MAX_DIMS+1)
	integer stride
	integer next_pos
	integer ind (DP_MAX_DIMS+1)
	integer sub_sizes (DP_MAX_DIMS+1)
	integer num_reads 
	integer initial_skip
	




	call dp_open (unit, 'READ', filename, text, 'REWIND', 'OLD', iostat)


	part_shape = shape (arr)
	full_shape = dp_proc_arr_size * part_shape

	

	full_start_coords = part_shape*dp_coords

	initial_skip = 0
	skip_mult = 1
	do i = 1, DP_MAX_DIMS
		initial_skip = initial_skip + full_start_coords(i) * skip_mult
		skip_mult = skip_mult * full_shape(i)
	ENDDO


	if (hdr) then
		call dp_skip_header (unit, text, iostat)
	endif



	call dp_skip_elements_R (unit, text, initial_skip, iostat)

	
	!! figure out how many element to skip when going to next
	!! start point in each dimension
	dim_skips = 0
	stride = full_shape(1)
	next_pos = part_shape(1)
	num_reads = 1
	sub_sizes = 1
	do i = 2, DP_MAX_DIMS+1
		sub_sizes(i) = sub_sizes(i-1)*part_shape(i)
		num_reads = num_reads * part_shape(i)
		dim_skips(i) = stride - next_pos
		next_pos = stride * part_shape(i)
		stride = stride * full_shape(i)
	enddo


	if (text) then
	    call make_fmt_R (fmt_str)
	    adv = 'YES'
	else
	    write (fmt_str, '(A, I12, A)') '(', part_shape(1), 'A)'
	    adv = 'NO'
	endif


	!! set index array (ind) to zero
	ind = 0



	!! Do all but the last read followed by appropriate skipping.
	!!   We omit the last iteration so that we don't skip past
	!!   the end of the array
	do n = 1, num_reads - 1
		read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))



		do i = 1, DP_MAX_DIMS-1
			!! For each dimension if we're at the end of the
			!!   the range of the dimension, execute the skip
			!!   for that dimension.
			if (MOD(n, sub_sizes(i)) .eq. 0) then



				call dp_skip_elements_R &
                                        (unit, text, dim_skips(i+1), iostat)
				ind(i+1) = ind(i+1)+1
				ind(0:i) = 0
			endif
		enddo

	enddo

	!! Do the last read.
	read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))


	close (unit)

end subroutine DP_READ_ARRAY_4D_R






!!!
!!!  End of REAL array IO routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!	public DP_WRITE_ARRAY_1D_D 
!	public DP_WRITE_ARRAY_2D_D 
!	public DP_WRITE_ARRAY_3D_D 
!	public DP_WRITE_ARRAY_4D_D 
!
!	public DP_READ_ARRAY_1D_D 
!	public DP_READ_ARRAY_2D_D 
!	public DP_READ_ARRAY_3D_D 
!	public DP_READ_ARRAY_4D_D 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of DOUBLE PRECISION array IO routines
!!!






subroutine DP_SKIP_ELEMENTS_D (unit, text, skip, iostat)
integer unit
logical text
integer skip
integer iostat

	DOUBLE PRECISION d
	integer i
	


	if (text)  then
		do i = 1, skip
			read (unit, *, iostat=iostat) d
		enddo
	else

		do i = 1, skip
			read (unit, '(A)', iostat=iostat, advance='NO') d
		enddo

	endif



end subroutine dp_skip_elements_D




subroutine DP_WRITE_ARRAY_1D_D (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
DOUBLE PRECISION arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims = 1
	dims(1) = size(arr,1)

	call dp_write_array_4d_wrapper_D &
		(unit, filename, text, hdr, arr, dims, 1, status, iostat)

end subroutine DP_WRITE_ARRAY_1D_D


subroutine DP_WRITE_ARRAY_2D_D (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
DOUBLE PRECISION arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_D &
		(unit, filename, text, hdr, arr, dims, 2, status, iostat)

end subroutine DP_WRITE_ARRAY_2D_D


subroutine DP_WRITE_ARRAY_3D_D (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
DOUBLE PRECISION arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_D &
		(unit, filename, text, hdr, arr, dims, 3, status, iostat)

end subroutine DP_WRITE_ARRAY_3D_D






subroutine DP_WRITE_ARRAY_4D_WRAPPER_D (unit, filename, text, hdr, arr, dims, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
integer dims(DP_MAX_DIMS)
DOUBLE PRECISION arr(0:dims(1)-1, 0:dims(2)-1, 0:dims(3)-1, 0:dims(4)-1)
integer num_true_dims
integer iostat

    call DP_WRITE_ARRAY_4D_INNER_D &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_WRAPPER_D




subroutine DP_WRITE_ARRAY_4D_D &
	(unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
DOUBLE PRECISION arr(0:,0:,0:,0:)
integer iostat

	call DP_WRITE_ARRAY_4D_INNER_D &
		(unit, filename, text, hdr, arr, 4, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_D


subroutine DP_WRITE_ARRAY_4D_INNER_D &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
DOUBLE PRECISION arr(0:,0:,0:,0:)	! dim dependency
integer num_true_dims
integer iostat


	integer ierr
	logical is_first_rank

	integer*4 stat(MPI_STATUS_SIZE)
	integer next_coords(DP_MAX_DIMS)
	integer next_rank
	integer first_coords(DP_MAX_DIMS)
	integer first_rank

	integer i, j, k, l, jj
	integer iproc, jproc, kproc, lproc
	integer dummy



	DOUBLE PRECISION , allocatable :: local_col(:)
	character*(128) fmt_str


! MUST COORDINATE PROCESSES

	is_first_rank = .true.
	do i = 1, DP_MAX_DIMS
		is_first_rank = is_first_rank .and. (dp_coords(i).eq.0)
	enddo

	if (text) then
		call make_fmt_D (fmt_str)
	endif

	if (dp_num_procs .eq. 1) then

		call dp_open (unit, 'READWRITE', filename, text,  &
			'REWIND', status, iostat)

		if (hdr) then
			call dp_write_hdr &
				(unit, text, shape(arr), num_true_dims, iostat)
		endif

		if (text) then
			write (unit, fmt=fmt_str) arr
		else
			write (fmt_str, '(A, I12, A)') &
				'(', size(arr), 'A)'
			write (unit, fmt=fmt_str, advance='NO') arr
		endif
		call dp_close (unit)

	else ! we have multiple processes

		! The first rank is responsible for all I/O it receives
		! all data to write from other ranks via MPI_RECV
		if (is_first_rank) then
			call dp_open (unit, 'READWRITE', filename, text,  &
				'REWIND', status, iostat)
			if (hdr) then
				call dp_write_hdr (unit, text, shape(arr), &
					num_true_dims, iostat)
			endif

			allocate (local_col(0:size(arr,1)-1))

			!
			! Loop through processors and j & k dimensions
			! within array.  Write out j,kth column from
			! each processor in turn.
			! Note that this could be made more efficient by
			! figuring out when more than a
			! single column is contiguous, and thus can be
			! handled as a unit.
			!
		  DO lproc = 0, dp_proc_arr_size(4) - 1
		    next_coords(4) = lproc

		    DO l = 0, size (arr, 4) - 1

			DO kproc = 0, dp_proc_arr_size(3)-1
			  next_coords(3) = kproc

			  DO k = 0, size (arr,3)-1

			    DO jproc = 0, dp_proc_arr_size(2)-1
			      next_coords(2) = jproc

			      DO j = 0, size (arr,2)-1

			        DO iproc = 0, dp_proc_arr_size(1)-1
			            next_coords(1) = iproc

		    	            call MPI_CART_RANK (dp_comm_cart, &
							next_coords,	&
							next_rank,	&
							ierr)

			            ! get arr (:,j,k,l) from 
				    !     proc at (iproc, jproc, kproc, lproc)

		    	            if (next_rank .eq. dp_rank) then
					if (text) then
					   write (unit, fmt=fmt_str)arr(:,j,k,l)
					else
					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					       advance='NO') arr(:,j,k,l)

					endif
				    else
					!send msg to next_rank
					!requesting column j
					!recv col j into local_col

					call MPI_SENDRECV ( &
						  j,			&
						  1,			&
						  MPI_INTEGER,		&
						  next_rank,		&
						  DP_REQUEST_ARR_COL,	&
						  local_col,		&
						  size(arr,1),		&
						  MPI_DOUBLE_PRECISION,		&
						  next_rank,		&
						  DP_ARR_COL,		&
						  dp_comm_cart,		&
						  stat,			&
						  ierr)

					if (text) then
					   write (unit, fmt=fmt_str) local_col
					else

					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					      advance='NO') local_col

					endif
				    endif

			        ENDDO
			      ENDDO
			    ENDDO
			  ENDDO
			ENDDO

		    ENDDO

		  ENDDO

			deallocate (local_col)
			close (unit)
		  

		else  ! we are not the first rank
			first_coords = 0
			call MPI_CART_RANK (dp_comm_cart, first_coords,	&
							first_rank,	&
							ierr)
		    do l = 0, size(arr,4)-1
			do k = 0, size(arr, 3)-1
			    do j = 0, size(arr,2)-1

				! recv msg requesting column
				call MPI_RECV ( &
				  jj,			      &
				  1,			      &
				  MPI_INTEGER,		      &
				  first_rank,		      &
				  DP_REQUEST_ARR_COL, dp_comm_cart, &
				  status, ierr)

				! send arr(:,j,k,l)
				call MPI_SEND (		      &
				  arr(:,j,k,l),		      &
				  size(arr,1),		      &
				  MPI_DOUBLE_PRECISION,	      	      &
				  first_rank,	 	      &
				  DP_ARR_COL,		      &
				  dp_comm_cart,		      &
				  ierr)

			    ENDDO !! j
							
			ENDDO  !! k

		    ENDDO  !! l

		endif
	endif
			

end subroutine DP_WRITE_ARRAY_4D_INNER_D



subroutine DP_READ_ARRAY_1D_D (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
DOUBLE PRECISION arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_D &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_1D_D



subroutine DP_READ_ARRAY_2D_D (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
DOUBLE PRECISION arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_D &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_2D_D


subroutine DP_READ_ARRAY_3D_D (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
DOUBLE PRECISION arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_D &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_3D_D



subroutine DP_READ_ARRAY_4D_WRAPPER_D &
	(unit, filename, text, hdr, dims, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
integer dims(DP_MAX_DIMS)
DOUBLE PRECISION arr(0:dims(1)-1,0:dims(2)-1,0:dims(3)-1,0:dims(4)-1)
integer iostat

    call DP_READ_ARRAY_4D_D (unit, filename, text, hdr, arr, iostat)

end subroutine DP_READ_ARRAY_4D_WRAPPER_D


subroutine DP_READ_ARRAY_4D_D (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr			! read a header?
DOUBLE PRECISION arr(0:,0:,0:,0:)
integer iostat

! EACH PROCESS can do this for itself although this approach
! might be slowed down by contention for disk operation.  Then
! again it my be speeded up by good disk caching.

	!! To revise this for higher dimensionality, need only change
	!! DP_MAX_DIMS and the declaration and indexing of arr

	integer*4 status(MPI_STATUS_SIZE)
	integer skip, i, j, k, l, n

	character*(128) fmt_str
	character*(3) adv

	integer part_shape (DP_MAX_DIMS)
	integer full_shape (DP_MAX_DIMS)
	integer full_start_coords (DP_MAX_DIMS)
	integer skip_mult
	
	integer dim_skips (DP_MAX_DIMS+1)
	integer stride
	integer next_pos
	integer ind (DP_MAX_DIMS+1)
	integer sub_sizes (DP_MAX_DIMS+1)
	integer num_reads 
	integer initial_skip
	




	call dp_open (unit, 'READ', filename, text, 'REWIND', 'OLD', iostat)


	part_shape = shape (arr)
	full_shape = dp_proc_arr_size * part_shape

	

	full_start_coords = part_shape*dp_coords

	initial_skip = 0
	skip_mult = 1
	do i = 1, DP_MAX_DIMS
		initial_skip = initial_skip + full_start_coords(i) * skip_mult
		skip_mult = skip_mult * full_shape(i)
	ENDDO


	if (hdr) then
		call dp_skip_header (unit, text, iostat)
	endif



	call dp_skip_elements_D (unit, text, initial_skip, iostat)

	
	!! figure out how many element to skip when going to next
	!! start point in each dimension
	dim_skips = 0
	stride = full_shape(1)
	next_pos = part_shape(1)
	num_reads = 1
	sub_sizes = 1
	do i = 2, DP_MAX_DIMS+1
		sub_sizes(i) = sub_sizes(i-1)*part_shape(i)
		num_reads = num_reads * part_shape(i)
		dim_skips(i) = stride - next_pos
		next_pos = stride * part_shape(i)
		stride = stride * full_shape(i)
	enddo


	if (text) then
	    call make_fmt_D (fmt_str)
	    adv = 'YES'
	else
	    write (fmt_str, '(A, I12, A)') '(', part_shape(1), 'A)'
	    adv = 'NO'
	endif


	!! set index array (ind) to zero
	ind = 0



	!! Do all but the last read followed by appropriate skipping.
	!!   We omit the last iteration so that we don't skip past
	!!   the end of the array
	do n = 1, num_reads - 1
		read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))



		do i = 1, DP_MAX_DIMS-1
			!! For each dimension if we're at the end of the
			!!   the range of the dimension, execute the skip
			!!   for that dimension.
			if (MOD(n, sub_sizes(i)) .eq. 0) then



				call dp_skip_elements_D &
                                        (unit, text, dim_skips(i+1), iostat)
				ind(i+1) = ind(i+1)+1
				ind(0:i) = 0
			endif
		enddo

	enddo

	!! Do the last read.
	read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))


	close (unit)

end subroutine DP_READ_ARRAY_4D_D






!!!
!!!  End of DOUBLE PRECISION array IO routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!	public DP_WRITE_ARRAY_1D_C 
!	public DP_WRITE_ARRAY_2D_C 
!	public DP_WRITE_ARRAY_3D_C 
!	public DP_WRITE_ARRAY_4D_C 
!
!	public DP_READ_ARRAY_1D_C 
!	public DP_READ_ARRAY_2D_C 
!	public DP_READ_ARRAY_3D_C 
!	public DP_READ_ARRAY_4D_C 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of COMPLEX array IO routines
!!!






subroutine DP_SKIP_ELEMENTS_C (unit, text, skip, iostat)
integer unit
logical text
integer skip
integer iostat

	COMPLEX d
	integer i
	


	if (text)  then
		do i = 1, skip
			read (unit, *, iostat=iostat) d
		enddo
	else

		do i = 1, skip
			read (unit, '(A)', iostat=iostat, advance='NO') d
		enddo

	endif



end subroutine dp_skip_elements_C




subroutine DP_WRITE_ARRAY_1D_C (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
COMPLEX arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims = 1
	dims(1) = size(arr,1)

	call dp_write_array_4d_wrapper_C &
		(unit, filename, text, hdr, arr, dims, 1, status, iostat)

end subroutine DP_WRITE_ARRAY_1D_C


subroutine DP_WRITE_ARRAY_2D_C (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
COMPLEX arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_C &
		(unit, filename, text, hdr, arr, dims, 2, status, iostat)

end subroutine DP_WRITE_ARRAY_2D_C


subroutine DP_WRITE_ARRAY_3D_C (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
COMPLEX arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_C &
		(unit, filename, text, hdr, arr, dims, 3, status, iostat)

end subroutine DP_WRITE_ARRAY_3D_C






subroutine DP_WRITE_ARRAY_4D_WRAPPER_C (unit, filename, text, hdr, arr, dims, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
integer dims(DP_MAX_DIMS)
COMPLEX arr(0:dims(1)-1, 0:dims(2)-1, 0:dims(3)-1, 0:dims(4)-1)
integer num_true_dims
integer iostat

    call DP_WRITE_ARRAY_4D_INNER_C &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_WRAPPER_C




subroutine DP_WRITE_ARRAY_4D_C &
	(unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
COMPLEX arr(0:,0:,0:,0:)
integer iostat

	call DP_WRITE_ARRAY_4D_INNER_C &
		(unit, filename, text, hdr, arr, 4, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_C


subroutine DP_WRITE_ARRAY_4D_INNER_C &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
COMPLEX arr(0:,0:,0:,0:)	! dim dependency
integer num_true_dims
integer iostat


	integer ierr
	logical is_first_rank

	integer*4 stat(MPI_STATUS_SIZE)
	integer next_coords(DP_MAX_DIMS)
	integer next_rank
	integer first_coords(DP_MAX_DIMS)
	integer first_rank

	integer i, j, k, l, jj
	integer iproc, jproc, kproc, lproc
	integer dummy



	COMPLEX , allocatable :: local_col(:)
	character*(128) fmt_str


! MUST COORDINATE PROCESSES

	is_first_rank = .true.
	do i = 1, DP_MAX_DIMS
		is_first_rank = is_first_rank .and. (dp_coords(i).eq.0)
	enddo

	if (text) then
		call make_fmt_C (fmt_str)
	endif

	if (dp_num_procs .eq. 1) then

		call dp_open (unit, 'READWRITE', filename, text,  &
			'REWIND', status, iostat)

		if (hdr) then
			call dp_write_hdr &
				(unit, text, shape(arr), num_true_dims, iostat)
		endif

		if (text) then
			write (unit, fmt=fmt_str) arr
		else
			write (fmt_str, '(A, I12, A)') &
				'(', size(arr), 'A)'
			write (unit, fmt=fmt_str, advance='NO') arr
		endif
		call dp_close (unit)

	else ! we have multiple processes

		! The first rank is responsible for all I/O it receives
		! all data to write from other ranks via MPI_RECV
		if (is_first_rank) then
			call dp_open (unit, 'READWRITE', filename, text,  &
				'REWIND', status, iostat)
			if (hdr) then
				call dp_write_hdr (unit, text, shape(arr), &
					num_true_dims, iostat)
			endif

			allocate (local_col(0:size(arr,1)-1))

			!
			! Loop through processors and j & k dimensions
			! within array.  Write out j,kth column from
			! each processor in turn.
			! Note that this could be made more efficient by
			! figuring out when more than a
			! single column is contiguous, and thus can be
			! handled as a unit.
			!
		  DO lproc = 0, dp_proc_arr_size(4) - 1
		    next_coords(4) = lproc

		    DO l = 0, size (arr, 4) - 1

			DO kproc = 0, dp_proc_arr_size(3)-1
			  next_coords(3) = kproc

			  DO k = 0, size (arr,3)-1

			    DO jproc = 0, dp_proc_arr_size(2)-1
			      next_coords(2) = jproc

			      DO j = 0, size (arr,2)-1

			        DO iproc = 0, dp_proc_arr_size(1)-1
			            next_coords(1) = iproc

		    	            call MPI_CART_RANK (dp_comm_cart, &
							next_coords,	&
							next_rank,	&
							ierr)

			            ! get arr (:,j,k,l) from 
				    !     proc at (iproc, jproc, kproc, lproc)

		    	            if (next_rank .eq. dp_rank) then
					if (text) then
					   write (unit, fmt=fmt_str)arr(:,j,k,l)
					else
					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					       advance='NO') arr(:,j,k,l)

					endif
				    else
					!send msg to next_rank
					!requesting column j
					!recv col j into local_col

					call MPI_SENDRECV ( &
						  j,			&
						  1,			&
						  MPI_INTEGER,		&
						  next_rank,		&
						  DP_REQUEST_ARR_COL,	&
						  local_col,		&
						  size(arr,1),		&
						  MPI_COMPLEX,		&
						  next_rank,		&
						  DP_ARR_COL,		&
						  dp_comm_cart,		&
						  stat,			&
						  ierr)

					if (text) then
					   write (unit, fmt=fmt_str) local_col
					else

					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					      advance='NO') local_col

					endif
				    endif

			        ENDDO
			      ENDDO
			    ENDDO
			  ENDDO
			ENDDO

		    ENDDO

		  ENDDO

			deallocate (local_col)
			close (unit)
		  

		else  ! we are not the first rank
			first_coords = 0
			call MPI_CART_RANK (dp_comm_cart, first_coords,	&
							first_rank,	&
							ierr)
		    do l = 0, size(arr,4)-1
			do k = 0, size(arr, 3)-1
			    do j = 0, size(arr,2)-1

				! recv msg requesting column
				call MPI_RECV ( &
				  jj,			      &
				  1,			      &
				  MPI_INTEGER,		      &
				  first_rank,		      &
				  DP_REQUEST_ARR_COL, dp_comm_cart, &
				  status, ierr)

				! send arr(:,j,k,l)
				call MPI_SEND (		      &
				  arr(:,j,k,l),		      &
				  size(arr,1),		      &
				  MPI_COMPLEX,	      	      &
				  first_rank,	 	      &
				  DP_ARR_COL,		      &
				  dp_comm_cart,		      &
				  ierr)

			    ENDDO !! j
							
			ENDDO  !! k

		    ENDDO  !! l

		endif
	endif
			

end subroutine DP_WRITE_ARRAY_4D_INNER_C



subroutine DP_READ_ARRAY_1D_C (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
COMPLEX arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_C &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_1D_C



subroutine DP_READ_ARRAY_2D_C (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
COMPLEX arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_C &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_2D_C


subroutine DP_READ_ARRAY_3D_C (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
COMPLEX arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_C &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_3D_C



subroutine DP_READ_ARRAY_4D_WRAPPER_C &
	(unit, filename, text, hdr, dims, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
integer dims(DP_MAX_DIMS)
COMPLEX arr(0:dims(1)-1,0:dims(2)-1,0:dims(3)-1,0:dims(4)-1)
integer iostat

    call DP_READ_ARRAY_4D_C (unit, filename, text, hdr, arr, iostat)

end subroutine DP_READ_ARRAY_4D_WRAPPER_C


subroutine DP_READ_ARRAY_4D_C (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr			! read a header?
COMPLEX arr(0:,0:,0:,0:)
integer iostat

! EACH PROCESS can do this for itself although this approach
! might be slowed down by contention for disk operation.  Then
! again it my be speeded up by good disk caching.

	!! To revise this for higher dimensionality, need only change
	!! DP_MAX_DIMS and the declaration and indexing of arr

	integer*4 status(MPI_STATUS_SIZE)
	integer skip, i, j, k, l, n

	character*(128) fmt_str
	character*(3) adv

	integer part_shape (DP_MAX_DIMS)
	integer full_shape (DP_MAX_DIMS)
	integer full_start_coords (DP_MAX_DIMS)
	integer skip_mult
	
	integer dim_skips (DP_MAX_DIMS+1)
	integer stride
	integer next_pos
	integer ind (DP_MAX_DIMS+1)
	integer sub_sizes (DP_MAX_DIMS+1)
	integer num_reads 
	integer initial_skip
	




	call dp_open (unit, 'READ', filename, text, 'REWIND', 'OLD', iostat)


	part_shape = shape (arr)
	full_shape = dp_proc_arr_size * part_shape

	

	full_start_coords = part_shape*dp_coords

	initial_skip = 0
	skip_mult = 1
	do i = 1, DP_MAX_DIMS
		initial_skip = initial_skip + full_start_coords(i) * skip_mult
		skip_mult = skip_mult * full_shape(i)
	ENDDO


	if (hdr) then
		call dp_skip_header (unit, text, iostat)
	endif



	call dp_skip_elements_C (unit, text, initial_skip, iostat)

	
	!! figure out how many element to skip when going to next
	!! start point in each dimension
	dim_skips = 0
	stride = full_shape(1)
	next_pos = part_shape(1)
	num_reads = 1
	sub_sizes = 1
	do i = 2, DP_MAX_DIMS+1
		sub_sizes(i) = sub_sizes(i-1)*part_shape(i)
		num_reads = num_reads * part_shape(i)
		dim_skips(i) = stride - next_pos
		next_pos = stride * part_shape(i)
		stride = stride * full_shape(i)
	enddo


	if (text) then
	    call make_fmt_C (fmt_str)
	    adv = 'YES'
	else
	    write (fmt_str, '(A, I12, A)') '(', part_shape(1), 'A)'
	    adv = 'NO'
	endif


	!! set index array (ind) to zero
	ind = 0



	!! Do all but the last read followed by appropriate skipping.
	!!   We omit the last iteration so that we don't skip past
	!!   the end of the array
	do n = 1, num_reads - 1
		read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))



		do i = 1, DP_MAX_DIMS-1
			!! For each dimension if we're at the end of the
			!!   the range of the dimension, execute the skip
			!!   for that dimension.
			if (MOD(n, sub_sizes(i)) .eq. 0) then



				call dp_skip_elements_C &
                                        (unit, text, dim_skips(i+1), iostat)
				ind(i+1) = ind(i+1)+1
				ind(0:i) = 0
			endif
		enddo

	enddo

	!! Do the last read.
	read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))


	close (unit)

end subroutine DP_READ_ARRAY_4D_C






!!!
!!!  End of COMPLEX array IO routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!	public DP_WRITE_ARRAY_1D_I 
!	public DP_WRITE_ARRAY_2D_I 
!	public DP_WRITE_ARRAY_3D_I 
!	public DP_WRITE_ARRAY_4D_I 
!
!	public DP_READ_ARRAY_1D_I 
!	public DP_READ_ARRAY_2D_I 
!	public DP_READ_ARRAY_3D_I 
!	public DP_READ_ARRAY_4D_I 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of INTEGER array IO routines
!!!






subroutine DP_SKIP_ELEMENTS_I (unit, text, skip, iostat)
integer unit
logical text
integer skip
integer iostat

	INTEGER d
	integer i
	


	if (text)  then
		do i = 1, skip
			read (unit, *, iostat=iostat) d
		enddo
	else

		do i = 1, skip
			read (unit, '(A)', iostat=iostat, advance='NO') d
		enddo

	endif



end subroutine dp_skip_elements_I




subroutine DP_WRITE_ARRAY_1D_I (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
INTEGER arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims = 1
	dims(1) = size(arr,1)

	call dp_write_array_4d_wrapper_I &
		(unit, filename, text, hdr, arr, dims, 1, status, iostat)

end subroutine DP_WRITE_ARRAY_1D_I


subroutine DP_WRITE_ARRAY_2D_I (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
INTEGER arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_I &
		(unit, filename, text, hdr, arr, dims, 2, status, iostat)

end subroutine DP_WRITE_ARRAY_2D_I


subroutine DP_WRITE_ARRAY_3D_I (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
INTEGER arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_I &
		(unit, filename, text, hdr, arr, dims, 3, status, iostat)

end subroutine DP_WRITE_ARRAY_3D_I






subroutine DP_WRITE_ARRAY_4D_WRAPPER_I (unit, filename, text, hdr, arr, dims, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
integer dims(DP_MAX_DIMS)
INTEGER arr(0:dims(1)-1, 0:dims(2)-1, 0:dims(3)-1, 0:dims(4)-1)
integer num_true_dims
integer iostat

    call DP_WRITE_ARRAY_4D_INNER_I &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_WRAPPER_I




subroutine DP_WRITE_ARRAY_4D_I &
	(unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
INTEGER arr(0:,0:,0:,0:)
integer iostat

	call DP_WRITE_ARRAY_4D_INNER_I &
		(unit, filename, text, hdr, arr, 4, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_I


subroutine DP_WRITE_ARRAY_4D_INNER_I &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
INTEGER arr(0:,0:,0:,0:)	! dim dependency
integer num_true_dims
integer iostat


	integer ierr
	logical is_first_rank

	integer*4 stat(MPI_STATUS_SIZE)
	integer next_coords(DP_MAX_DIMS)
	integer next_rank
	integer first_coords(DP_MAX_DIMS)
	integer first_rank

	integer i, j, k, l, jj
	integer iproc, jproc, kproc, lproc
	integer dummy



	INTEGER , allocatable :: local_col(:)
	character*(128) fmt_str


! MUST COORDINATE PROCESSES

	is_first_rank = .true.
	do i = 1, DP_MAX_DIMS
		is_first_rank = is_first_rank .and. (dp_coords(i).eq.0)
	enddo

	if (text) then
		call make_fmt_I (fmt_str)
	endif

	if (dp_num_procs .eq. 1) then

		call dp_open (unit, 'READWRITE', filename, text,  &
			'REWIND', status, iostat)

		if (hdr) then
			call dp_write_hdr &
				(unit, text, shape(arr), num_true_dims, iostat)
		endif

		if (text) then
			write (unit, fmt=fmt_str) arr
		else
			write (fmt_str, '(A, I12, A)') &
				'(', size(arr), 'A)'
			write (unit, fmt=fmt_str, advance='NO') arr
		endif
		call dp_close (unit)

	else ! we have multiple processes

		! The first rank is responsible for all I/O it receives
		! all data to write from other ranks via MPI_RECV
		if (is_first_rank) then
			call dp_open (unit, 'READWRITE', filename, text,  &
				'REWIND', status, iostat)
			if (hdr) then
				call dp_write_hdr (unit, text, shape(arr), &
					num_true_dims, iostat)
			endif

			allocate (local_col(0:size(arr,1)-1))

			!
			! Loop through processors and j & k dimensions
			! within array.  Write out j,kth column from
			! each processor in turn.
			! Note that this could be made more efficient by
			! figuring out when more than a
			! single column is contiguous, and thus can be
			! handled as a unit.
			!
		  DO lproc = 0, dp_proc_arr_size(4) - 1
		    next_coords(4) = lproc

		    DO l = 0, size (arr, 4) - 1

			DO kproc = 0, dp_proc_arr_size(3)-1
			  next_coords(3) = kproc

			  DO k = 0, size (arr,3)-1

			    DO jproc = 0, dp_proc_arr_size(2)-1
			      next_coords(2) = jproc

			      DO j = 0, size (arr,2)-1

			        DO iproc = 0, dp_proc_arr_size(1)-1
			            next_coords(1) = iproc

		    	            call MPI_CART_RANK (dp_comm_cart, &
							next_coords,	&
							next_rank,	&
							ierr)

			            ! get arr (:,j,k,l) from 
				    !     proc at (iproc, jproc, kproc, lproc)

		    	            if (next_rank .eq. dp_rank) then
					if (text) then
					   write (unit, fmt=fmt_str)arr(:,j,k,l)
					else
					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					       advance='NO') arr(:,j,k,l)

					endif
				    else
					!send msg to next_rank
					!requesting column j
					!recv col j into local_col

					call MPI_SENDRECV ( &
						  j,			&
						  1,			&
						  MPI_INTEGER,		&
						  next_rank,		&
						  DP_REQUEST_ARR_COL,	&
						  local_col,		&
						  size(arr,1),		&
						  MPI_INTEGER,		&
						  next_rank,		&
						  DP_ARR_COL,		&
						  dp_comm_cart,		&
						  stat,			&
						  ierr)

					if (text) then
					   write (unit, fmt=fmt_str) local_col
					else

					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					      advance='NO') local_col

					endif
				    endif

			        ENDDO
			      ENDDO
			    ENDDO
			  ENDDO
			ENDDO

		    ENDDO

		  ENDDO

			deallocate (local_col)
			close (unit)
		  

		else  ! we are not the first rank
			first_coords = 0
			call MPI_CART_RANK (dp_comm_cart, first_coords,	&
							first_rank,	&
							ierr)
		    do l = 0, size(arr,4)-1
			do k = 0, size(arr, 3)-1
			    do j = 0, size(arr,2)-1

				! recv msg requesting column
				call MPI_RECV ( &
				  jj,			      &
				  1,			      &
				  MPI_INTEGER,		      &
				  first_rank,		      &
				  DP_REQUEST_ARR_COL, dp_comm_cart, &
				  status, ierr)

				! send arr(:,j,k,l)
				call MPI_SEND (		      &
				  arr(:,j,k,l),		      &
				  size(arr,1),		      &
				  MPI_INTEGER,	      	      &
				  first_rank,	 	      &
				  DP_ARR_COL,		      &
				  dp_comm_cart,		      &
				  ierr)

			    ENDDO !! j
							
			ENDDO  !! k

		    ENDDO  !! l

		endif
	endif
			

end subroutine DP_WRITE_ARRAY_4D_INNER_I



subroutine DP_READ_ARRAY_1D_I (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
INTEGER arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_I &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_1D_I



subroutine DP_READ_ARRAY_2D_I (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
INTEGER arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_I &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_2D_I


subroutine DP_READ_ARRAY_3D_I (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
INTEGER arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_I &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_3D_I



subroutine DP_READ_ARRAY_4D_WRAPPER_I &
	(unit, filename, text, hdr, dims, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
integer dims(DP_MAX_DIMS)
INTEGER arr(0:dims(1)-1,0:dims(2)-1,0:dims(3)-1,0:dims(4)-1)
integer iostat

    call DP_READ_ARRAY_4D_I (unit, filename, text, hdr, arr, iostat)

end subroutine DP_READ_ARRAY_4D_WRAPPER_I


subroutine DP_READ_ARRAY_4D_I (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr			! read a header?
INTEGER arr(0:,0:,0:,0:)
integer iostat

! EACH PROCESS can do this for itself although this approach
! might be slowed down by contention for disk operation.  Then
! again it my be speeded up by good disk caching.

	!! To revise this for higher dimensionality, need only change
	!! DP_MAX_DIMS and the declaration and indexing of arr

	integer*4 status(MPI_STATUS_SIZE)
	integer skip, i, j, k, l, n

	character*(128) fmt_str
	character*(3) adv

	integer part_shape (DP_MAX_DIMS)
	integer full_shape (DP_MAX_DIMS)
	integer full_start_coords (DP_MAX_DIMS)
	integer skip_mult
	
	integer dim_skips (DP_MAX_DIMS+1)
	integer stride
	integer next_pos
	integer ind (DP_MAX_DIMS+1)
	integer sub_sizes (DP_MAX_DIMS+1)
	integer num_reads 
	integer initial_skip
	




	call dp_open (unit, 'READ', filename, text, 'REWIND', 'OLD', iostat)


	part_shape = shape (arr)
	full_shape = dp_proc_arr_size * part_shape

	

	full_start_coords = part_shape*dp_coords

	initial_skip = 0
	skip_mult = 1
	do i = 1, DP_MAX_DIMS
		initial_skip = initial_skip + full_start_coords(i) * skip_mult
		skip_mult = skip_mult * full_shape(i)
	ENDDO


	if (hdr) then
		call dp_skip_header (unit, text, iostat)
	endif



	call dp_skip_elements_I (unit, text, initial_skip, iostat)

	
	!! figure out how many element to skip when going to next
	!! start point in each dimension
	dim_skips = 0
	stride = full_shape(1)
	next_pos = part_shape(1)
	num_reads = 1
	sub_sizes = 1
	do i = 2, DP_MAX_DIMS+1
		sub_sizes(i) = sub_sizes(i-1)*part_shape(i)
		num_reads = num_reads * part_shape(i)
		dim_skips(i) = stride - next_pos
		next_pos = stride * part_shape(i)
		stride = stride * full_shape(i)
	enddo


	if (text) then
	    call make_fmt_I (fmt_str)
	    adv = 'YES'
	else
	    write (fmt_str, '(A, I12, A)') '(', part_shape(1), 'A)'
	    adv = 'NO'
	endif


	!! set index array (ind) to zero
	ind = 0



	!! Do all but the last read followed by appropriate skipping.
	!!   We omit the last iteration so that we don't skip past
	!!   the end of the array
	do n = 1, num_reads - 1
		read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))



		do i = 1, DP_MAX_DIMS-1
			!! For each dimension if we're at the end of the
			!!   the range of the dimension, execute the skip
			!!   for that dimension.
			if (MOD(n, sub_sizes(i)) .eq. 0) then



				call dp_skip_elements_I &
                                        (unit, text, dim_skips(i+1), iostat)
				ind(i+1) = ind(i+1)+1
				ind(0:i) = 0
			endif
		enddo

	enddo

	!! Do the last read.
	read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))


	close (unit)

end subroutine DP_READ_ARRAY_4D_I






!!!
!!!  End of INTEGER array IO routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!	public DP_WRITE_ARRAY_1D_L 
!	public DP_WRITE_ARRAY_2D_L 
!	public DP_WRITE_ARRAY_3D_L 
!	public DP_WRITE_ARRAY_4D_L 
!
!	public DP_READ_ARRAY_1D_L 
!	public DP_READ_ARRAY_2D_L 
!	public DP_READ_ARRAY_3D_L 
!	public DP_READ_ARRAY_4D_L 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of LOGICAL array IO routines
!!!






subroutine DP_SKIP_ELEMENTS_L (unit, text, skip, iostat)
integer unit
logical text
integer skip
integer iostat

	LOGICAL d
	integer i
	


	if (text)  then
		do i = 1, skip
			read (unit, *, iostat=iostat) d
		enddo
	else

		do i = 1, skip
			read (unit, '(A)', iostat=iostat, advance='NO') d
		enddo

	endif



end subroutine dp_skip_elements_L




subroutine DP_WRITE_ARRAY_1D_L (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
LOGICAL arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims = 1
	dims(1) = size(arr,1)

	call dp_write_array_4d_wrapper_L &
		(unit, filename, text, hdr, arr, dims, 1, status, iostat)

end subroutine DP_WRITE_ARRAY_1D_L


subroutine DP_WRITE_ARRAY_2D_L (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
LOGICAL arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_L &
		(unit, filename, text, hdr, arr, dims, 2, status, iostat)

end subroutine DP_WRITE_ARRAY_2D_L


subroutine DP_WRITE_ARRAY_3D_L (unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
LOGICAL arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call dp_write_array_4d_wrapper_L &
		(unit, filename, text, hdr, arr, dims, 3, status, iostat)

end subroutine DP_WRITE_ARRAY_3D_L






subroutine DP_WRITE_ARRAY_4D_WRAPPER_L (unit, filename, text, hdr, arr, dims, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr
integer dims(DP_MAX_DIMS)
LOGICAL arr(0:dims(1)-1, 0:dims(2)-1, 0:dims(3)-1, 0:dims(4)-1)
integer num_true_dims
integer iostat

    call DP_WRITE_ARRAY_4D_INNER_L &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_WRAPPER_L




subroutine DP_WRITE_ARRAY_4D_L &
	(unit, filename, text, hdr, arr, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
LOGICAL arr(0:,0:,0:,0:)
integer iostat

	call DP_WRITE_ARRAY_4D_INNER_L &
		(unit, filename, text, hdr, arr, 4, status, iostat)

end subroutine DP_WRITE_ARRAY_4D_L


subroutine DP_WRITE_ARRAY_4D_INNER_L &
	(unit, filename, text, hdr, arr, num_true_dims, status, iostat)
integer unit
character*(*) filename
logical text
character*(*) status
logical hdr			! write a header?
LOGICAL arr(0:,0:,0:,0:)	! dim dependency
integer num_true_dims
integer iostat


	integer ierr
	logical is_first_rank

	integer*4 stat(MPI_STATUS_SIZE)
	integer next_coords(DP_MAX_DIMS)
	integer next_rank
	integer first_coords(DP_MAX_DIMS)
	integer first_rank

	integer i, j, k, l, jj
	integer iproc, jproc, kproc, lproc
	integer dummy



	LOGICAL , allocatable :: local_col(:)
	character*(128) fmt_str


! MUST COORDINATE PROCESSES

	is_first_rank = .true.
	do i = 1, DP_MAX_DIMS
		is_first_rank = is_first_rank .and. (dp_coords(i).eq.0)
	enddo

	if (text) then
		call make_fmt_L (fmt_str)
	endif

	if (dp_num_procs .eq. 1) then

		call dp_open (unit, 'READWRITE', filename, text,  &
			'REWIND', status, iostat)

		if (hdr) then
			call dp_write_hdr &
				(unit, text, shape(arr), num_true_dims, iostat)
		endif

		if (text) then
			write (unit, fmt=fmt_str) arr
		else
			write (fmt_str, '(A, I12, A)') &
				'(', size(arr), 'A)'
			write (unit, fmt=fmt_str, advance='NO') arr
		endif
		call dp_close (unit)

	else ! we have multiple processes

		! The first rank is responsible for all I/O it receives
		! all data to write from other ranks via MPI_RECV
		if (is_first_rank) then
			call dp_open (unit, 'READWRITE', filename, text,  &
				'REWIND', status, iostat)
			if (hdr) then
				call dp_write_hdr (unit, text, shape(arr), &
					num_true_dims, iostat)
			endif

			allocate (local_col(0:size(arr,1)-1))

			!
			! Loop through processors and j & k dimensions
			! within array.  Write out j,kth column from
			! each processor in turn.
			! Note that this could be made more efficient by
			! figuring out when more than a
			! single column is contiguous, and thus can be
			! handled as a unit.
			!
		  DO lproc = 0, dp_proc_arr_size(4) - 1
		    next_coords(4) = lproc

		    DO l = 0, size (arr, 4) - 1

			DO kproc = 0, dp_proc_arr_size(3)-1
			  next_coords(3) = kproc

			  DO k = 0, size (arr,3)-1

			    DO jproc = 0, dp_proc_arr_size(2)-1
			      next_coords(2) = jproc

			      DO j = 0, size (arr,2)-1

			        DO iproc = 0, dp_proc_arr_size(1)-1
			            next_coords(1) = iproc

		    	            call MPI_CART_RANK (dp_comm_cart, &
							next_coords,	&
							next_rank,	&
							ierr)

			            ! get arr (:,j,k,l) from 
				    !     proc at (iproc, jproc, kproc, lproc)

		    	            if (next_rank .eq. dp_rank) then
					if (text) then
					   write (unit, fmt=fmt_str)arr(:,j,k,l)
					else
					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					       advance='NO') arr(:,j,k,l)

					endif
				    else
					!send msg to next_rank
					!requesting column j
					!recv col j into local_col

					call MPI_SENDRECV ( &
						  j,			&
						  1,			&
						  MPI_INTEGER,		&
						  next_rank,		&
						  DP_REQUEST_ARR_COL,	&
						  local_col,		&
						  size(arr,1),		&
						  MPI_LOGICAL,		&
						  next_rank,		&
						  DP_ARR_COL,		&
						  dp_comm_cart,		&
						  stat,			&
						  ierr)

					if (text) then
					   write (unit, fmt=fmt_str) local_col
					else

					   write (fmt_str, &
					     '(A, I12, A)') &
					     '(',  &
					      size(arr,1),&
					      'A)'
					   write (unit, fmt=fmt_str, &
					      advance='NO') local_col

					endif
				    endif

			        ENDDO
			      ENDDO
			    ENDDO
			  ENDDO
			ENDDO

		    ENDDO

		  ENDDO

			deallocate (local_col)
			close (unit)
		  

		else  ! we are not the first rank
			first_coords = 0
			call MPI_CART_RANK (dp_comm_cart, first_coords,	&
							first_rank,	&
							ierr)
		    do l = 0, size(arr,4)-1
			do k = 0, size(arr, 3)-1
			    do j = 0, size(arr,2)-1

				! recv msg requesting column
				call MPI_RECV ( &
				  jj,			      &
				  1,			      &
				  MPI_INTEGER,		      &
				  first_rank,		      &
				  DP_REQUEST_ARR_COL, dp_comm_cart, &
				  status, ierr)

				! send arr(:,j,k,l)
				call MPI_SEND (		      &
				  arr(:,j,k,l),		      &
				  size(arr,1),		      &
				  MPI_LOGICAL,	      	      &
				  first_rank,	 	      &
				  DP_ARR_COL,		      &
				  dp_comm_cart,		      &
				  ierr)

			    ENDDO !! j
							
			ENDDO  !! k

		    ENDDO  !! l

		endif
	endif
			

end subroutine DP_WRITE_ARRAY_4D_INNER_L



subroutine DP_READ_ARRAY_1D_L (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
LOGICAL arr(0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_L &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_1D_L



subroutine DP_READ_ARRAY_2D_L (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
LOGICAL arr(0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_L &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_2D_L


subroutine DP_READ_ARRAY_3D_L (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
LOGICAL arr(0:,0:,0:)
integer iostat

	integer dims(DP_MAX_DIMS)

	dims(1) = size(arr,1)
	dims(2) = size(arr,2)
	dims(3) = size(arr,3)
	dims(4:DP_MAX_DIMS) = 1

	call DP_READ_ARRAY_4D_WRAPPER_L &
		(unit, filename, text, hdr, dims, arr, iostat)

end subroutine DP_READ_ARRAY_3D_L



subroutine DP_READ_ARRAY_4D_WRAPPER_L &
	(unit, filename, text, hdr, dims, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr
integer dims(DP_MAX_DIMS)
LOGICAL arr(0:dims(1)-1,0:dims(2)-1,0:dims(3)-1,0:dims(4)-1)
integer iostat

    call DP_READ_ARRAY_4D_L (unit, filename, text, hdr, arr, iostat)

end subroutine DP_READ_ARRAY_4D_WRAPPER_L


subroutine DP_READ_ARRAY_4D_L (unit, filename, text, hdr, arr, iostat)
integer unit
character*(*) filename
logical text
logical hdr			! read a header?
LOGICAL arr(0:,0:,0:,0:)
integer iostat

! EACH PROCESS can do this for itself although this approach
! might be slowed down by contention for disk operation.  Then
! again it my be speeded up by good disk caching.

	!! To revise this for higher dimensionality, need only change
	!! DP_MAX_DIMS and the declaration and indexing of arr

	integer*4 status(MPI_STATUS_SIZE)
	integer skip, i, j, k, l, n

	character*(128) fmt_str
	character*(3) adv

	integer part_shape (DP_MAX_DIMS)
	integer full_shape (DP_MAX_DIMS)
	integer full_start_coords (DP_MAX_DIMS)
	integer skip_mult
	
	integer dim_skips (DP_MAX_DIMS+1)
	integer stride
	integer next_pos
	integer ind (DP_MAX_DIMS+1)
	integer sub_sizes (DP_MAX_DIMS+1)
	integer num_reads 
	integer initial_skip
	




	call dp_open (unit, 'READ', filename, text, 'REWIND', 'OLD', iostat)


	part_shape = shape (arr)
	full_shape = dp_proc_arr_size * part_shape

	

	full_start_coords = part_shape*dp_coords

	initial_skip = 0
	skip_mult = 1
	do i = 1, DP_MAX_DIMS
		initial_skip = initial_skip + full_start_coords(i) * skip_mult
		skip_mult = skip_mult * full_shape(i)
	ENDDO


	if (hdr) then
		call dp_skip_header (unit, text, iostat)
	endif



	call dp_skip_elements_L (unit, text, initial_skip, iostat)

	
	!! figure out how many element to skip when going to next
	!! start point in each dimension
	dim_skips = 0
	stride = full_shape(1)
	next_pos = part_shape(1)
	num_reads = 1
	sub_sizes = 1
	do i = 2, DP_MAX_DIMS+1
		sub_sizes(i) = sub_sizes(i-1)*part_shape(i)
		num_reads = num_reads * part_shape(i)
		dim_skips(i) = stride - next_pos
		next_pos = stride * part_shape(i)
		stride = stride * full_shape(i)
	enddo


	if (text) then
	    call make_fmt_L (fmt_str)
	    adv = 'YES'
	else
	    write (fmt_str, '(A, I12, A)') '(', part_shape(1), 'A)'
	    adv = 'NO'
	endif


	!! set index array (ind) to zero
	ind = 0



	!! Do all but the last read followed by appropriate skipping.
	!!   We omit the last iteration so that we don't skip past
	!!   the end of the array
	do n = 1, num_reads - 1
		read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))



		do i = 1, DP_MAX_DIMS-1
			!! For each dimension if we're at the end of the
			!!   the range of the dimension, execute the skip
			!!   for that dimension.
			if (MOD(n, sub_sizes(i)) .eq. 0) then



				call dp_skip_elements_L &
                                        (unit, text, dim_skips(i+1), iostat)
				ind(i+1) = ind(i+1)+1
				ind(0:i) = 0
			endif
		enddo

	enddo

	!! Do the last read.
	read(unit, fmt=fmt_str, advance=adv, iostat=iostat) &
				arr(:,ind(1),ind(2),ind(3))


	close (unit)

end subroutine DP_READ_ARRAY_4D_L






!!!
!!!  End of LOGICAL array IO routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

