
!		module procedure dp_maxval_1D_R 
!		module procedure dp_minval_1D_R 
!		module procedure dp_maxloc_1D_R 
!		module procedure dp_minloc_1D_R 

!		public dp_maxval_1D_R 
!		public dp_minval_1D_R 
!		public dp_maxloc_1D_R 
!		public dp_minloc_1D_R 


FUNCTION dp_maxval_1D_R (arr, mask) RESULT (reduction)
REAL reduction
REAL arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	REAL partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_REAL, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_1D_R


FUNCTION dp_minval_1D_R (arr, mask) RESULT (reduction)
REAL reduction
REAL arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	REAL partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_REAL, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_1D_R


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_1D_R 

FUNCTION dp_maxloc_1D_R (arr, mask) RESULT (result_loc)
integer result_loc(1)
REAL, intent(IN) :: arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	REAL partial, complete
	integer local_loc(1)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(1+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 1
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 1
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 1, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_1D_R




FUNCTION dp_minloc_1D_R (arr, mask) RESULT (result_loc)
integer result_loc(1)
REAL, intent(IN) :: arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	REAL partial, complete
	integer local_loc(1)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(1+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 1
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 1
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 1, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_1D_R






!		module procedure dp_maxval_2D_R 
!		module procedure dp_minval_2D_R 
!		module procedure dp_maxloc_2D_R 
!		module procedure dp_minloc_2D_R 

!		public dp_maxval_2D_R 
!		public dp_minval_2D_R 
!		public dp_maxloc_2D_R 
!		public dp_minloc_2D_R 


FUNCTION dp_maxval_2D_R (arr, mask) RESULT (reduction)
REAL reduction
REAL arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	REAL partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_REAL, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_2D_R


FUNCTION dp_minval_2D_R (arr, mask) RESULT (reduction)
REAL reduction
REAL arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	REAL partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_REAL, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_2D_R


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_2D_R 

FUNCTION dp_maxloc_2D_R (arr, mask) RESULT (result_loc)
integer result_loc(2)
REAL, intent(IN) :: arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	REAL partial, complete
	integer local_loc(2)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(2+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 2
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 2
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 2, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_2D_R




FUNCTION dp_minloc_2D_R (arr, mask) RESULT (result_loc)
integer result_loc(2)
REAL, intent(IN) :: arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	REAL partial, complete
	integer local_loc(2)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(2+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 2
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 2
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 2, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_2D_R






!		module procedure dp_maxval_3D_R 
!		module procedure dp_minval_3D_R 
!		module procedure dp_maxloc_3D_R 
!		module procedure dp_minloc_3D_R 

!		public dp_maxval_3D_R 
!		public dp_minval_3D_R 
!		public dp_maxloc_3D_R 
!		public dp_minloc_3D_R 


FUNCTION dp_maxval_3D_R (arr, mask) RESULT (reduction)
REAL reduction
REAL arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	REAL partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_REAL, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_3D_R


FUNCTION dp_minval_3D_R (arr, mask) RESULT (reduction)
REAL reduction
REAL arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	REAL partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_REAL, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_3D_R


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_3D_R 

FUNCTION dp_maxloc_3D_R (arr, mask) RESULT (result_loc)
integer result_loc(3)
REAL, intent(IN) :: arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	REAL partial, complete
	integer local_loc(3)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(3+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 3
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 3
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 3, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_3D_R




FUNCTION dp_minloc_3D_R (arr, mask) RESULT (result_loc)
integer result_loc(3)
REAL, intent(IN) :: arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	REAL partial, complete
	integer local_loc(3)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(3+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 3
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 3
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 3, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_3D_R






!		module procedure dp_maxval_4D_R 
!		module procedure dp_minval_4D_R 
!		module procedure dp_maxloc_4D_R 
!		module procedure dp_minloc_4D_R 

!		public dp_maxval_4D_R 
!		public dp_minval_4D_R 
!		public dp_maxloc_4D_R 
!		public dp_minloc_4D_R 


FUNCTION dp_maxval_4D_R (arr, mask) RESULT (reduction)
REAL reduction
REAL arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	REAL partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_REAL, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_4D_R


FUNCTION dp_minval_4D_R (arr, mask) RESULT (reduction)
REAL reduction
REAL arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	REAL partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_REAL, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_4D_R


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_4D_R 

FUNCTION dp_maxloc_4D_R (arr, mask) RESULT (result_loc)
integer result_loc(4)
REAL, intent(IN) :: arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	REAL partial, complete
	integer local_loc(4)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(4+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 4
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 4
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 4, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_4D_R




FUNCTION dp_minloc_4D_R (arr, mask) RESULT (result_loc)
integer result_loc(4)
REAL, intent(IN) :: arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	REAL partial, complete
	integer local_loc(4)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(4+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 4
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 4
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 4, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_4D_R






!		module procedure dp_maxval_1D_D 
!		module procedure dp_minval_1D_D 
!		module procedure dp_maxloc_1D_D 
!		module procedure dp_minloc_1D_D 

!		public dp_maxval_1D_D 
!		public dp_minval_1D_D 
!		public dp_maxloc_1D_D 
!		public dp_minloc_1D_D 


FUNCTION dp_maxval_1D_D (arr, mask) RESULT (reduction)
DOUBLE PRECISION reduction
DOUBLE PRECISION arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	DOUBLE PRECISION partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_DOUBLE_PRECISION, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_1D_D


FUNCTION dp_minval_1D_D (arr, mask) RESULT (reduction)
DOUBLE PRECISION reduction
DOUBLE PRECISION arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	DOUBLE PRECISION partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_DOUBLE_PRECISION, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_1D_D


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_1D_D 

FUNCTION dp_maxloc_1D_D (arr, mask) RESULT (result_loc)
integer result_loc(1)
DOUBLE PRECISION, intent(IN) :: arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	DOUBLE PRECISION partial, complete
	integer local_loc(1)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(1+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 1
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 1
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 1, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_1D_D




FUNCTION dp_minloc_1D_D (arr, mask) RESULT (result_loc)
integer result_loc(1)
DOUBLE PRECISION, intent(IN) :: arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	DOUBLE PRECISION partial, complete
	integer local_loc(1)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(1+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 1
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 1
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 1, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_1D_D






!		module procedure dp_maxval_2D_D 
!		module procedure dp_minval_2D_D 
!		module procedure dp_maxloc_2D_D 
!		module procedure dp_minloc_2D_D 

!		public dp_maxval_2D_D 
!		public dp_minval_2D_D 
!		public dp_maxloc_2D_D 
!		public dp_minloc_2D_D 


FUNCTION dp_maxval_2D_D (arr, mask) RESULT (reduction)
DOUBLE PRECISION reduction
DOUBLE PRECISION arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	DOUBLE PRECISION partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_DOUBLE_PRECISION, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_2D_D


FUNCTION dp_minval_2D_D (arr, mask) RESULT (reduction)
DOUBLE PRECISION reduction
DOUBLE PRECISION arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	DOUBLE PRECISION partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_DOUBLE_PRECISION, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_2D_D


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_2D_D 

FUNCTION dp_maxloc_2D_D (arr, mask) RESULT (result_loc)
integer result_loc(2)
DOUBLE PRECISION, intent(IN) :: arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	DOUBLE PRECISION partial, complete
	integer local_loc(2)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(2+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 2
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 2
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 2, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_2D_D




FUNCTION dp_minloc_2D_D (arr, mask) RESULT (result_loc)
integer result_loc(2)
DOUBLE PRECISION, intent(IN) :: arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	DOUBLE PRECISION partial, complete
	integer local_loc(2)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(2+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 2
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 2
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 2, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_2D_D






!		module procedure dp_maxval_3D_D 
!		module procedure dp_minval_3D_D 
!		module procedure dp_maxloc_3D_D 
!		module procedure dp_minloc_3D_D 

!		public dp_maxval_3D_D 
!		public dp_minval_3D_D 
!		public dp_maxloc_3D_D 
!		public dp_minloc_3D_D 


FUNCTION dp_maxval_3D_D (arr, mask) RESULT (reduction)
DOUBLE PRECISION reduction
DOUBLE PRECISION arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	DOUBLE PRECISION partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_DOUBLE_PRECISION, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_3D_D


FUNCTION dp_minval_3D_D (arr, mask) RESULT (reduction)
DOUBLE PRECISION reduction
DOUBLE PRECISION arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	DOUBLE PRECISION partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_DOUBLE_PRECISION, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_3D_D


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_3D_D 

FUNCTION dp_maxloc_3D_D (arr, mask) RESULT (result_loc)
integer result_loc(3)
DOUBLE PRECISION, intent(IN) :: arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	DOUBLE PRECISION partial, complete
	integer local_loc(3)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(3+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 3
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 3
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 3, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_3D_D




FUNCTION dp_minloc_3D_D (arr, mask) RESULT (result_loc)
integer result_loc(3)
DOUBLE PRECISION, intent(IN) :: arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	DOUBLE PRECISION partial, complete
	integer local_loc(3)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(3+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 3
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 3
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 3, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_3D_D






!		module procedure dp_maxval_4D_D 
!		module procedure dp_minval_4D_D 
!		module procedure dp_maxloc_4D_D 
!		module procedure dp_minloc_4D_D 

!		public dp_maxval_4D_D 
!		public dp_minval_4D_D 
!		public dp_maxloc_4D_D 
!		public dp_minloc_4D_D 


FUNCTION dp_maxval_4D_D (arr, mask) RESULT (reduction)
DOUBLE PRECISION reduction
DOUBLE PRECISION arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	DOUBLE PRECISION partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_DOUBLE_PRECISION, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_4D_D


FUNCTION dp_minval_4D_D (arr, mask) RESULT (reduction)
DOUBLE PRECISION reduction
DOUBLE PRECISION arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	DOUBLE PRECISION partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_DOUBLE_PRECISION, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_4D_D


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_4D_D 

FUNCTION dp_maxloc_4D_D (arr, mask) RESULT (result_loc)
integer result_loc(4)
DOUBLE PRECISION, intent(IN) :: arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	DOUBLE PRECISION partial, complete
	integer local_loc(4)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(4+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 4
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 4
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 4, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_4D_D




FUNCTION dp_minloc_4D_D (arr, mask) RESULT (result_loc)
integer result_loc(4)
DOUBLE PRECISION, intent(IN) :: arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	DOUBLE PRECISION partial, complete
	integer local_loc(4)
	integer ierr
	DOUBLE PRECISION in(2), out(2)
	integer global_index
	integer stride(4+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 4
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 4
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2DOUBLE_PRECISION, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 4, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_4D_D






!		module procedure dp_maxval_1D_I 
!		module procedure dp_minval_1D_I 
!		module procedure dp_maxloc_1D_I 
!		module procedure dp_minloc_1D_I 

!		public dp_maxval_1D_I 
!		public dp_minval_1D_I 
!		public dp_maxloc_1D_I 
!		public dp_minloc_1D_I 


FUNCTION dp_maxval_1D_I (arr, mask) RESULT (reduction)
INTEGER reduction
INTEGER arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	INTEGER partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_INTEGER, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_1D_I


FUNCTION dp_minval_1D_I (arr, mask) RESULT (reduction)
INTEGER reduction
INTEGER arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	INTEGER partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_INTEGER, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_1D_I


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_1D_I 

FUNCTION dp_maxloc_1D_I (arr, mask) RESULT (result_loc)
integer result_loc(1)
INTEGER, intent(IN) :: arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	INTEGER partial, complete
	integer local_loc(1)
	integer ierr
	INTEGER in(2), out(2)
	integer global_index
	integer stride(1+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 1
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 1
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2INTEGER, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 1, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_1D_I




FUNCTION dp_minloc_1D_I (arr, mask) RESULT (result_loc)
integer result_loc(1)
INTEGER, intent(IN) :: arr (0:)
logical, intent(IN), OPTIONAL :: mask (0:)

	INTEGER partial, complete
	integer local_loc(1)
	integer ierr
	INTEGER in(2), out(2)
	integer global_index
	integer stride(1+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 1
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 1
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2INTEGER, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 1, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_1D_I






!		module procedure dp_maxval_2D_I 
!		module procedure dp_minval_2D_I 
!		module procedure dp_maxloc_2D_I 
!		module procedure dp_minloc_2D_I 

!		public dp_maxval_2D_I 
!		public dp_minval_2D_I 
!		public dp_maxloc_2D_I 
!		public dp_minloc_2D_I 


FUNCTION dp_maxval_2D_I (arr, mask) RESULT (reduction)
INTEGER reduction
INTEGER arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	INTEGER partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_INTEGER, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_2D_I


FUNCTION dp_minval_2D_I (arr, mask) RESULT (reduction)
INTEGER reduction
INTEGER arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	INTEGER partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_INTEGER, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_2D_I


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_2D_I 

FUNCTION dp_maxloc_2D_I (arr, mask) RESULT (result_loc)
integer result_loc(2)
INTEGER, intent(IN) :: arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	INTEGER partial, complete
	integer local_loc(2)
	integer ierr
	INTEGER in(2), out(2)
	integer global_index
	integer stride(2+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 2
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 2
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2INTEGER, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 2, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_2D_I




FUNCTION dp_minloc_2D_I (arr, mask) RESULT (result_loc)
integer result_loc(2)
INTEGER, intent(IN) :: arr (0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:)

	INTEGER partial, complete
	integer local_loc(2)
	integer ierr
	INTEGER in(2), out(2)
	integer global_index
	integer stride(2+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 2
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 2
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2INTEGER, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 2, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_2D_I






!		module procedure dp_maxval_3D_I 
!		module procedure dp_minval_3D_I 
!		module procedure dp_maxloc_3D_I 
!		module procedure dp_minloc_3D_I 

!		public dp_maxval_3D_I 
!		public dp_minval_3D_I 
!		public dp_maxloc_3D_I 
!		public dp_minloc_3D_I 


FUNCTION dp_maxval_3D_I (arr, mask) RESULT (reduction)
INTEGER reduction
INTEGER arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	INTEGER partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_INTEGER, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_3D_I


FUNCTION dp_minval_3D_I (arr, mask) RESULT (reduction)
INTEGER reduction
INTEGER arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	INTEGER partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_INTEGER, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_3D_I


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_3D_I 

FUNCTION dp_maxloc_3D_I (arr, mask) RESULT (result_loc)
integer result_loc(3)
INTEGER, intent(IN) :: arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	INTEGER partial, complete
	integer local_loc(3)
	integer ierr
	INTEGER in(2), out(2)
	integer global_index
	integer stride(3+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 3
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 3
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2INTEGER, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 3, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_3D_I




FUNCTION dp_minloc_3D_I (arr, mask) RESULT (result_loc)
integer result_loc(3)
INTEGER, intent(IN) :: arr (0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:)

	INTEGER partial, complete
	integer local_loc(3)
	integer ierr
	INTEGER in(2), out(2)
	integer global_index
	integer stride(3+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 3
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 3
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2INTEGER, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 3, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_3D_I






!		module procedure dp_maxval_4D_I 
!		module procedure dp_minval_4D_I 
!		module procedure dp_maxloc_4D_I 
!		module procedure dp_minloc_4D_I 

!		public dp_maxval_4D_I 
!		public dp_minval_4D_I 
!		public dp_maxloc_4D_I 
!		public dp_minloc_4D_I 


FUNCTION dp_maxval_4D_I (arr, mask) RESULT (reduction)
INTEGER reduction
INTEGER arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	INTEGER partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MAXVAL (arr, mask=mask)

	else

		partial = MAXVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_INTEGER, &
					MPI_MAX, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_maxval_4D_I


FUNCTION dp_minval_4D_I (arr, mask) RESULT (reduction)
INTEGER reduction
INTEGER arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	INTEGER partial, complete
	integer ierr

	if (dp_num_procs .eq. 1) then

		reduction = MINVAL (arr, mask=mask)

	else

		partial = MINVAL (arr, mask=mask)

		call MPI_ALLREDUCE (partial, complete, 1, &
					MPI_INTEGER, &
					MPI_MIN, dp_comm_cart, ierr)
		! should check ierr

		reduction = complete
	endif

END FUNCTION dp_minval_4D_I


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! should handle other logical types
!! does this properly handle optional mask?
!! functions above should have mask
!! does this handle different array bases the same way that MAXLOC does?

!	dp_maxloc_4D_I 

FUNCTION dp_maxloc_4D_I (arr, mask) RESULT (result_loc)
integer result_loc(4)
INTEGER, intent(IN) :: arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	INTEGER partial, complete
	integer local_loc(4)
	integer ierr
	INTEGER in(2), out(2)
	integer global_index
	integer stride(4+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MAXLOC (arr, mask=mask)

	else

        	in(1) = MAXVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif


		if ( use_mask ) then

       			local_loc = MAXLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 4
			    global_index = global_index + &
		  	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
			    stride(i) * size(arr,i) * dp_proc_arr_size(i)

			enddo

			in(2) = global_index

		else

			stride(1) = 1
			do i = 1, 4
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1
		endif


		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2INTEGER, &
					MPI_MAXLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 4, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo

			result_loc = result_loc + 1
		else
			result_loc = 0
		endif


	endif

END FUNCTION dp_maxloc_4D_I




FUNCTION dp_minloc_4D_I (arr, mask) RESULT (result_loc)
integer result_loc(4)
INTEGER, intent(IN) :: arr (0:,0:,0:,0:)
logical, intent(IN), OPTIONAL :: mask (0:,0:,0:,0:)

	INTEGER partial, complete
	integer local_loc(4)
	integer ierr
	INTEGER in(2), out(2)
	integer global_index
	integer stride(4+1)
	integer i
	logical use_mask

	if (dp_num_procs .eq. 1) then

		result_loc = MINLOC (arr, mask=mask)

	else

       		in(1) = MINVAL (arr, mask=mask)

		if (present(mask)) then
			use_mask = any (mask)
		else
			use_mask = .TRUE.
		endif

		if ( use_mask ) then

       			local_loc = MINLOC (arr, mask=mask) - 1

			
			global_index = 0
			stride(1) = 1
			do i = 1, 4
			    global_index = global_index + &
		   	    stride(i) * (local_loc(i)+size(arr,i)*dp_coords(i))
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = global_index
		else
			stride(1) = 1
			do i = 1, 4
			    stride(i+1) = &
		 	    stride(i) * size(arr,i) * dp_proc_arr_size(i)
			enddo

			in(2) = -1

		endif

		call MPI_ALLREDUCE (in, out, 1, &
					MPI_2INTEGER, &
					MPI_MINLOC, dp_comm_cart, ierr)
		! should check ierr

		
		global_index = out(2)

		if (global_index .ne. -1) then
			do i = 4, 1, -1
				result_loc(i) = global_index / stride(i)
				global_index = MOD (global_index, stride(i))
			enddo
			result_loc = result_loc + 1
		else
			result_loc = 0
		endif
	

	endif

END FUNCTION dp_minloc_4D_I





