
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 1D REAL array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 1D REAL array.
!!
!!	public DP_CSHIFT_1D_R 
!!
FUNCTION DP_CSHIFT_1D_R (arr, shift, dim) &
                                                       RESULT (arr_out)
	REAL, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	REAL arr_out (0:size(arr,1)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_R &
			(arr, dist, ierr)
		call do_shift_R  &
		   (offset, .TRUE., 0.0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_1D_R








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 1D REAL array.
!!
!!	public DP_EOSHIFT_1D_R 
!!
FUNCTION DP_EOSHIFT_1D_R &
		(arr, shift, boundary, dim) RESULT (arr_out)
	REAL, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	REAL, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	REAL arr_out &
		(0:size(arr,1)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_R &
			(arr, dist, ierr)
		call do_shift_R  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_1D_R


!!!
!!!  End of 1D REAL array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 2D REAL array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 2D REAL array.
!!
!!	public DP_CSHIFT_2D_R 
!!
FUNCTION DP_CSHIFT_2D_R (arr, shift, dim) &
                                                       RESULT (arr_out)
	REAL, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	REAL arr_out (0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_R &
			(arr, dist, ierr)
		call do_shift_R  &
		   (offset, .TRUE., 0.0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_2D_R








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 2D REAL array.
!!
!!	public DP_EOSHIFT_2D_R 
!!
FUNCTION DP_EOSHIFT_2D_R &
		(arr, shift, boundary, dim) RESULT (arr_out)
	REAL, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	REAL, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	REAL arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_R &
			(arr, dist, ierr)
		call do_shift_R  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_2D_R


!!!
!!!  End of 2D REAL array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 3D REAL array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 3D REAL array.
!!
!!	public DP_CSHIFT_3D_R 
!!
FUNCTION DP_CSHIFT_3D_R (arr, shift, dim) &
                                                       RESULT (arr_out)
	REAL, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	REAL arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_R &
			(arr, dist, ierr)
		call do_shift_R  &
		   (offset, .TRUE., 0.0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_3D_R








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 3D REAL array.
!!
!!	public DP_EOSHIFT_3D_R 
!!
FUNCTION DP_EOSHIFT_3D_R &
		(arr, shift, boundary, dim) RESULT (arr_out)
	REAL, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	REAL, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	REAL arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_R &
			(arr, dist, ierr)
		call do_shift_R  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_3D_R


!!!
!!!  End of 3D REAL array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 4D REAL array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 4D REAL array.
!!
!!	public DP_CSHIFT_4D_R 
!!
FUNCTION DP_CSHIFT_4D_R (arr, shift, dim) &
                                                       RESULT (arr_out)
	REAL, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	REAL arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_R &
			(arr, dist, ierr)
		call do_shift_R  &
		   (offset, .TRUE., 0.0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_4D_R








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 4D REAL array.
!!
!!	public DP_EOSHIFT_4D_R 
!!
FUNCTION DP_EOSHIFT_4D_R &
		(arr, shift, boundary, dim) RESULT (arr_out)
	REAL, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	REAL, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	REAL arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_R &
			(arr, dist, ierr)
		call do_shift_R  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_4D_R


!!!
!!!  End of 4D REAL array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 1D DOUBLE PRECISION array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 1D DOUBLE PRECISION array.
!!
!!	public DP_CSHIFT_1D_D 
!!
FUNCTION DP_CSHIFT_1D_D (arr, shift, dim) &
                                                       RESULT (arr_out)
	DOUBLE PRECISION, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	DOUBLE PRECISION arr_out (0:size(arr,1)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_D &
			(arr, dist, ierr)
		call do_shift_D  &
		   (offset, .TRUE., 0.0D0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_1D_D








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 1D DOUBLE PRECISION array.
!!
!!	public DP_EOSHIFT_1D_D 
!!
FUNCTION DP_EOSHIFT_1D_D &
		(arr, shift, boundary, dim) RESULT (arr_out)
	DOUBLE PRECISION, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	DOUBLE PRECISION, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	DOUBLE PRECISION arr_out &
		(0:size(arr,1)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_D &
			(arr, dist, ierr)
		call do_shift_D  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_1D_D


!!!
!!!  End of 1D DOUBLE PRECISION array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 2D DOUBLE PRECISION array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 2D DOUBLE PRECISION array.
!!
!!	public DP_CSHIFT_2D_D 
!!
FUNCTION DP_CSHIFT_2D_D (arr, shift, dim) &
                                                       RESULT (arr_out)
	DOUBLE PRECISION, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	DOUBLE PRECISION arr_out (0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_D &
			(arr, dist, ierr)
		call do_shift_D  &
		   (offset, .TRUE., 0.0D0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_2D_D








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 2D DOUBLE PRECISION array.
!!
!!	public DP_EOSHIFT_2D_D 
!!
FUNCTION DP_EOSHIFT_2D_D &
		(arr, shift, boundary, dim) RESULT (arr_out)
	DOUBLE PRECISION, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	DOUBLE PRECISION, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	DOUBLE PRECISION arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_D &
			(arr, dist, ierr)
		call do_shift_D  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_2D_D


!!!
!!!  End of 2D DOUBLE PRECISION array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 3D DOUBLE PRECISION array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 3D DOUBLE PRECISION array.
!!
!!	public DP_CSHIFT_3D_D 
!!
FUNCTION DP_CSHIFT_3D_D (arr, shift, dim) &
                                                       RESULT (arr_out)
	DOUBLE PRECISION, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	DOUBLE PRECISION arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_D &
			(arr, dist, ierr)
		call do_shift_D  &
		   (offset, .TRUE., 0.0D0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_3D_D








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 3D DOUBLE PRECISION array.
!!
!!	public DP_EOSHIFT_3D_D 
!!
FUNCTION DP_EOSHIFT_3D_D &
		(arr, shift, boundary, dim) RESULT (arr_out)
	DOUBLE PRECISION, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	DOUBLE PRECISION, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	DOUBLE PRECISION arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_D &
			(arr, dist, ierr)
		call do_shift_D  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_3D_D


!!!
!!!  End of 3D DOUBLE PRECISION array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 4D DOUBLE PRECISION array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 4D DOUBLE PRECISION array.
!!
!!	public DP_CSHIFT_4D_D 
!!
FUNCTION DP_CSHIFT_4D_D (arr, shift, dim) &
                                                       RESULT (arr_out)
	DOUBLE PRECISION, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	DOUBLE PRECISION arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_D &
			(arr, dist, ierr)
		call do_shift_D  &
		   (offset, .TRUE., 0.0D0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_4D_D








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 4D DOUBLE PRECISION array.
!!
!!	public DP_EOSHIFT_4D_D 
!!
FUNCTION DP_EOSHIFT_4D_D &
		(arr, shift, boundary, dim) RESULT (arr_out)
	DOUBLE PRECISION, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	DOUBLE PRECISION, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	DOUBLE PRECISION arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_D &
			(arr, dist, ierr)
		call do_shift_D  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_4D_D


!!!
!!!  End of 4D DOUBLE PRECISION array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 1D COMPLEX array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 1D COMPLEX array.
!!
!!	public DP_CSHIFT_1D_C 
!!
FUNCTION DP_CSHIFT_1D_C (arr, shift, dim) &
                                                       RESULT (arr_out)
	COMPLEX, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	COMPLEX arr_out (0:size(arr,1)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_C &
			(arr, dist, ierr)
		call do_shift_C  &
		   (offset, .TRUE., (0.0,0.0), arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_1D_C








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 1D COMPLEX array.
!!
!!	public DP_EOSHIFT_1D_C 
!!
FUNCTION DP_EOSHIFT_1D_C &
		(arr, shift, boundary, dim) RESULT (arr_out)
	COMPLEX, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	COMPLEX, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	COMPLEX arr_out &
		(0:size(arr,1)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_C &
			(arr, dist, ierr)
		call do_shift_C  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_1D_C


!!!
!!!  End of 1D COMPLEX array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 2D COMPLEX array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 2D COMPLEX array.
!!
!!	public DP_CSHIFT_2D_C 
!!
FUNCTION DP_CSHIFT_2D_C (arr, shift, dim) &
                                                       RESULT (arr_out)
	COMPLEX, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	COMPLEX arr_out (0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_C &
			(arr, dist, ierr)
		call do_shift_C  &
		   (offset, .TRUE., (0.0,0.0), arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_2D_C








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 2D COMPLEX array.
!!
!!	public DP_EOSHIFT_2D_C 
!!
FUNCTION DP_EOSHIFT_2D_C &
		(arr, shift, boundary, dim) RESULT (arr_out)
	COMPLEX, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	COMPLEX, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	COMPLEX arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_C &
			(arr, dist, ierr)
		call do_shift_C  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_2D_C


!!!
!!!  End of 2D COMPLEX array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 3D COMPLEX array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 3D COMPLEX array.
!!
!!	public DP_CSHIFT_3D_C 
!!
FUNCTION DP_CSHIFT_3D_C (arr, shift, dim) &
                                                       RESULT (arr_out)
	COMPLEX, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	COMPLEX arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_C &
			(arr, dist, ierr)
		call do_shift_C  &
		   (offset, .TRUE., (0.0,0.0), arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_3D_C








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 3D COMPLEX array.
!!
!!	public DP_EOSHIFT_3D_C 
!!
FUNCTION DP_EOSHIFT_3D_C &
		(arr, shift, boundary, dim) RESULT (arr_out)
	COMPLEX, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	COMPLEX, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	COMPLEX arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_C &
			(arr, dist, ierr)
		call do_shift_C  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_3D_C


!!!
!!!  End of 3D COMPLEX array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 4D COMPLEX array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 4D COMPLEX array.
!!
!!	public DP_CSHIFT_4D_C 
!!
FUNCTION DP_CSHIFT_4D_C (arr, shift, dim) &
                                                       RESULT (arr_out)
	COMPLEX, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	COMPLEX arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_C &
			(arr, dist, ierr)
		call do_shift_C  &
		   (offset, .TRUE., (0.0,0.0), arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_4D_C








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 4D COMPLEX array.
!!
!!	public DP_EOSHIFT_4D_C 
!!
FUNCTION DP_EOSHIFT_4D_C &
		(arr, shift, boundary, dim) RESULT (arr_out)
	COMPLEX, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	COMPLEX, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	COMPLEX arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_C &
			(arr, dist, ierr)
		call do_shift_C  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_4D_C


!!!
!!!  End of 4D COMPLEX array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 1D INTEGER array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 1D INTEGER array.
!!
!!	public DP_CSHIFT_1D_I 
!!
FUNCTION DP_CSHIFT_1D_I (arr, shift, dim) &
                                                       RESULT (arr_out)
	INTEGER, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	INTEGER arr_out (0:size(arr,1)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_I &
			(arr, dist, ierr)
		call do_shift_I  &
		   (offset, .TRUE., 0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_1D_I








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 1D INTEGER array.
!!
!!	public DP_EOSHIFT_1D_I 
!!
FUNCTION DP_EOSHIFT_1D_I &
		(arr, shift, boundary, dim) RESULT (arr_out)
	INTEGER, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	INTEGER, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	INTEGER arr_out &
		(0:size(arr,1)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_I &
			(arr, dist, ierr)
		call do_shift_I  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_1D_I


!!!
!!!  End of 1D INTEGER array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 2D INTEGER array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 2D INTEGER array.
!!
!!	public DP_CSHIFT_2D_I 
!!
FUNCTION DP_CSHIFT_2D_I (arr, shift, dim) &
                                                       RESULT (arr_out)
	INTEGER, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	INTEGER arr_out (0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_I &
			(arr, dist, ierr)
		call do_shift_I  &
		   (offset, .TRUE., 0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_2D_I








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 2D INTEGER array.
!!
!!	public DP_EOSHIFT_2D_I 
!!
FUNCTION DP_EOSHIFT_2D_I &
		(arr, shift, boundary, dim) RESULT (arr_out)
	INTEGER, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	INTEGER, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	INTEGER arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_I &
			(arr, dist, ierr)
		call do_shift_I  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_2D_I


!!!
!!!  End of 2D INTEGER array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 3D INTEGER array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 3D INTEGER array.
!!
!!	public DP_CSHIFT_3D_I 
!!
FUNCTION DP_CSHIFT_3D_I (arr, shift, dim) &
                                                       RESULT (arr_out)
	INTEGER, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	INTEGER arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_I &
			(arr, dist, ierr)
		call do_shift_I  &
		   (offset, .TRUE., 0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_3D_I








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 3D INTEGER array.
!!
!!	public DP_EOSHIFT_3D_I 
!!
FUNCTION DP_EOSHIFT_3D_I &
		(arr, shift, boundary, dim) RESULT (arr_out)
	INTEGER, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	INTEGER, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	INTEGER arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_I &
			(arr, dist, ierr)
		call do_shift_I  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_3D_I


!!!
!!!  End of 3D INTEGER array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 4D INTEGER array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 4D INTEGER array.
!!
!!	public DP_CSHIFT_4D_I 
!!
FUNCTION DP_CSHIFT_4D_I (arr, shift, dim) &
                                                       RESULT (arr_out)
	INTEGER, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	INTEGER arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_I &
			(arr, dist, ierr)
		call do_shift_I  &
		   (offset, .TRUE., 0, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_4D_I








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 4D INTEGER array.
!!
!!	public DP_EOSHIFT_4D_I 
!!
FUNCTION DP_EOSHIFT_4D_I &
		(arr, shift, boundary, dim) RESULT (arr_out)
	INTEGER, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	INTEGER, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	INTEGER arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_I &
			(arr, dist, ierr)
		call do_shift_I  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_4D_I


!!!
!!!  End of 4D INTEGER array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 1D LOGICAL array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 1D LOGICAL array.
!!
!!	public DP_CSHIFT_1D_L 
!!
FUNCTION DP_CSHIFT_1D_L (arr, shift, dim) &
                                                       RESULT (arr_out)
	LOGICAL, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	LOGICAL arr_out (0:size(arr,1)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_L &
			(arr, dist, ierr)
		call do_shift_L  &
		   (offset, .TRUE., .TRUE., arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_1D_L








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 1D LOGICAL array.
!!
!!	public DP_EOSHIFT_1D_L 
!!
FUNCTION DP_EOSHIFT_1D_L &
		(arr, shift, boundary, dim) RESULT (arr_out)
	LOGICAL, intent(IN) :: arr (0:)
	integer, intent(IN) :: shift
	LOGICAL, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	LOGICAL arr_out &
		(0:size(arr,1)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 1 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_1D_L &
			(arr, dist, ierr)
		call do_shift_L  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_1D_L


!!!
!!!  End of 1D LOGICAL array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 2D LOGICAL array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 2D LOGICAL array.
!!
!!	public DP_CSHIFT_2D_L 
!!
FUNCTION DP_CSHIFT_2D_L (arr, shift, dim) &
                                                       RESULT (arr_out)
	LOGICAL, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	LOGICAL arr_out (0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_L &
			(arr, dist, ierr)
		call do_shift_L  &
		   (offset, .TRUE., .TRUE., arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_2D_L








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 2D LOGICAL array.
!!
!!	public DP_EOSHIFT_2D_L 
!!
FUNCTION DP_EOSHIFT_2D_L &
		(arr, shift, boundary, dim) RESULT (arr_out)
	LOGICAL, intent(IN) :: arr (0:,0:)
	integer, intent(IN) :: shift
	LOGICAL, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	LOGICAL arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 2 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_2D_L &
			(arr, dist, ierr)
		call do_shift_L  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_2D_L


!!!
!!!  End of 2D LOGICAL array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 3D LOGICAL array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 3D LOGICAL array.
!!
!!	public DP_CSHIFT_3D_L 
!!
FUNCTION DP_CSHIFT_3D_L (arr, shift, dim) &
                                                       RESULT (arr_out)
	LOGICAL, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	LOGICAL arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_L &
			(arr, dist, ierr)
		call do_shift_L  &
		   (offset, .TRUE., .TRUE., arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_3D_L








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 3D LOGICAL array.
!!
!!	public DP_EOSHIFT_3D_L 
!!
FUNCTION DP_EOSHIFT_3D_L &
		(arr, shift, boundary, dim) RESULT (arr_out)
	LOGICAL, intent(IN) :: arr (0:,0:,0:)
	integer, intent(IN) :: shift
	LOGICAL, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	LOGICAL arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 3 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_3D_L &
			(arr, dist, ierr)
		call do_shift_L  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_3D_L


!!!
!!!  End of 3D LOGICAL array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of 4D LOGICAL array shift routines
!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do a circular shift on a 4D LOGICAL array.
!!
!!	public DP_CSHIFT_4D_L 
!!
FUNCTION DP_CSHIFT_4D_L (arr, shift, dim) &
                                                       RESULT (arr_out)
	LOGICAL, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	LOGICAL arr_out (0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_cshift_ierr = 0

	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_L &
			(arr, dist, ierr)
		call do_shift_L  &
		   (offset, .TRUE., .TRUE., arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_CSHIFT_4D_L








!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do an end-off shift on a 4D LOGICAL array.
!!
!!	public DP_EOSHIFT_4D_L 
!!
FUNCTION DP_EOSHIFT_4D_L &
		(arr, shift, boundary, dim) RESULT (arr_out)
	LOGICAL, intent(IN) :: arr (0:,0:,0:,0:)
	integer, intent(IN) :: shift
	LOGICAL, intent(IN) :: boundary
	integer, intent(IN) :: dim

	integer arr_size(DP_MAX_DIMS)
	integer offset(DP_MAX_DIMS)
	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)
	integer ii
	integer ierr

	type (array_distribution), pointer :: dist

	LOGICAL arr_out &
		(0:size(arr,1)-1, 0:size(arr,2)-1, 0:size(arr,3)-1, 0:size(arr,4)-1)

	dp_eoshift_ierr = 0


	if (shift .eq. 0) then

		arr_out = arr

	else

		offset = 0
		offset(dim) = shift
		arr_size = 1
		do ii = 1, 4 
			arr_size(ii) = size(arr,ii)
		enddo

		call dp_get_arr_dist_4D_L &
			(arr, dist, ierr)
		call do_shift_L  &
	    	 (offset, .FALSE., boundary, arr_size, dist, arr, arr_out, ierr)

	endif  ! end of clause for non-zero shift


end function DP_EOSHIFT_4D_L


!!!
!!!  End of 4D LOGICAL array shift routines
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


