
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of REAL array shift routines for arrays of max dimension
!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do the indicated regular block moves on an REAL array.  Outgoing data
!! is copies from one array.  Incoming data is copied to another array.
!!
!! NOTE: all arrays are treated as 4D.  If it is not 4D in the calling routine,
!!       the higher dimensions should be indicated to have size 1 in arr_size.
!!
SUBROUTINE do_reg_blk_moves_R &
	(arr_in, arr_size, arr_out, blk_mv, num_moves, do_wraps, fill, ierr)

	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	REAL, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	REAL, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the block moves
	type (regular_blk_move_type), intent(IN) :: blk_mv (0:num_moves-1)

	!! the number of block moves
	integer, intent(IN) :: num_moves

	!! indicates whether data that wraps around to the other side of the
	!! global array should be moved.
	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	REAL, intent(IN) :: fill

	!! error code
	integer, intent(OUT) :: ierr


	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	integer ii


	integer i
	integer total_elements

	integer*4 status(MPI_STATUS_SIZE)

   do i = 0, num_moves-1





	if (blk_mv(i)%targ_rank .eq. dp_rank) then
	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
		arr_in (                                                &
			blk_mv(i)%src_coords(1):			&
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	&
		        blk_mv(i)%src_coords(2):			&
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,  &
		        blk_mv(i)%src_coords(3):			&
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,  &
		        blk_mv(i)%src_coords(4):			&
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1)

	    else

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
	    	fill
	    endif

	else

	  total_elements = 1
	  do ii = 1, DP_MAX_DIMS
	    total_elements = total_elements * blk_mv(i)%size(ii)
	  enddo

	  if (total_elements .ne. 0) then  ! there is data to move


	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		call MPI_SENDRECV (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_REAL,               			    &
			blk_mv(i)%targ_rank, 				    &
			0,  			    			    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_REAL,               			    &
			blk_mv(i)%src_rank,       			    &
			MPI_ANY_TAG, 				    	    &
			dp_comm_cart, 				    	    &
			status, 				    	    &
			ierr)

	    else   ! .not. do_wraps and one of the directions wraps

		! Note that this section could be done with MPI_SENDRECV
		! and appropriate use of MPI_PROC_NULL.  I don't do that
		! because the recieve and send arrays will be packed to and
		! and unpacked from contiguous memory as part of constructing
		! the subroutine call.  This inefficiency is avoided in the
		! code below.
		!
		if (blk_mv(i)%incoming_wraps) then
	   	    arr_out(blk_mv(i)%targ_coords(1):			  &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,   &
		        blk_mv(i)%targ_coords(2):			  &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,   &
		        blk_mv(i)%targ_coords(3):			  &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,   &
		        blk_mv(i)%targ_coords(4):			  &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1) = &
		    fill
		else
		    call MPI_RECV (					    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_REAL,			            	    &
			blk_mv(i)%src_rank, 				    &
			MPI_ANY_TAG, 					    &
			dp_comm_cart, 					    &
			status, 					    &
			ierr)
		endif

		if (.not. blk_mv(i)%outgoing_wraps) then
		    call MPI_SEND (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_REAL,  	        			    &
			blk_mv(i)%targ_rank, 				    &
			0, 			 			    &
			dp_comm_cart, 					    &
			ierr)
		endif

	    endif  ! done with data transfer


	  endif ! there was some data to transfer

	endif

    enddo
		
    ! set ierr

end subroutine do_reg_blk_moves_R




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Why not have a dp_shift_circ and dp_shift_eo that does a shift
!! along multiple dimensions?  Don't make it function.  Maybe:
!!
!! subroutine dp_shift_circ (arr_in, shifts, arr_out, ierr)
!!   where shifts is a 1D array of length equal to the number of
!!   dimensions of arr_in.  arr_out is same type and conformable to arr_in
!!
!! For _eo version just add a fill value.  Need only be a scalar.
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


subroutine do_irreg_shift_R &
	(offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	REAL, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	REAL, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	REAL, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	ierr = 0



	return
end subroutine do_irreg_shift_R




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_shift_R &
   (offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	REAL, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	REAL, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	REAL, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	!!!!!!!!!!!!

	logical regular
	

	if ( associated (dist) ) then
		regular = dist%all_even
	else
		regular = .TRUE.
	endif


	if (regular) then

		!! regularly partitioned arrays
		call do_reg_shift_R  &
		    (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	else

		!! possibly irregularly partitioned arrays
		call do_irreg_shift_R &
	  		(offset, do_wraps, fill, arr_size, dist, &
				arr_in, arr_out, ierr)

	endif


	return

end subroutine do_shift_R



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_reg_shift_R &
   (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	REAL, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	REAL, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	REAL, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr


	!!!!!!!!!!!

	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)


	call get_reg_blk_moves (offset, arr_size, blk_mv, ierr)
	if (ierr .ne. 0) then
		return
	endif

	call do_reg_blk_moves_R 	&
		(arr_in, arr_size, 		&
		 arr_out, 			&
		 blk_mv, DP_MAX_MV_BLOCKS,	&
		 do_wraps, fill,		&
		 ierr)


	return

end subroutine do_reg_shift_R





!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of DOUBLE PRECISION array shift routines for arrays of max dimension
!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do the indicated regular block moves on an DOUBLE PRECISION array.  Outgoing data
!! is copies from one array.  Incoming data is copied to another array.
!!
!! NOTE: all arrays are treated as 4D.  If it is not 4D in the calling routine,
!!       the higher dimensions should be indicated to have size 1 in arr_size.
!!
SUBROUTINE do_reg_blk_moves_D &
	(arr_in, arr_size, arr_out, blk_mv, num_moves, do_wraps, fill, ierr)

	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	DOUBLE PRECISION, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	DOUBLE PRECISION, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the block moves
	type (regular_blk_move_type), intent(IN) :: blk_mv (0:num_moves-1)

	!! the number of block moves
	integer, intent(IN) :: num_moves

	!! indicates whether data that wraps around to the other side of the
	!! global array should be moved.
	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	DOUBLE PRECISION, intent(IN) :: fill

	!! error code
	integer, intent(OUT) :: ierr


	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	integer ii


	integer i
	integer total_elements

	integer*4 status(MPI_STATUS_SIZE)

   do i = 0, num_moves-1





	if (blk_mv(i)%targ_rank .eq. dp_rank) then
	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
		arr_in (                                                &
			blk_mv(i)%src_coords(1):			&
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	&
		        blk_mv(i)%src_coords(2):			&
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,  &
		        blk_mv(i)%src_coords(3):			&
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,  &
		        blk_mv(i)%src_coords(4):			&
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1)

	    else

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
	    	fill
	    endif

	else

	  total_elements = 1
	  do ii = 1, DP_MAX_DIMS
	    total_elements = total_elements * blk_mv(i)%size(ii)
	  enddo

	  if (total_elements .ne. 0) then  ! there is data to move


	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		call MPI_SENDRECV (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_DOUBLE_PRECISION,               			    &
			blk_mv(i)%targ_rank, 				    &
			0,  			    			    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_DOUBLE_PRECISION,               			    &
			blk_mv(i)%src_rank,       			    &
			MPI_ANY_TAG, 				    	    &
			dp_comm_cart, 				    	    &
			status, 				    	    &
			ierr)

	    else   ! .not. do_wraps and one of the directions wraps

		! Note that this section could be done with MPI_SENDRECV
		! and appropriate use of MPI_PROC_NULL.  I don't do that
		! because the recieve and send arrays will be packed to and
		! and unpacked from contiguous memory as part of constructing
		! the subroutine call.  This inefficiency is avoided in the
		! code below.
		!
		if (blk_mv(i)%incoming_wraps) then
	   	    arr_out(blk_mv(i)%targ_coords(1):			  &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,   &
		        blk_mv(i)%targ_coords(2):			  &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,   &
		        blk_mv(i)%targ_coords(3):			  &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,   &
		        blk_mv(i)%targ_coords(4):			  &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1) = &
		    fill
		else
		    call MPI_RECV (					    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_DOUBLE_PRECISION,			            	    &
			blk_mv(i)%src_rank, 				    &
			MPI_ANY_TAG, 					    &
			dp_comm_cart, 					    &
			status, 					    &
			ierr)
		endif

		if (.not. blk_mv(i)%outgoing_wraps) then
		    call MPI_SEND (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_DOUBLE_PRECISION,  	        			    &
			blk_mv(i)%targ_rank, 				    &
			0, 			 			    &
			dp_comm_cart, 					    &
			ierr)
		endif

	    endif  ! done with data transfer


	  endif ! there was some data to transfer

	endif

    enddo
		
    ! set ierr

end subroutine do_reg_blk_moves_D




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Why not have a dp_shift_circ and dp_shift_eo that does a shift
!! along multiple dimensions?  Don't make it function.  Maybe:
!!
!! subroutine dp_shift_circ (arr_in, shifts, arr_out, ierr)
!!   where shifts is a 1D array of length equal to the number of
!!   dimensions of arr_in.  arr_out is same type and conformable to arr_in
!!
!! For _eo version just add a fill value.  Need only be a scalar.
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


subroutine do_irreg_shift_D &
	(offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	DOUBLE PRECISION, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	DOUBLE PRECISION, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	DOUBLE PRECISION, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	ierr = 0



	return
end subroutine do_irreg_shift_D




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_shift_D &
   (offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	DOUBLE PRECISION, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	DOUBLE PRECISION, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	DOUBLE PRECISION, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	!!!!!!!!!!!!

	logical regular
	

	if ( associated (dist) ) then
		regular = dist%all_even
	else
		regular = .TRUE.
	endif


	if (regular) then

		!! regularly partitioned arrays
		call do_reg_shift_D  &
		    (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	else

		!! possibly irregularly partitioned arrays
		call do_irreg_shift_D &
	  		(offset, do_wraps, fill, arr_size, dist, &
				arr_in, arr_out, ierr)

	endif


	return

end subroutine do_shift_D



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_reg_shift_D &
   (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	DOUBLE PRECISION, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	DOUBLE PRECISION, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	DOUBLE PRECISION, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr


	!!!!!!!!!!!

	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)


	call get_reg_blk_moves (offset, arr_size, blk_mv, ierr)
	if (ierr .ne. 0) then
		return
	endif

	call do_reg_blk_moves_D 	&
		(arr_in, arr_size, 		&
		 arr_out, 			&
		 blk_mv, DP_MAX_MV_BLOCKS,	&
		 do_wraps, fill,		&
		 ierr)


	return

end subroutine do_reg_shift_D





!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of COMPLEX array shift routines for arrays of max dimension
!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do the indicated regular block moves on an COMPLEX array.  Outgoing data
!! is copies from one array.  Incoming data is copied to another array.
!!
!! NOTE: all arrays are treated as 4D.  If it is not 4D in the calling routine,
!!       the higher dimensions should be indicated to have size 1 in arr_size.
!!
SUBROUTINE do_reg_blk_moves_C &
	(arr_in, arr_size, arr_out, blk_mv, num_moves, do_wraps, fill, ierr)

	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	COMPLEX, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	COMPLEX, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the block moves
	type (regular_blk_move_type), intent(IN) :: blk_mv (0:num_moves-1)

	!! the number of block moves
	integer, intent(IN) :: num_moves

	!! indicates whether data that wraps around to the other side of the
	!! global array should be moved.
	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	COMPLEX, intent(IN) :: fill

	!! error code
	integer, intent(OUT) :: ierr


	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	integer ii


	integer i
	integer total_elements

	integer*4 status(MPI_STATUS_SIZE)

   do i = 0, num_moves-1





	if (blk_mv(i)%targ_rank .eq. dp_rank) then
	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
		arr_in (                                                &
			blk_mv(i)%src_coords(1):			&
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	&
		        blk_mv(i)%src_coords(2):			&
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,  &
		        blk_mv(i)%src_coords(3):			&
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,  &
		        blk_mv(i)%src_coords(4):			&
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1)

	    else

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
	    	fill
	    endif

	else

	  total_elements = 1
	  do ii = 1, DP_MAX_DIMS
	    total_elements = total_elements * blk_mv(i)%size(ii)
	  enddo

	  if (total_elements .ne. 0) then  ! there is data to move


	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		call MPI_SENDRECV (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_COMPLEX,               			    &
			blk_mv(i)%targ_rank, 				    &
			0,  			    			    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_COMPLEX,               			    &
			blk_mv(i)%src_rank,       			    &
			MPI_ANY_TAG, 				    	    &
			dp_comm_cart, 				    	    &
			status, 				    	    &
			ierr)

	    else   ! .not. do_wraps and one of the directions wraps

		! Note that this section could be done with MPI_SENDRECV
		! and appropriate use of MPI_PROC_NULL.  I don't do that
		! because the recieve and send arrays will be packed to and
		! and unpacked from contiguous memory as part of constructing
		! the subroutine call.  This inefficiency is avoided in the
		! code below.
		!
		if (blk_mv(i)%incoming_wraps) then
	   	    arr_out(blk_mv(i)%targ_coords(1):			  &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,   &
		        blk_mv(i)%targ_coords(2):			  &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,   &
		        blk_mv(i)%targ_coords(3):			  &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,   &
		        blk_mv(i)%targ_coords(4):			  &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1) = &
		    fill
		else
		    call MPI_RECV (					    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_COMPLEX,			            	    &
			blk_mv(i)%src_rank, 				    &
			MPI_ANY_TAG, 					    &
			dp_comm_cart, 					    &
			status, 					    &
			ierr)
		endif

		if (.not. blk_mv(i)%outgoing_wraps) then
		    call MPI_SEND (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_COMPLEX,  	        			    &
			blk_mv(i)%targ_rank, 				    &
			0, 			 			    &
			dp_comm_cart, 					    &
			ierr)
		endif

	    endif  ! done with data transfer


	  endif ! there was some data to transfer

	endif

    enddo
		
    ! set ierr

end subroutine do_reg_blk_moves_C




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Why not have a dp_shift_circ and dp_shift_eo that does a shift
!! along multiple dimensions?  Don't make it function.  Maybe:
!!
!! subroutine dp_shift_circ (arr_in, shifts, arr_out, ierr)
!!   where shifts is a 1D array of length equal to the number of
!!   dimensions of arr_in.  arr_out is same type and conformable to arr_in
!!
!! For _eo version just add a fill value.  Need only be a scalar.
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


subroutine do_irreg_shift_C &
	(offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	COMPLEX, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	COMPLEX, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	COMPLEX, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	ierr = 0



	return
end subroutine do_irreg_shift_C




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_shift_C &
   (offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	COMPLEX, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	COMPLEX, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	COMPLEX, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	!!!!!!!!!!!!

	logical regular
	

	if ( associated (dist) ) then
		regular = dist%all_even
	else
		regular = .TRUE.
	endif


	if (regular) then

		!! regularly partitioned arrays
		call do_reg_shift_C  &
		    (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	else

		!! possibly irregularly partitioned arrays
		call do_irreg_shift_C &
	  		(offset, do_wraps, fill, arr_size, dist, &
				arr_in, arr_out, ierr)

	endif


	return

end subroutine do_shift_C



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_reg_shift_C &
   (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	COMPLEX, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	COMPLEX, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	COMPLEX, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr


	!!!!!!!!!!!

	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)


	call get_reg_blk_moves (offset, arr_size, blk_mv, ierr)
	if (ierr .ne. 0) then
		return
	endif

	call do_reg_blk_moves_C 	&
		(arr_in, arr_size, 		&
		 arr_out, 			&
		 blk_mv, DP_MAX_MV_BLOCKS,	&
		 do_wraps, fill,		&
		 ierr)


	return

end subroutine do_reg_shift_C





!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of INTEGER array shift routines for arrays of max dimension
!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do the indicated regular block moves on an INTEGER array.  Outgoing data
!! is copies from one array.  Incoming data is copied to another array.
!!
!! NOTE: all arrays are treated as 4D.  If it is not 4D in the calling routine,
!!       the higher dimensions should be indicated to have size 1 in arr_size.
!!
SUBROUTINE do_reg_blk_moves_I &
	(arr_in, arr_size, arr_out, blk_mv, num_moves, do_wraps, fill, ierr)

	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	INTEGER, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	INTEGER, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the block moves
	type (regular_blk_move_type), intent(IN) :: blk_mv (0:num_moves-1)

	!! the number of block moves
	integer, intent(IN) :: num_moves

	!! indicates whether data that wraps around to the other side of the
	!! global array should be moved.
	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	INTEGER, intent(IN) :: fill

	!! error code
	integer, intent(OUT) :: ierr


	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	integer ii


	integer i
	integer total_elements

	integer*4 status(MPI_STATUS_SIZE)

   do i = 0, num_moves-1





	if (blk_mv(i)%targ_rank .eq. dp_rank) then
	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
		arr_in (                                                &
			blk_mv(i)%src_coords(1):			&
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	&
		        blk_mv(i)%src_coords(2):			&
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,  &
		        blk_mv(i)%src_coords(3):			&
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,  &
		        blk_mv(i)%src_coords(4):			&
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1)

	    else

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
	    	fill
	    endif

	else

	  total_elements = 1
	  do ii = 1, DP_MAX_DIMS
	    total_elements = total_elements * blk_mv(i)%size(ii)
	  enddo

	  if (total_elements .ne. 0) then  ! there is data to move


	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		call MPI_SENDRECV (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_INTEGER,               			    &
			blk_mv(i)%targ_rank, 				    &
			0,  			    			    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_INTEGER,               			    &
			blk_mv(i)%src_rank,       			    &
			MPI_ANY_TAG, 				    	    &
			dp_comm_cart, 				    	    &
			status, 				    	    &
			ierr)

	    else   ! .not. do_wraps and one of the directions wraps

		! Note that this section could be done with MPI_SENDRECV
		! and appropriate use of MPI_PROC_NULL.  I don't do that
		! because the recieve and send arrays will be packed to and
		! and unpacked from contiguous memory as part of constructing
		! the subroutine call.  This inefficiency is avoided in the
		! code below.
		!
		if (blk_mv(i)%incoming_wraps) then
	   	    arr_out(blk_mv(i)%targ_coords(1):			  &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,   &
		        blk_mv(i)%targ_coords(2):			  &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,   &
		        blk_mv(i)%targ_coords(3):			  &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,   &
		        blk_mv(i)%targ_coords(4):			  &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1) = &
		    fill
		else
		    call MPI_RECV (					    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_INTEGER,			            	    &
			blk_mv(i)%src_rank, 				    &
			MPI_ANY_TAG, 					    &
			dp_comm_cart, 					    &
			status, 					    &
			ierr)
		endif

		if (.not. blk_mv(i)%outgoing_wraps) then
		    call MPI_SEND (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_INTEGER,  	        			    &
			blk_mv(i)%targ_rank, 				    &
			0, 			 			    &
			dp_comm_cart, 					    &
			ierr)
		endif

	    endif  ! done with data transfer


	  endif ! there was some data to transfer

	endif

    enddo
		
    ! set ierr

end subroutine do_reg_blk_moves_I




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Why not have a dp_shift_circ and dp_shift_eo that does a shift
!! along multiple dimensions?  Don't make it function.  Maybe:
!!
!! subroutine dp_shift_circ (arr_in, shifts, arr_out, ierr)
!!   where shifts is a 1D array of length equal to the number of
!!   dimensions of arr_in.  arr_out is same type and conformable to arr_in
!!
!! For _eo version just add a fill value.  Need only be a scalar.
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


subroutine do_irreg_shift_I &
	(offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	INTEGER, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	INTEGER, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	INTEGER, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	ierr = 0



	return
end subroutine do_irreg_shift_I




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_shift_I &
   (offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	INTEGER, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	INTEGER, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	INTEGER, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	!!!!!!!!!!!!

	logical regular
	

	if ( associated (dist) ) then
		regular = dist%all_even
	else
		regular = .TRUE.
	endif


	if (regular) then

		!! regularly partitioned arrays
		call do_reg_shift_I  &
		    (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	else

		!! possibly irregularly partitioned arrays
		call do_irreg_shift_I &
	  		(offset, do_wraps, fill, arr_size, dist, &
				arr_in, arr_out, ierr)

	endif


	return

end subroutine do_shift_I



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_reg_shift_I &
   (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	INTEGER, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	INTEGER, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	INTEGER, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr


	!!!!!!!!!!!

	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)


	call get_reg_blk_moves (offset, arr_size, blk_mv, ierr)
	if (ierr .ne. 0) then
		return
	endif

	call do_reg_blk_moves_I 	&
		(arr_in, arr_size, 		&
		 arr_out, 			&
		 blk_mv, DP_MAX_MV_BLOCKS,	&
		 do_wraps, fill,		&
		 ierr)


	return

end subroutine do_reg_shift_I





!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!!  Begining of LOGICAL array shift routines for arrays of max dimension
!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Do the indicated regular block moves on an LOGICAL array.  Outgoing data
!! is copies from one array.  Incoming data is copied to another array.
!!
!! NOTE: all arrays are treated as 4D.  If it is not 4D in the calling routine,
!!       the higher dimensions should be indicated to have size 1 in arr_size.
!!
SUBROUTINE do_reg_blk_moves_L &
	(arr_in, arr_size, arr_out, blk_mv, num_moves, do_wraps, fill, ierr)

	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	LOGICAL, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	LOGICAL, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the block moves
	type (regular_blk_move_type), intent(IN) :: blk_mv (0:num_moves-1)

	!! the number of block moves
	integer, intent(IN) :: num_moves

	!! indicates whether data that wraps around to the other side of the
	!! global array should be moved.
	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	LOGICAL, intent(IN) :: fill

	!! error code
	integer, intent(OUT) :: ierr


	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	integer ii


	integer i
	integer total_elements

	integer*4 status(MPI_STATUS_SIZE)

   do i = 0, num_moves-1





	if (blk_mv(i)%targ_rank .eq. dp_rank) then
	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
		arr_in (                                                &
			blk_mv(i)%src_coords(1):			&
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	&
		        blk_mv(i)%src_coords(2):			&
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,  &
		        blk_mv(i)%src_coords(3):			&
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,  &
		        blk_mv(i)%src_coords(4):			&
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1)

	    else

		arr_out(                                                &
			blk_mv(i)%targ_coords(1):			&
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1, &
		        blk_mv(i)%targ_coords(2):			&
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1, &
		        blk_mv(i)%targ_coords(3):			&
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1, &
		        blk_mv(i)%targ_coords(4):			&
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1  &
		       ) = &
	    	fill
	    endif

	else

	  total_elements = 1
	  do ii = 1, DP_MAX_DIMS
	    total_elements = total_elements * blk_mv(i)%size(ii)
	  enddo

	  if (total_elements .ne. 0) then  ! there is data to move


	    if (do_wraps .or. ( (.not. blk_mv(i)%incoming_wraps) .and.  &
				(.not. blk_mv(i)%outgoing_wraps) ) ) then

		call MPI_SENDRECV (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_LOGICAL,               			    &
			blk_mv(i)%targ_rank, 				    &
			0,  			    			    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_LOGICAL,               			    &
			blk_mv(i)%src_rank,       			    &
			MPI_ANY_TAG, 				    	    &
			dp_comm_cart, 				    	    &
			status, 				    	    &
			ierr)

	    else   ! .not. do_wraps and one of the directions wraps

		! Note that this section could be done with MPI_SENDRECV
		! and appropriate use of MPI_PROC_NULL.  I don't do that
		! because the recieve and send arrays will be packed to and
		! and unpacked from contiguous memory as part of constructing
		! the subroutine call.  This inefficiency is avoided in the
		! code below.
		!
		if (blk_mv(i)%incoming_wraps) then
	   	    arr_out(blk_mv(i)%targ_coords(1):			  &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,   &
		        blk_mv(i)%targ_coords(2):			  &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,   &
		        blk_mv(i)%targ_coords(3):			  &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,   &
		        blk_mv(i)%targ_coords(4):			  &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1) = &
		    fill
		else
		    call MPI_RECV (					    &
			arr_out(blk_mv(i)%targ_coords(1):		    &
			  blk_mv(i)%targ_coords(1)+blk_mv(i)%size(1)-1,     &
		        blk_mv(i)%targ_coords(2):			    &
			  blk_mv(i)%targ_coords(2)+blk_mv(i)%size(2)-1,     &
		        blk_mv(i)%targ_coords(3):			    &
			  blk_mv(i)%targ_coords(3)+blk_mv(i)%size(3)-1,     &
		        blk_mv(i)%targ_coords(4):			    &
			  blk_mv(i)%targ_coords(4)+blk_mv(i)%size(4)-1),    &
			total_elements, 				    &
			MPI_LOGICAL,			            	    &
			blk_mv(i)%src_rank, 				    &
			MPI_ANY_TAG, 					    &
			dp_comm_cart, 					    &
			status, 					    &
			ierr)
		endif

		if (.not. blk_mv(i)%outgoing_wraps) then
		    call MPI_SEND (					    &
			arr_in (blk_mv(i)%src_coords(1):		    &
			  blk_mv(i)%src_coords(1)+blk_mv(i)%size(1)-1,	    &
		        blk_mv(i)%src_coords(2):			    &
			  blk_mv(i)%src_coords(2)+blk_mv(i)%size(2)-1,      &
		        blk_mv(i)%src_coords(3):			    &
			  blk_mv(i)%src_coords(3)+blk_mv(i)%size(3)-1,      &
		        blk_mv(i)%src_coords(4):			    &
			  blk_mv(i)%src_coords(4)+blk_mv(i)%size(4)-1),     &
			total_elements,  				    &
			MPI_LOGICAL,  	        			    &
			blk_mv(i)%targ_rank, 				    &
			0, 			 			    &
			dp_comm_cart, 					    &
			ierr)
		endif

	    endif  ! done with data transfer


	  endif ! there was some data to transfer

	endif

    enddo
		
    ! set ierr

end subroutine do_reg_blk_moves_L




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Why not have a dp_shift_circ and dp_shift_eo that does a shift
!! along multiple dimensions?  Don't make it function.  Maybe:
!!
!! subroutine dp_shift_circ (arr_in, shifts, arr_out, ierr)
!!   where shifts is a 1D array of length equal to the number of
!!   dimensions of arr_in.  arr_out is same type and conformable to arr_in
!!
!! For _eo version just add a fill value.  Need only be a scalar.
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


subroutine do_irreg_shift_L &
	(offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	LOGICAL, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	LOGICAL, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	LOGICAL, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	ierr = 0



	return
end subroutine do_irreg_shift_L




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_shift_L &
   (offset, do_wraps, fill, arr_size, dist, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	LOGICAL, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the distribution of the global data array over the processor array
	type (array_distribution), pointer :: dist

	!! the incoming array
	LOGICAL, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	LOGICAL, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr

	!!!!!!!!!!!!

	logical regular
	

	if ( associated (dist) ) then
		regular = dist%all_even
	else
		regular = .TRUE.
	endif


	if (regular) then

		!! regularly partitioned arrays
		call do_reg_shift_L  &
		    (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	else

		!! possibly irregularly partitioned arrays
		call do_irreg_shift_L &
	  		(offset, do_wraps, fill, arr_size, dist, &
				arr_in, arr_out, ierr)

	endif


	return

end subroutine do_shift_L



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine does a shift on arrays that are regularly partitioned
!! among the processes.  It handles both circular and end-off shifts.
!!
subroutine do_reg_shift_L &
   (offset, do_wraps, fill, arr_size, arr_in, arr_out, ierr)

	integer, intent(IN) :: offset(DP_MAX_DIMS)

	logical, intent(IN) :: do_wraps

	!! if wraps are not done, then the target space is filled with 
	!! this value.
	LOGICAL, intent(IN) :: fill


	!! the size of the incoming and outgoing arrays.
	integer, intent(IN) :: arr_size(DP_MAX_DIMS)

	!! the incoming array
	LOGICAL, intent(IN) :: arr_in  &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	!! the outgoing array
	LOGICAL, intent(OUT) :: arr_out &
	   (0:arr_size(1)-1, 0:arr_size(2)-1, 0:arr_size(3)-1, 0:arr_size(4)-1)

	integer, intent(OUT) :: ierr


	!!!!!!!!!!!

	type (regular_blk_move_type) :: blk_mv (0:DP_MAX_MV_BLOCKS-1)


	call get_reg_blk_moves (offset, arr_size, blk_mv, ierr)
	if (ierr .ne. 0) then
		return
	endif

	call do_reg_blk_moves_L 	&
		(arr_in, arr_size, 		&
		 arr_out, 			&
		 blk_mv, DP_MAX_MV_BLOCKS,	&
		 do_wraps, fill,		&
		 ierr)


	return

end subroutine do_reg_shift_L




