

	module dp_uni

	use DP_INTERNAL

	implicit none



!       The following statement indicates that all variables and routines
!       are private (inaccessible outside of the module) by default.
!       Items can be made accessible only by explicitly declaring them public.
	private



!	These are the only public routines
        public dp_uni_init, dp_uni



CONTAINS

! RANDOM NUMBER GENERATOR

	function dp_uni_init (seedvec)

!IN:
	integer seedvec(0:)  ! a vector of integer seeds; processor picks seed at dp_rand place in it.
!OUT:
	real dp_uni_init     ! floating point value of seed used

        IMPLICIT NONE
        integer, parameter :: k = 24
        integer ii,jj
        integer i1,j1,k1,l1,m1
        integer seed_par
        real s,t
      

        seed_par = seedvec(dp_rank)
        dp_uni_c = dp_uni_csave
        dp_uni_i = 17
        dp_uni_j = 5




 
!          Set up ...
!          Convert seed_par to four smallish positive integers.

        i1 = MOD(ABS(seed_par),177)+1
        j1 = MOD(ABS(seed_par),167)+1
        k1 = MOD(ABS(seed_par),157)+1
        l1 = MOD(ABS(seed_par),147)+1

!              Generate random bit pattern in array based on given seed.
 
        DO 2 ii = 1,17
          s = 0.0
          t = 0.5
!             Do for each of the bits of mantissa of word 
!             Loop  over K bits, where K is defaulted to 24 
          DO 3 jj = 1,k
                  m1 = MOD(MOD(i1*j1,179)*k1,179)
                  i1 = j1
                  j1 = k1
                  k1 = m1
                  l1 = MOD(53*l1+1,169)
                  if(MOD(l1*m1,64) .ge. 32)s = s + t
    3             t = .5*t
    2   dp_u_array(ii) = s

	dp_uni_init = float(seed_par)

        RETURN 
	end function dp_uni_init



	function dp_uni()

	real dp_uni(0:dp_block_sizeM(0),0:dp_block_sizeM(1))

        implicit none


!***PURPOSE  THIS ROUTINE GENERATES PARALLEL (2-DIMENSION) REAL  
!            (SINGLE PRECISION) UNIFORM RANDOM NUMBERS ON [0,1)

!***DESCRIPTION
!        Computes real (single precision) uniform numbers on [0,1).
!           From the book, "Numerical Methods and Software" by
!                D. Kahaner, C. Moler, S. Nash
!                Prentice Hall, 1988
! 
!       USAGE: 
!              To initialize the generator 
!            fseed = dp_uni_init (seedvec)

!                  seedvec             is a vector of seeds; vector must be at least as long as the number of processors
!                  will return floating point value of ISEED. 
! 
!               Subsequently 
!                       array = dp_uni() 
!                  will return a real uniform array on [0,1) 
 
!                One initialization is necessary, but any number of initializations with 
!                   dp_uni_init() are allowed.
 
!***END PROLOGUE 


      
      integer index1, index2
    
!   Basic generator is Fibonacci

      do index2 = 0, dp_block_sizeM(1)
        do index1 = 0, dp_block_sizeM(0)
 
          dp_uni(index1,index2)  = dp_u_array(dp_uni_i) - dp_u_array(dp_uni_j)
          if(dp_uni(index1,index2) .lt. 0.0)dp_uni(index1,index2) = dp_uni(index1,index2) + 1.0
          dp_u_array(dp_uni_i) = dp_uni(index1,index2)
          dp_uni_i = dp_uni_i - 1
          if(dp_uni_i .eq. 0)dp_uni_i = 17
          dp_uni_j = dp_uni_j - 1
          if(dp_uni_j .eq. 0)dp_uni_j = 17

!   Second generator is congruential

          dp_uni_c = dp_uni_c - dp_uni_cd 
          if(dp_uni_c .lt. 0.0) dp_uni_c = dp_uni_c + dp_uni_cm 

!   Combination generator
 
          dp_uni(index1,index2) = dp_uni(index1,index2) - dp_uni_c
          if(dp_uni(index1,index2) .lt. 0.0)dp_uni(index1,index2) = dp_uni(index1,index2) + 1.0

        enddo
      enddo

        RETURN
	end function dp_uni





	end module dp_uni
