
	module dp_util



	use DP_INTERNAL

	use DP_RANDOM

	use DP_ALLOC


	implicit none

!       The following statement indicates that all variables and routines
!       are private (inaccessible outside of the module) by default.
!       Items can be made accessible only by explicitly declaring them public.
	private



!	These are the public routines
	public 	dp_initialize, 		&
		dp_finalize, 		&
		dp_query_layout, 	&
		dp_sync, 		&
		dp_get_num_procs, 	&
		dp_setup_proc_array, 	&
		dp_init


CONTAINS





SUBROUTINE INIT_UNI ()
        integer tmp_seeds(0:dp_maxproc)
        real fake_seed

	! initialize random number genearator in case user doesn't
        dp_seeds (0:dp_maxproc) =  &
		(/ 305, 306, 307, 308, 309, 310, 311, 312, &
		   313, 314, 315, 316, 317, 318, 319, 320, &
		   321, 322, 323, 324, 325, 326, 327, 328, &
		   329, 330, 331, 332, 333, 334, 335, 336, &
		   337, 338, 339, 340, 341, 342, 343, 344, &
		   345, 346, 347, 348, 349, 350, 351, 352, &
		   353, 354, 355, 356, 357, 358, 359, 360, &
		   361, 362, 363, 364, 365, 366, 367, 368, 369 /)

        dp_u_array (1:17)  =  &
		(/ 0.8668672834288,  0.3697986366357,  0.8008968294805, &
		   0.4173889774680,  0.8254561579836,  0.9640965269077, &
		   0.4508667414265,  0.6451309529668,  0.1645456024730, &
		   0.2787901807898,  0.06761531340295, 0.9663226330820, &
		   0.01963343943798, 0.02947398211399, 0.1636231515294, &
		   0.3976343250467,  0.2631008574685 /)

        tmp_seeds = dp_seeds
        dp_uni_c = dp_uni_csave
        dp_uni_i = 17
        dp_uni_j = 5

        fake_seed = dp_uni_init(tmp_seeds)   
	! end random number generator initialization

	return

END SUBROUTINE INIT_UNI


SUBROUTINE DP_INIT (num_procs, this_proc, ierror)
integer, intent (OUT) :: num_procs
integer, intent (OUT) :: this_proc
integer, intent (OUT) :: ierror

	if (dp_initialized) then

		if (dp_comm_cart .ne. MPI_COMM_NULL) then

			call MPI_COMM_FREE (dp_comm_cart)
			dp_comm_cart = MPI_COMM_NULL

		else  ! this happens when previous initialization
		      ! specified only one processor

			if ( (.not. dp_mpi_initialized ) .and.    &
			     (.not. dp_serial_run)) then
				call MPI_INIT(ierror)
				! check ierror
				dp_mpi_initialized = .TRUE.
			endif

		endif

	else
		call init_uni ()   ! initialize random number generator

		call dp_initialize_array_lists () ! Initializes the linked lists
					  	  ! of array descriptors.
		if ((.not. dp_mpi_initialized) .and. (.not. dp_serial_run)) then
			call MPI_INIT(ierror)
			! check ierror
			dp_mpi_initialized = .TRUE.
		endif
	endif

	if (dp_serial_run) then
		dp_num_procs = 1
		num_procs = 1
		this_proc = 0
		dp_rank = 0
		ierror = DP_SUCCESS
	else

		call MPI_COMM_SIZE (MPI_COMM_WORLD, num_procs, ierror)
		!c	should test ierr

		call MPI_COMM_RANK (MPI_COMM_WORLD, this_proc, ierror)
		!c	should test ierr

		dp_rank = this_proc

		dp_num_procs = num_procs
	endif

	dp_initialized = .true.

	ierror = DP_SUCCESS

	return

end SUBROUTINE DP_INIT 








SUBROUTINE DP_SETUP_PROC_ARRAY (proc_ndims, proc_dims, &
                my_rank, my_coords, ierr)
!c
	integer, intent(IN) ::  proc_ndims
	integer, intent(IN) ::  proc_dims(proc_ndims)

	integer, intent(OUT) ::  my_rank
	integer, intent(OUT) ::  my_coords(proc_ndims)
	integer, intent(OUT) ::  ierr

!c	local
	logical periods(DP_MAX_DIMS)
	integer ierror
	integer num_ranks
	integer nprocs
	integer i


	if (.not. dp_initialized) then
		ierr = DP_NOT_INITIALIZED
		return
	endif

	nprocs = 1
	DO i = 1, proc_ndims
		nprocs = nprocs * proc_dims(i)
	ENDDO

	if (nprocs .ne. dp_num_procs) then
		ierr = DP_BAD_NUM_PROCS
		return
	endif

	if (dp_comm_cart .ne. MPI_COMM_NULL) then
		call MPI_COMM_FREE (dp_comm_cart)
		dp_comm_cart = MPI_COMM_NULL
	endif



	if (nprocs .eq. 1) then
		dp_comm_cart = MPI_COMM_NULL
		dp_proc_arr_size = 1
		dp_coords = 0
		my_coords = 0
		my_rank = dp_rank
		ierr = DP_SUCCESS
		dp_initialized = .true. 

	else   !!   more than one processor


		if ( (proc_ndims .lt. 1) .or.    &
		     (proc_ndims .gt. DP_MAX_DIMS) ) then
		!		error !!!!
			ierr = DP_BAD_PROC_NDIMS
			return
		endif

		dp_proc_arr_size = 1
		dp_proc_arr_size(1:proc_ndims) = proc_dims


		periods = .true.

		call MPI_CART_CREATE (MPI_COMM_WORLD, DP_MAX_DIMS,  &
			dp_proc_arr_size, periods, .false., dp_comm_cart, ierr)
		!c	should test ierr

		call MPI_COMM_RANK(MPI_COMM_WORLD, my_rank, ierr)
		!c	should test ierr

		dp_rank = my_rank


		call MPI_CART_COORDS (dp_comm_cart, my_rank, DP_MAX_DIMS, &
				dp_coords, ierr)
		my_coords = dp_coords(1:proc_ndims)
		!c	should test ierr


		ierr = DP_SUCCESS
		dp_initialized = .true.
	endif

	return


end subroutine DP_SETUP_PROC_ARRAY





SUBROUTINE DP_INITIALIZE (proc_ndims, proc_dims, &
		my_rank, my_coords, ierr)
!c
!c	IN:
	integer proc_ndims
	integer proc_dims(proc_ndims)

!c	OUT:
	integer my_rank
	integer my_coords(proc_ndims)
	integer ierr

!c	local
	logical periods(DP_MAX_DIMS)
	integer ierror
	integer num_ranks
	integer nprocs
	integer i


	call dp_init (num_ranks, my_rank, ierr) 
	! check ierror

	call dp_setup_proc_array (proc_ndims, proc_dims, &
                my_rank, my_coords, ierr)
	! check ierror

	return

end subroutine dp_initialize
	

SUBROUTINE DP_RUN_SERIAL (ierror)
integer, intent(OUT) :: ierror

	if (dp_initialized) then
		ierror = DP_BAD_CALL_SEQUENCE
		return
	endif
	dp_serial_run = .true.

end subroutine DP_RUN_SERIAL


SUBROUTINE DP_FINALIZE (ierror)
	integer ierror


	if (.not. dp_initialized) then
		ierror = DP_NOT_INITIALIZED
		return
	endif

	if (dp_mpi_initialized) then
		call MPI_FINALIZE(ierror)
	endif

	dp_initialized = .false.

	return
end subroutine dp_finalize





SUBROUTINE DP_QUERY_LAYOUT (  &
		ndims, full_size, index_base, block_size, block_index, ierr)
!	IN:
	integer, intent(IN) :: ndims
	integer, intent(IN) :: full_size(ndims)
	integer, intent(IN) :: index_base
	
!	OUT:
	integer, intent(OUT) :: block_size (ndims)
	integer, intent(OUT) :: block_index (0:1, ndims)
	integer, intent(OUT) :: ierr

!       LOCAL
	integer i

	if (.not. dp_initialized) then
		ierr = -1
		return
	endif

	do i = 1, ndims

print *, "query: full_size = ", i, '|', full_size

		call dp_get_local_array_block (i, dp_coords(i), &
			index_base, index_base + full_size(i) - 1,  &
			block_index(0,i), block_index(1,i), ierr)
print *, "done query: full_size = ", full_size(i)
		block_size(i) = 1 + block_index(1,i) - block_index(0,i)
print *, "done2 query: full_size = ", full_size(i)

	enddo


	ierr = 0
	return
end subroutine dp_query_layout


SUBROUTINE DP_SYNC ()

	integer ierr

	if (dp_initialized .and. (dp_num_procs .gt. 1) ) then

		CALL MPI_Barrier (dp_comm_cart, ierr) 
		! should test ierr

	endif

end subroutine dp_sync




SUBROUTINE DP_GET_NUM_PROCS (num_procs, ierr)
	integer, intent (OUT) :: num_procs
	integer, intent (OUT) :: ierr


	if (dp_serial_run) then
		ierr = DP_SUCCESS
		num_procs = 1
	else

		if ( .not. dp_mpi_initialized ) then
			call MPI_INIT(ierr)
			! check ierr
			dp_mpi_initialized = .TRUE.
		endif
	

		call MPI_COMM_SIZE (MPI_COMM_WORLD, num_procs, ierr)
	endif

	return

end SUBROUTINE DP_GET_NUM_PROCS



	end module dp_util
