
module DPARLIB


	use dp_io
	use dp_reduce_arith
	use dp_reduce_maxmin
	use dp_reduce_log
	use dp_shift
	use dp_random
	use dp_util
	use dp_alloc
	use dp_public_constants

	!!
	!! public DP_MAX_DIMS 
	!! public DP_SUCCESS
	!! public DP_NOT_INITIALIZED
	!! public DP_INCONSISTENT_ARRAY_SIZE
	!! public DP_BAD_NUM_PROCS
	!! public DP_BAD_PROC_NDIMS
	!! public DP_BAD_CALL_SEQUENCE



	! The following routines are public:
	!	DP_INITIALIZE
	!	DP_FINALIZE
	!	DP_QUERY_LAYOUT
	!	DP_UNI_INIT
	!	DP_UNI

	! These are generic interfaces to type-specific routines:
	!	DP_CSHIFT
	!	DP_EOSHIFT
	! 	DP_SUM
	!	DP_MAXVAL
	!	DP_WRITE_ARRAY
	!	DP_READ_ARRAY


	! Here are the generic interface blocks:


	interface DP_CSHIFT

		module procedure DP_CSHIFT_1D_D
		module procedure DP_CSHIFT_2D_D
		module procedure DP_CSHIFT_3D_D
		module procedure DP_CSHIFT_4D_D

		module procedure DP_CSHIFT_1D_R
		module procedure DP_CSHIFT_2D_R
		module procedure DP_CSHIFT_3D_R
		module procedure DP_CSHIFT_4D_R


#if 0
!		module procedure DP_CSHIFT_1D_R4
!		module procedure DP_CSHIFT_2D_R4
!		module procedure DP_CSHIFT_3D_R4
!		module procedure DP_CSHIFT_4D_R4

!		module procedure DP_CSHIFT_1D_R8
!		module procedure DP_CSHIFT_2D_R8
!		module procedure DP_CSHIFT_3D_R8
!		module procedure DP_CSHIFT_4D_R8

		module procedure DP_CSHIFT_1D_R16
		module procedure DP_CSHIFT_2D_R16
		module procedure DP_CSHIFT_3D_R16
		module procedure DP_CSHIFT_4D_R16
#endif


		module procedure DP_CSHIFT_1D_I
		module procedure DP_CSHIFT_2D_I
		module procedure DP_CSHIFT_3D_I
		module procedure DP_CSHIFT_4D_I

#if 0
		module procedure DP_CSHIFT_1D_I1
		module procedure DP_CSHIFT_2D_I1
		module procedure DP_CSHIFT_3D_I1
		module procedure DP_CSHIFT_4D_I1

		module procedure DP_CSHIFT_1D_I2
		module procedure DP_CSHIFT_2D_I2
		module procedure DP_CSHIFT_3D_I2
		module procedure DP_CSHIFT_4D_I2

!		module procedure DP_CSHIFT_1D_I4
!		module procedure DP_CSHIFT_2D_I4
!		module procedure DP_CSHIFT_3D_I4
!		module procedure DP_CSHIFT_4D_I4
#endif


		module procedure DP_CSHIFT_1D_L
		module procedure DP_CSHIFT_2D_L
		module procedure DP_CSHIFT_3D_L
		module procedure DP_CSHIFT_4D_L

#if 0
		module procedure DP_CSHIFT_1D_L1
		module procedure DP_CSHIFT_2D_L1
		module procedure DP_CSHIFT_3D_L1
		module procedure DP_CSHIFT_4D_L1

		module procedure DP_CSHIFT_1D_L2
		module procedure DP_CSHIFT_2D_L2
		module procedure DP_CSHIFT_3D_L2
		module procedure DP_CSHIFT_4D_L2

!		module procedure DP_CSHIFT_1D_L4
!		module procedure DP_CSHIFT_2D_L4
!		module procedure DP_CSHIFT_3D_L4
!		module procedure DP_CSHIFT_4D_L4
#endif

		module procedure DP_CSHIFT_1D_C
		module procedure DP_CSHIFT_2D_C
		module procedure DP_CSHIFT_3D_C
		module procedure DP_CSHIFT_4D_C

#if 0
!		module procedure DP_CSHIFT_1D_C8
!		module procedure DP_CSHIFT_2D_C8
!		module procedure DP_CSHIFT_3D_C8
!		module procedure DP_CSHIFT_4D_C8

		module procedure DP_CSHIFT_1D_C16
		module procedure DP_CSHIFT_2D_C16
		module procedure DP_CSHIFT_3D_C16
		module procedure DP_CSHIFT_4D_C16

		module procedure DP_CSHIFT_1D_C32
		module procedure DP_CSHIFT_2D_C32
		module procedure DP_CSHIFT_3D_C32
		module procedure DP_CSHIFT_4D_C32
#endif

	end interface


	interface DP_EOSHIFT

		module procedure DP_EOSHIFT_1D_D
		module procedure DP_EOSHIFT_2D_D
		module procedure DP_EOSHIFT_3D_D
		module procedure DP_EOSHIFT_4D_D

		module procedure DP_EOSHIFT_1D_R
		module procedure DP_EOSHIFT_2D_R
		module procedure DP_EOSHIFT_3D_R
		module procedure DP_EOSHIFT_4D_R

#if 0
!		module procedure DP_EOSHIFT_1D_R4
!		module procedure DP_EOSHIFT_2D_R4
!		module procedure DP_EOSHIFT_3D_R4
!		module procedure DP_EOSHIFT_4D_R4

!		module procedure DP_EOSHIFT_1D_R8
!		module procedure DP_EOSHIFT_2D_R8
!		module procedure DP_EOSHIFT_3D_R8
!		module procedure DP_EOSHIFT_4D_R8

		module procedure DP_EOSHIFT_1D_R16
		module procedure DP_EOSHIFT_2D_R16
		module procedure DP_EOSHIFT_3D_R16
		module procedure DP_EOSHIFT_4D_R16
#endif

		module procedure DP_EOSHIFT_1D_I
		module procedure DP_EOSHIFT_2D_I
		module procedure DP_EOSHIFT_3D_I
		module procedure DP_EOSHIFT_4D_I

#if 0
		module procedure DP_EOSHIFT_1D_I1
		module procedure DP_EOSHIFT_2D_I1
		module procedure DP_EOSHIFT_3D_I1
		module procedure DP_EOSHIFT_4D_I1

		module procedure DP_EOSHIFT_1D_I2
		module procedure DP_EOSHIFT_2D_I2
		module procedure DP_EOSHIFT_3D_I2
		module procedure DP_EOSHIFT_4D_I2

!		module procedure DP_EOSHIFT_1D_I4
!		module procedure DP_EOSHIFT_2D_I4
!		module procedure DP_EOSHIFT_3D_I4
!		module procedure DP_EOSHIFT_4D_I4
#endif

		module procedure DP_EOSHIFT_1D_L
		module procedure DP_EOSHIFT_2D_L
		module procedure DP_EOSHIFT_3D_L
		module procedure DP_EOSHIFT_4D_L

#if 0
		module procedure DP_EOSHIFT_1D_L1
		module procedure DP_EOSHIFT_2D_L1
		module procedure DP_EOSHIFT_3D_L1
		module procedure DP_EOSHIFT_4D_L1

		module procedure DP_EOSHIFT_1D_L2
		module procedure DP_EOSHIFT_2D_L2
		module procedure DP_EOSHIFT_3D_L2
		module procedure DP_EOSHIFT_4D_L2

!		module procedure DP_EOSHIFT_1D_L4
!		module procedure DP_EOSHIFT_2D_L4
!		module procedure DP_EOSHIFT_3D_L4
!		module procedure DP_EOSHIFT_4D_L4
#endif

		module procedure DP_EOSHIFT_1D_C
		module procedure DP_EOSHIFT_2D_C
		module procedure DP_EOSHIFT_3D_C
		module procedure DP_EOSHIFT_4D_C

#if 0
!		module procedure DP_EOSHIFT_1D_C8
!		module procedure DP_EOSHIFT_2D_C8
!		module procedure DP_EOSHIFT_3D_C8
!		module procedure DP_EOSHIFT_4D_C8

		module procedure DP_EOSHIFT_1D_C16
		module procedure DP_EOSHIFT_2D_C16
		module procedure DP_EOSHIFT_3D_C16
		module procedure DP_EOSHIFT_4D_C16

		module procedure DP_EOSHIFT_1D_C32
		module procedure DP_EOSHIFT_2D_C32
		module procedure DP_EOSHIFT_3D_C32
		module procedure DP_EOSHIFT_4D_C32
#endif

	end interface


	interface DP_UNI

		module procedure DP_UNI_1D
		module procedure DP_UNI_2D
		module procedure DP_UNI_3D
		module procedure DP_UNI_4D
		module procedure DP_UNI_SCALAR

	end interface



	interface DP_SUM

               module procedure dp_sum_1D_R 
               module procedure dp_sum_2D_R 
               module procedure dp_sum_3D_R 
               module procedure dp_sum_4D_R 

#if 0
!               module procedure dp_sum_1D_R4 
!               module procedure dp_sum_2D_R4 
!               module procedure dp_sum_3D_R4 
!               module procedure dp_sum_4D_R4 
!               module procedure dp_sum_1D_R8 
!               module procedure dp_sum_2D_R8 
!               module procedure dp_sum_3D_R8 
!               module procedure dp_sum_4D_R8 
               module procedure dp_sum_1D_R16 
               module procedure dp_sum_2D_R16 
               module procedure dp_sum_3D_R16 
               module procedure dp_sum_4D_R16 
#endif

               module procedure dp_sum_1D_D 
               module procedure dp_sum_2D_D 
               module procedure dp_sum_3D_D 
               module procedure dp_sum_4D_D 
               module procedure dp_sum_1D_I 
               module procedure dp_sum_2D_I 
               module procedure dp_sum_3D_I 
               module procedure dp_sum_4D_I 

#if 0
               module procedure dp_sum_1D_I1 
               module procedure dp_sum_2D_I1 
               module procedure dp_sum_3D_I1 
               module procedure dp_sum_4D_I1 
               module procedure dp_sum_1D_I2 
               module procedure dp_sum_2D_I2 
               module procedure dp_sum_3D_I2 
               module procedure dp_sum_4D_I2 
!               module procedure dp_sum_1D_I4 
!               module procedure dp_sum_2D_I4 
!               module procedure dp_sum_3D_I4 
!               module procedure dp_sum_4D_I4 
#endif

               module procedure dp_sum_1D_C
               module procedure dp_sum_2D_C
               module procedure dp_sum_3D_C
               module procedure dp_sum_4D_C

#if 0
!               module procedure dp_sum_1D_C8
!               module procedure dp_sum_2D_C8
!               module procedure dp_sum_3D_C8
!               module procedure dp_sum_4D_C8

               module procedure dp_sum_1D_C16
               module procedure dp_sum_2D_C16
               module procedure dp_sum_3D_C16
               module procedure dp_sum_4D_C16

               module procedure dp_sum_1D_C32
               module procedure dp_sum_2D_C32
               module procedure dp_sum_3D_C32
               module procedure dp_sum_4D_C32
#endif

	end interface



	interface DP_PRODUCT
               module procedure dp_product_1D_R 
               module procedure dp_product_2D_R 
               module procedure dp_product_3D_R 
               module procedure dp_product_4D_R 

#if 0
!               module procedure dp_product_1D_R4 
!               module procedure dp_product_2D_R4 
!               module procedure dp_product_3D_R4 
!               module procedure dp_product_4D_R4 
!               module procedure dp_product_1D_R8 
!               module procedure dp_product_2D_R8 
!               module procedure dp_product_3D_R8 
!               module procedure dp_product_4D_R8 
               module procedure dp_product_1D_R16 
               module procedure dp_product_2D_R16 
               module procedure dp_product_3D_R16 
               module procedure dp_product_4D_R16 
#endif

               module procedure dp_product_1D_D 
               module procedure dp_product_2D_D 
               module procedure dp_product_3D_D 
               module procedure dp_product_4D_D 
               module procedure dp_product_1D_I 
               module procedure dp_product_2D_I 
               module procedure dp_product_3D_I 
               module procedure dp_product_4D_I 

#if 0
               module procedure dp_product_1D_I1 
               module procedure dp_product_2D_I1 
               module procedure dp_product_3D_I1 
               module procedure dp_product_4D_I1 
               module procedure dp_product_1D_I2 
               module procedure dp_product_2D_I2 
               module procedure dp_product_3D_I2 
               module procedure dp_product_4D_I2 
!               module procedure dp_product_1D_I4 
!               module procedure dp_product_2D_I4 
!               module procedure dp_product_3D_I4 
!               module procedure dp_product_4D_I4 
#endif


               module procedure dp_product_1D_C
               module procedure dp_product_2D_C
               module procedure dp_product_3D_C
               module procedure dp_product_4D_C

#if 0
!               module procedure dp_product_1D_C8
!               module procedure dp_product_2D_C8
!               module procedure dp_product_3D_C8
!               module procedure dp_product_4D_C8

               module procedure dp_product_1D_C16
               module procedure dp_product_2D_C16
               module procedure dp_product_3D_C16
               module procedure dp_product_4D_C16

               module procedure dp_product_1D_C32
               module procedure dp_product_2D_C32
               module procedure dp_product_3D_C32
               module procedure dp_product_4D_C32
#endif


	end interface




	interface DP_DOT_PRODUCT
               module procedure dp_dot_product_1D_R 
               module procedure dp_dot_product_2D_R 
               module procedure dp_dot_product_3D_R 
               module procedure dp_dot_product_4D_R 

#if 0
!               module procedure dp_dot_product_1D_R4 
!               module procedure dp_dot_product_2D_R4 
!               module procedure dp_dot_product_3D_R4 
!               module procedure dp_dot_product_4D_R4 
!               module procedure dp_dot_product_1D_R8 
!               module procedure dp_dot_product_2D_R8 
!               module procedure dp_dot_product_3D_R8 
!               module procedure dp_dot_product_4D_R8 
               module procedure dp_dot_product_1D_R16 
               module procedure dp_dot_product_2D_R16 
               module procedure dp_dot_product_3D_R16 
               module procedure dp_dot_product_4D_R16 
#endif

               module procedure dp_dot_product_1D_D 
               module procedure dp_dot_product_2D_D 
               module procedure dp_dot_product_3D_D 
               module procedure dp_dot_product_4D_D 
               module procedure dp_dot_product_1D_I 
               module procedure dp_dot_product_2D_I 
               module procedure dp_dot_product_3D_I 
               module procedure dp_dot_product_4D_I 

#if 0
               module procedure dp_dot_product_1D_I1 
               module procedure dp_dot_product_2D_I1 
               module procedure dp_dot_product_3D_I1 
               module procedure dp_dot_product_4D_I1 
               module procedure dp_dot_product_1D_I2 
               module procedure dp_dot_product_2D_I2 
               module procedure dp_dot_product_3D_I2 
               module procedure dp_dot_product_4D_I2 
!               module procedure dp_dot_product_1D_I4 
!               module procedure dp_dot_product_2D_I4 
!               module procedure dp_dot_product_3D_I4 
!               module procedure dp_dot_product_4D_I4 
#endif


               module procedure dp_dot_product_1D_C
               module procedure dp_dot_product_2D_C
               module procedure dp_dot_product_3D_C
               module procedure dp_dot_product_4D_C

#if 0
!               module procedure dp_dot_product_1D_C8
!               module procedure dp_dot_product_2D_C8
!               module procedure dp_dot_product_3D_C8
!               module procedure dp_dot_product_4D_C8

               module procedure dp_dot_product_1D_C16
               module procedure dp_dot_product_2D_C16
               module procedure dp_dot_product_3D_C16
               module procedure dp_dot_product_4D_C16

               module procedure dp_dot_product_1D_C32
               module procedure dp_dot_product_2D_C32
               module procedure dp_dot_product_3D_C32
               module procedure dp_dot_product_4D_C32
#endif


	end interface




	interface DP_MINVAL

               module procedure dp_minval_1D_R 
               module procedure dp_minval_2D_R 
               module procedure dp_minval_3D_R 
               module procedure dp_minval_4D_R 

#if 0
!               module procedure dp_minval_1D_R4 
!               module procedure dp_minval_2D_R4 
!               module procedure dp_minval_3D_R4 
!               module procedure dp_minval_4D_R4 
!               module procedure dp_minval_1D_R8 
!               module procedure dp_minval_2D_R8 
!               module procedure dp_minval_3D_R8 
!               module procedure dp_minval_4D_R8 
               module procedure dp_minval_1D_R16 
               module procedure dp_minval_2D_R16 
               module procedure dp_minval_3D_R16 
               module procedure dp_minval_4D_R16 
#endif

               module procedure dp_minval_1D_D 
               module procedure dp_minval_2D_D 
               module procedure dp_minval_3D_D 
               module procedure dp_minval_4D_D 
               module procedure dp_minval_1D_I 
               module procedure dp_minval_2D_I 
               module procedure dp_minval_3D_I 
               module procedure dp_minval_4D_I 

#if 0
               module procedure dp_minval_1D_I1 
               module procedure dp_minval_2D_I1 
               module procedure dp_minval_3D_I1 
               module procedure dp_minval_4D_I1 
               module procedure dp_minval_1D_I2 
               module procedure dp_minval_2D_I2 
               module procedure dp_minval_3D_I2 
               module procedure dp_minval_4D_I2 
!               module procedure dp_minval_1D_I4 
!               module procedure dp_minval_2D_I4 
!               module procedure dp_minval_3D_I4 
!               module procedure dp_minval_4D_I4 
#endif


	end interface




	interface DP_MAXVAL

               module procedure dp_maxval_1D_R 
               module procedure dp_maxval_2D_R 
               module procedure dp_maxval_3D_R 
               module procedure dp_maxval_4D_R 

#if 0
!               module procedure dp_maxval_1D_R4 
!               module procedure dp_maxval_2D_R4 
!               module procedure dp_maxval_3D_R4 
!               module procedure dp_maxval_4D_R4 
!               module procedure dp_maxval_1D_R8 
!               module procedure dp_maxval_2D_R8 
!               module procedure dp_maxval_3D_R8 
!               module procedure dp_maxval_4D_R8 
               module procedure dp_maxval_1D_R16 
               module procedure dp_maxval_2D_R16 
               module procedure dp_maxval_3D_R16 
               module procedure dp_maxval_4D_R16 
#endif

               module procedure dp_maxval_1D_D 
               module procedure dp_maxval_2D_D 
               module procedure dp_maxval_3D_D 
               module procedure dp_maxval_4D_D 
               module procedure dp_maxval_1D_I 
               module procedure dp_maxval_2D_I 
               module procedure dp_maxval_3D_I 
               module procedure dp_maxval_4D_I 

#if 0
               module procedure dp_maxval_1D_I1 
               module procedure dp_maxval_2D_I1 
               module procedure dp_maxval_3D_I1 
               module procedure dp_maxval_4D_I1 
               module procedure dp_maxval_1D_I2 
               module procedure dp_maxval_2D_I2 
               module procedure dp_maxval_3D_I2 
               module procedure dp_maxval_4D_I2 
!               module procedure dp_maxval_1D_I4 
!               module procedure dp_maxval_2D_I4 
!               module procedure dp_maxval_3D_I4 
!               module procedure dp_maxval_4D_I4 
#endif


	end interface




	interface DP_MAXLOC

               module procedure dp_maxloc_1D_R 
               module procedure dp_maxloc_2D_R 
               module procedure dp_maxloc_3D_R 
               module procedure dp_maxloc_4D_R 

#if 0
!               module procedure dp_maxloc_1D_R4 
!               module procedure dp_maxloc_2D_R4 
!               module procedure dp_maxloc_3D_R4 
!               module procedure dp_maxloc_4D_R4 
!               module procedure dp_maxloc_1D_R8 
!               module procedure dp_maxloc_2D_R8 
!               module procedure dp_maxloc_3D_R8 
!               module procedure dp_maxloc_4D_R8 
               module procedure dp_maxloc_1D_R16 
               module procedure dp_maxloc_2D_R16 
               module procedure dp_maxloc_3D_R16 
               module procedure dp_maxloc_4D_R16 
#endif

               module procedure dp_maxloc_1D_D 
               module procedure dp_maxloc_2D_D 
               module procedure dp_maxloc_3D_D 
               module procedure dp_maxloc_4D_D 
               module procedure dp_maxloc_1D_I 
               module procedure dp_maxloc_2D_I 
               module procedure dp_maxloc_3D_I 
               module procedure dp_maxloc_4D_I 

#if 0
               module procedure dp_maxloc_1D_I1 
               module procedure dp_maxloc_2D_I1 
               module procedure dp_maxloc_3D_I1 
               module procedure dp_maxloc_4D_I1 
               module procedure dp_maxloc_1D_I2 
               module procedure dp_maxloc_2D_I2 
               module procedure dp_maxloc_3D_I2 
               module procedure dp_maxloc_4D_I2 
!               module procedure dp_maxloc_1D_I4 
!               module procedure dp_maxloc_2D_I4 
!               module procedure dp_maxloc_3D_I4 
!               module procedure dp_maxloc_4D_I4 
#endif


	end interface




	interface DP_MINLOC

               module procedure dp_minloc_1D_R 
               module procedure dp_minloc_2D_R 
               module procedure dp_minloc_3D_R 
               module procedure dp_minloc_4D_R 

#if 0
!               module procedure dp_minloc_1D_R4 
!               module procedure dp_minloc_2D_R4 
!               module procedure dp_minloc_3D_R4 
!               module procedure dp_minloc_4D_R4 
!               module procedure dp_minloc_1D_R8 
!               module procedure dp_minloc_2D_R8 
!               module procedure dp_minloc_3D_R8 
!               module procedure dp_minloc_4D_R8 
               module procedure dp_minloc_1D_R16 
               module procedure dp_minloc_2D_R16 
               module procedure dp_minloc_3D_R16 
               module procedure dp_minloc_4D_R16 
#endif

               module procedure dp_minloc_1D_D 
               module procedure dp_minloc_2D_D 
               module procedure dp_minloc_3D_D 
               module procedure dp_minloc_4D_D 
               module procedure dp_minloc_1D_I 
               module procedure dp_minloc_2D_I 
               module procedure dp_minloc_3D_I 
               module procedure dp_minloc_4D_I 

#if 0
               module procedure dp_minloc_1D_I1 
               module procedure dp_minloc_2D_I1 
               module procedure dp_minloc_3D_I1 
               module procedure dp_minloc_4D_I1 
               module procedure dp_minloc_1D_I2 
               module procedure dp_minloc_2D_I2 
               module procedure dp_minloc_3D_I2 
               module procedure dp_minloc_4D_I2 
!               module procedure dp_minloc_1D_I4 
!               module procedure dp_minloc_2D_I4 
!               module procedure dp_minloc_3D_I4 
!               module procedure dp_minloc_4D_I4 
#endif


	end interface




	interface DP_ALL

               module procedure dp_all_1D 
               module procedure dp_all_2D 
               module procedure dp_all_3D 
               module procedure dp_all_4D 

	end interface




	interface DP_ANY

               module procedure dp_any_1D 
               module procedure dp_any_2D 
               module procedure dp_any_3D 
               module procedure dp_any_4D 

	end interface




	interface DP_COUNT
		module procedure DP_COUNT_1D 
		module procedure DP_COUNT_2D 
		module procedure DP_COUNT_3D 
		module procedure DP_COUNT_4D 
	end interface





	interface DP_WRITE_ARRAY

		module procedure DP_WRITE_ARRAY_1D_D 
		module procedure DP_WRITE_ARRAY_2D_D 
		module procedure DP_WRITE_ARRAY_3D_D 
		module procedure DP_WRITE_ARRAY_4D_D 

		module procedure DP_WRITE_ARRAY_1D_R 
		module procedure DP_WRITE_ARRAY_2D_R 
		module procedure DP_WRITE_ARRAY_3D_R 
		module procedure DP_WRITE_ARRAY_4D_R 

#if 0
!		module procedure DP_WRITE_ARRAY_1D_R4
!		module procedure DP_WRITE_ARRAY_2D_R4
!		module procedure DP_WRITE_ARRAY_3D_R4
!		module procedure DP_WRITE_ARRAY_4D_R4

!		module procedure DP_WRITE_ARRAY_1D_R8
!		module procedure DP_WRITE_ARRAY_2D_R8
!		module procedure DP_WRITE_ARRAY_3D_R8
!		module procedure DP_WRITE_ARRAY_4D_R8

		module procedure DP_WRITE_ARRAY_1D_R16
		module procedure DP_WRITE_ARRAY_2D_R16
		module procedure DP_WRITE_ARRAY_3D_R16
		module procedure DP_WRITE_ARRAY_4D_R16
#endif


		module procedure DP_WRITE_ARRAY_1D_I 
		module procedure DP_WRITE_ARRAY_2D_I 
		module procedure DP_WRITE_ARRAY_3D_I 
		module procedure DP_WRITE_ARRAY_4D_I 

#if 0
		module procedure DP_WRITE_ARRAY_1D_I1
		module procedure DP_WRITE_ARRAY_2D_I1
		module procedure DP_WRITE_ARRAY_3D_I1
		module procedure DP_WRITE_ARRAY_4D_I1

		module procedure DP_WRITE_ARRAY_1D_I2
		module procedure DP_WRITE_ARRAY_2D_I2
		module procedure DP_WRITE_ARRAY_3D_I2
		module procedure DP_WRITE_ARRAY_4D_I2

!		module procedure DP_WRITE_ARRAY_1D_I4
!		module procedure DP_WRITE_ARRAY_2D_I4
!		module procedure DP_WRITE_ARRAY_3D_I4
!		module procedure DP_WRITE_ARRAY_4D_I4
#endif

		module procedure DP_WRITE_ARRAY_1D_L 
		module procedure DP_WRITE_ARRAY_2D_L 
		module procedure DP_WRITE_ARRAY_3D_L 
		module procedure DP_WRITE_ARRAY_4D_L 

#if 0
		module procedure DP_WRITE_ARRAY_1D_L1
		module procedure DP_WRITE_ARRAY_2D_L1
		module procedure DP_WRITE_ARRAY_3D_L1
		module procedure DP_WRITE_ARRAY_4D_L1

		module procedure DP_WRITE_ARRAY_1D_L2
		module procedure DP_WRITE_ARRAY_2D_L2
		module procedure DP_WRITE_ARRAY_3D_L2
		module procedure DP_WRITE_ARRAY_4D_L2

!		module procedure DP_WRITE_ARRAY_1D_L4
!		module procedure DP_WRITE_ARRAY_2D_L4
!		module procedure DP_WRITE_ARRAY_3D_L4
!		module procedure DP_WRITE_ARRAY_4D_L4
#endif

		module procedure DP_WRITE_ARRAY_1D_C
		module procedure DP_WRITE_ARRAY_2D_C
		module procedure DP_WRITE_ARRAY_3D_C
		module procedure DP_WRITE_ARRAY_4D_C

#if 0
!		module procedure DP_WRITE_ARRAY_1D_C8
!		module procedure DP_WRITE_ARRAY_2D_C8
!		module procedure DP_WRITE_ARRAY_3D_C8
!		module procedure DP_WRITE_ARRAY_4D_C8

		module procedure DP_WRITE_ARRAY_1D_C16
		module procedure DP_WRITE_ARRAY_2D_C16
		module procedure DP_WRITE_ARRAY_3D_C16
		module procedure DP_WRITE_ARRAY_4D_C16

		module procedure DP_WRITE_ARRAY_1D_C32
		module procedure DP_WRITE_ARRAY_2D_C32
		module procedure DP_WRITE_ARRAY_3D_C32
		module procedure DP_WRITE_ARRAY_4D_C32
#endif


	end interface


	interface DP_READ_ARRAY

		module procedure DP_READ_ARRAY_1D_D
		module procedure DP_READ_ARRAY_2D_D
		module procedure DP_READ_ARRAY_3D_D
		module procedure DP_READ_ARRAY_4D_D

		module procedure DP_READ_ARRAY_1D_R
		module procedure DP_READ_ARRAY_2D_R
		module procedure DP_READ_ARRAY_3D_R
		module procedure DP_READ_ARRAY_4D_R

#if 0
!		module procedure DP_READ_ARRAY_1D_R4
!		module procedure DP_READ_ARRAY_2D_R4
!		module procedure DP_READ_ARRAY_3D_R4
!		module procedure DP_READ_ARRAY_4D_R4

!		module procedure DP_READ_ARRAY_1D_R8
!		module procedure DP_READ_ARRAY_2D_R8
!		module procedure DP_READ_ARRAY_3D_R8
!		module procedure DP_READ_ARRAY_4D_R8

		module procedure DP_READ_ARRAY_1D_R16
		module procedure DP_READ_ARRAY_2D_R16
		module procedure DP_READ_ARRAY_3D_R16
		module procedure DP_READ_ARRAY_4D_R16
#endif

		module procedure DP_READ_ARRAY_1D_I
		module procedure DP_READ_ARRAY_2D_I
		module procedure DP_READ_ARRAY_3D_I
		module procedure DP_READ_ARRAY_4D_I

#if 0
		module procedure DP_READ_ARRAY_1D_I1
		module procedure DP_READ_ARRAY_2D_I1
		module procedure DP_READ_ARRAY_3D_I1
		module procedure DP_READ_ARRAY_4D_I1

		module procedure DP_READ_ARRAY_1D_I2
		module procedure DP_READ_ARRAY_2D_I2
		module procedure DP_READ_ARRAY_3D_I2
		module procedure DP_READ_ARRAY_4D_I2

!		module procedure DP_READ_ARRAY_1D_I4
!		module procedure DP_READ_ARRAY_2D_I4
!		module procedure DP_READ_ARRAY_3D_I4
!		module procedure DP_READ_ARRAY_4D_I4

#endif
		module procedure DP_READ_ARRAY_1D_L
		module procedure DP_READ_ARRAY_2D_L
		module procedure DP_READ_ARRAY_3D_L
		module procedure DP_READ_ARRAY_4D_L

#if 0
		module procedure DP_READ_ARRAY_1D_L1
		module procedure DP_READ_ARRAY_2D_L1
		module procedure DP_READ_ARRAY_3D_L1
		module procedure DP_READ_ARRAY_4D_L1

		module procedure DP_READ_ARRAY_1D_L2
		module procedure DP_READ_ARRAY_2D_L2
		module procedure DP_READ_ARRAY_3D_L2
		module procedure DP_READ_ARRAY_4D_L2

!		module procedure DP_READ_ARRAY_1D_L4
!		module procedure DP_READ_ARRAY_2D_L4
!		module procedure DP_READ_ARRAY_3D_L4
!		module procedure DP_READ_ARRAY_4D_L4
#endif

		module procedure DP_READ_ARRAY_1D_C
		module procedure DP_READ_ARRAY_2D_C
		module procedure DP_READ_ARRAY_3D_C
		module procedure DP_READ_ARRAY_4D_C

#if 0
!		module procedure DP_READ_ARRAY_1D_C8
!		module procedure DP_READ_ARRAY_2D_C8
!		module procedure DP_READ_ARRAY_3D_C8
!		module procedure DP_READ_ARRAY_4D_C8

		module procedure DP_READ_ARRAY_1D_C16
		module procedure DP_READ_ARRAY_2D_C16
		module procedure DP_READ_ARRAY_3D_C16
		module procedure DP_READ_ARRAY_4D_C16

		module procedure DP_READ_ARRAY_1D_C32
		module procedure DP_READ_ARRAY_2D_C32
		module procedure DP_READ_ARRAY_3D_C32
		module procedure DP_READ_ARRAY_4D_C32
#endif

	end interface






	interface DP_ALLOC_ARRAY

		module procedure dp_alloc_array_A_1D_C
#if 0
		module procedure dp_alloc_array_A_1D_C16
		module procedure dp_alloc_array_A_1D_C32
#endif
		module procedure dp_alloc_array_A_1D_D
		module procedure dp_alloc_array_A_1D_I

#if 0
		module procedure dp_alloc_array_A_1D_I1
		module procedure dp_alloc_array_A_1D_I2
#endif
		module procedure dp_alloc_array_A_1D_R

#if 0
		module procedure dp_alloc_array_A_1D_R16
#endif
		module procedure dp_alloc_array_A_2D_C

#if 0
		module procedure dp_alloc_array_A_2D_C16
		module procedure dp_alloc_array_A_2D_C32
#endif
		module procedure dp_alloc_array_A_2D_D
		module procedure dp_alloc_array_A_2D_I

#if 0
		module procedure dp_alloc_array_A_2D_I1
		module procedure dp_alloc_array_A_2D_I2
#endif
		module procedure dp_alloc_array_A_2D_R

#if 0
		module procedure dp_alloc_array_A_2D_R16
#endif
		module procedure dp_alloc_array_A_3D_C

#if 0
		module procedure dp_alloc_array_A_3D_C16
		module procedure dp_alloc_array_A_3D_C32
#endif
		module procedure dp_alloc_array_A_3D_D
		module procedure dp_alloc_array_A_3D_I

#if 0
		module procedure dp_alloc_array_A_3D_I1
		module procedure dp_alloc_array_A_3D_I2
#endif
		module procedure dp_alloc_array_A_3D_R

#if 0
		module procedure dp_alloc_array_A_3D_R16
#endif
		module procedure dp_alloc_array_A_4D_C

#if 0
		module procedure dp_alloc_array_A_4D_C16
		module procedure dp_alloc_array_A_4D_C32
#endif
		module procedure dp_alloc_array_A_4D_D
		module procedure dp_alloc_array_A_4D_I

#if 0
		module procedure dp_alloc_array_A_4D_I1
		module procedure dp_alloc_array_A_4D_I2
#endif
		module procedure dp_alloc_array_A_4D_R

#if 0
		module procedure dp_alloc_array_A_4D_R16
#endif

		module procedure dp_alloc_array_B_1D_C

#if 0
		module procedure dp_alloc_array_B_1D_C16
		module procedure dp_alloc_array_B_1D_C32
#endif

		module procedure dp_alloc_array_B_1D_D
		module procedure dp_alloc_array_B_1D_I

#if 0
		module procedure dp_alloc_array_B_1D_I1
		module procedure dp_alloc_array_B_1D_I2
#endif

		module procedure dp_alloc_array_B_1D_R

#if 0
		module procedure dp_alloc_array_B_1D_R16
#endif

		module procedure dp_alloc_array_B_2D_C

#if 0
		module procedure dp_alloc_array_B_2D_C16
		module procedure dp_alloc_array_B_2D_C32
#endif

		module procedure dp_alloc_array_B_2D_D

		module procedure dp_alloc_array_B_2D_I

#if 0
		module procedure dp_alloc_array_B_2D_I1
		module procedure dp_alloc_array_B_2D_I2
#endif

		module procedure dp_alloc_array_B_2D_R

#if 0
		module procedure dp_alloc_array_B_2D_R16
#endif

		module procedure dp_alloc_array_B_3D_C

#if 0
		module procedure dp_alloc_array_B_3D_C16
		module procedure dp_alloc_array_B_3D_C32
#endif

		module procedure dp_alloc_array_B_3D_D
		module procedure dp_alloc_array_B_3D_I

#if 0
		module procedure dp_alloc_array_B_3D_I1
		module procedure dp_alloc_array_B_3D_I2
#endif

		module procedure dp_alloc_array_B_3D_R

#if 0
		module procedure dp_alloc_array_B_3D_R16
#endif

		module procedure dp_alloc_array_B_4D_C

#if 0
		module procedure dp_alloc_array_B_4D_C16
		module procedure dp_alloc_array_B_4D_C32
#endif

		module procedure dp_alloc_array_B_4D_D
		module procedure dp_alloc_array_B_4D_I

#if 0
		module procedure dp_alloc_array_B_4D_I1
		module procedure dp_alloc_array_B_4D_I2
#endif

		module procedure dp_alloc_array_B_4D_R

#if 0
		module procedure dp_alloc_array_B_4D_R16
#endif


	end interface


	interface dp_find_arr_desc

		module procedure dp_find_arr_desc_1D_C 
		module procedure dp_find_arr_desc_1D_D 
		module procedure dp_find_arr_desc_1D_I 
		module procedure dp_find_arr_desc_1D_R 
		module procedure dp_find_arr_desc_2D_C 
		module procedure dp_find_arr_desc_2D_D 
		module procedure dp_find_arr_desc_2D_I 
		module procedure dp_find_arr_desc_2D_R 
		module procedure dp_find_arr_desc_3D_C 
		module procedure dp_find_arr_desc_3D_D 
		module procedure dp_find_arr_desc_3D_I 
		module procedure dp_find_arr_desc_3D_R 
		module procedure dp_find_arr_desc_4D_C 
		module procedure dp_find_arr_desc_4D_D 
		module procedure dp_find_arr_desc_4D_I 
		module procedure dp_find_arr_desc_4D_R 


#if 0
		module procedure dp_find_arr_desc_1D_C16 
		module procedure dp_find_arr_desc_1D_C32 
		module procedure dp_find_arr_desc_1D_I1 
		module procedure dp_find_arr_desc_1D_I2 
		module procedure dp_find_arr_desc_1D_R16 
		module procedure dp_find_arr_desc_2D_C16 
		module procedure dp_find_arr_desc_2D_C32 
		module procedure dp_find_arr_desc_2D_I1 
		module procedure dp_find_arr_desc_2D_I2 
		module procedure dp_find_arr_desc_2D_R16 
		module procedure dp_find_arr_desc_3D_C16 
		module procedure dp_find_arr_desc_3D_C32 
		module procedure dp_find_arr_desc_3D_I1 
		module procedure dp_find_arr_desc_3D_I2 
		module procedure dp_find_arr_desc_3D_R16 
		module procedure dp_find_arr_desc_4D_C16 
		module procedure dp_find_arr_desc_4D_C32 
		module procedure dp_find_arr_desc_4D_I1 
		module procedure dp_find_arr_desc_4D_I2 
		module procedure dp_find_arr_desc_4D_R16 
#endif

	end interface



	interface dp_register_array

		module procedure dp_register_array_1D_C
		module procedure dp_register_array_1D_D
		module procedure dp_register_array_1D_I
		module procedure dp_register_array_1D_R
		module procedure dp_register_array_1D_L
		module procedure dp_register_array_2D_C
		module procedure dp_register_array_2D_D
		module procedure dp_register_array_2D_I
		module procedure dp_register_array_2D_R
		module procedure dp_register_array_2D_L
		module procedure dp_register_array_3D_C
		module procedure dp_register_array_3D_D
		module procedure dp_register_array_3D_I
		module procedure dp_register_array_3D_R
		module procedure dp_register_array_3D_L
		module procedure dp_register_array_4D_C
		module procedure dp_register_array_4D_D
		module procedure dp_register_array_4D_I
		module procedure dp_register_array_4D_R
		module procedure dp_register_array_4D_L


#if 0
		module procedure dp_register_array_1D_C16
		module procedure dp_register_array_1D_C32
		module procedure dp_register_array_1D_I1
		module procedure dp_register_array_1D_I2
		module procedure dp_register_array_1D_R16
		module procedure dp_register_array_1D_L1
		module procedure dp_register_array_1D_L2
		module procedure dp_register_array_2D_C16
		module procedure dp_register_array_2D_C32
		module procedure dp_register_array_2D_I1
		module procedure dp_register_array_2D_I2
		module procedure dp_register_array_2D_R16
		module procedure dp_register_array_2D_L1
		module procedure dp_register_array_2D_L2
		module procedure dp_register_array_3D_C16
		module procedure dp_register_array_3D_C32
		module procedure dp_register_array_3D_I1
		module procedure dp_register_array_3D_I2
		module procedure dp_register_array_3D_R16
		module procedure dp_register_array_3D_L1
		module procedure dp_register_array_3D_L2
		module procedure dp_register_array_4D_C16
		module procedure dp_register_array_4D_C32
		module procedure dp_register_array_4D_I1
		module procedure dp_register_array_4D_I2
		module procedure dp_register_array_4D_R16
		module procedure dp_register_array_4D_L1
		module procedure dp_register_array_4D_L2
#endif

	end interface


	



end module DPARLIB

















	module dp_internal

	implicit none

!IBM* SOURCEFORM (FIXED)
        include 'mpif.h'
!IBM* SOURCEFORM (FREE(F90))




!	Internal data items:

	integer, parameter :: DP_MAX_DIMS = 4
	integer, parameter :: DP_MAX_MV_BLOCKS = 16  ! 2**DP_MAX_DIMS

	logical :: dp_initialized = .FALSE.
	logical :: dp_mpi_initialized = .FALSE.
	logical :: dp_serial_run = .FALSE.

	integer :: dp_comm_cart = MPI_COMM_NULL
	integer :: dp_num_procs, dp_rank

	integer dp_coords(DP_MAX_DIMS)
	integer dp_proc_arr_size(DP_MAX_DIMS)

        integer dp_cshift_ierr
        integer dp_eoshift_ierr

        integer, parameter :: dp_maxproc = 64   ! maximum number of processors
        

!!! START RANDOM NUMBER GENERATOR
!   Stuff for random number generator; this will work for a machine 
!   size up to 64 (one can always hope...)
        integer dp_seeds (0: dp_maxproc)

!      Load data array in case user forgets to initialize.
!      This array is the result of calling UNI 100000 times
!         with ISEED=305 and K=64.
       real dp_u_array (1:17)  
       real*4, parameter  ::  dp_uni_denom = 16777216.0
       real*4, parameter  ::  dp_uni_csave = 362436.0/dp_uni_denom
       real*4, parameter  ::  dp_uni_cd    = 7654321.0/dp_uni_denom
       real*4, parameter  ::  dp_uni_cm    = 16777213.0/dp_uni_denom

      real*4  dp_uni_c 
      integer dp_uni_i 
      integer dp_uni_j
      

!!! END RANDOM NUMBER GENERATOR


!!! Some parameters used in the IO routines

	integer, parameter :: DP_WRITE_NEXT_ARR_SEGMENT = 1,  &
		      		DP_REQUEST_ARR_COL = 2,         &
		      		DP_ARR_COL = 3

	integer, parameter :: DP_DOUBLE_PRECISION_SIZE = 8

!!! END of stuff for IO routines




!!! Some error conditions

	integer, parameter :: DP_SUCCESS = 0
	integer, parameter :: DP_NOT_INITIALIZED = -1
	integer, parameter :: DP_INCONSISTENT_ARRAY_SIZE = -2
	integer, parameter :: DP_BAD_NUM_PROCS = -3
	integer, parameter :: DP_BAD_PROC_NDIMS = -4
	integer, parameter :: DP_BAD_CALL_SEQUENCE = -4


!!! 

	end module dp_internal


