

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 1D REAL global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 1D REAL global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_1D_R
!!
SUBROUTINE dp_register_array_1D_R &
                                           (arr, global_size, ierr)
REAL, target, intent(IN) :: arr (:)
integer, intent(IN) :: global_size(1)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_1D_R), &
           pointer :: arr_desc 




print *, "dp_register_array_1D_R A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_1D_R B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_1D_R C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_1D_R D"

arr_desc%arr => arr


print *, "dp_register_array_1D_R E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 1 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_1D_R F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:1) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:1) = ubound(arr)
arr_desc%distribution%global_size(1:1) = global_size


print *, "dp_register_array_1D_R G"


do i = 1, 1


print *, "dp_register_array_1D_R H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_1D_R I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_1D_R J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_1D_R K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 1
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_1D_R L"


!! Add this node to the appropriate linked list
call add_arr_desc_1D_R (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_1D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_1D_R (arr_desc) 
type (array_descriptor_1D_R), &
                                    pointer :: arr_desc

if (associated ( Arr_List_1D_R )) then
	arr_desc%next => Arr_List_1D_R
else
	NULLIFY (arr_desc%next)
endif

Arr_List_1D_R => arr_desc

!! call dp_print_array_desc_1D_R (arr_desc)

end SUBROUTINE add_arr_desc_1D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_1D_R 
!!
SUBROUTINE dp_find_arr_desc_1D_R (arr, arr_desc, ierr)
REAL, target, intent(IN) :: arr (:)
type (array_descriptor_1D_R), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_1D_R

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_1D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_1D_R
!!
SUBROUTINE dp_get_arr_dist_1D_R &
                                                (arr, dist, ierr)
REAL, target, intent(IN) :: arr (:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_1D_R), pointer :: arr_desc

call dp_find_arr_desc_1D_R (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_1D_R







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_1D_R
!!
subroutine dp_alloc_array_A_1D_R &
	(arr_ptr, global_size, local_size, ierr)

REAL, POINTER :: arr_ptr (:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (1)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (1)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 1)
integer :: local_bounds (2, 1)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_1D_R &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_1D_R 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_1D_R
!!
subroutine dp_alloc_array_B_1D_R &
	(arr_ptr, global_bounds, local_bounds, ierr)

REAL, POINTER :: arr_ptr (:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 1)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 1)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(1)
integer :: b (2, 1)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 1

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_1D_R (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_1D_R 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_1D_R 
!!
subroutine dp_print_array_desc_1D_R (arr_desc)
type (array_descriptor_1D_R) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 1D R"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_1D_R


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 2D REAL global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 2D REAL global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_2D_R
!!
SUBROUTINE dp_register_array_2D_R &
                                           (arr, global_size, ierr)
REAL, target, intent(IN) :: arr (:,:)
integer, intent(IN) :: global_size(2)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_2D_R), &
           pointer :: arr_desc 




print *, "dp_register_array_2D_R A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_2D_R B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_2D_R C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_2D_R D"

arr_desc%arr => arr


print *, "dp_register_array_2D_R E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 2 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_2D_R F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:2) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:2) = ubound(arr)
arr_desc%distribution%global_size(1:2) = global_size


print *, "dp_register_array_2D_R G"


do i = 1, 2


print *, "dp_register_array_2D_R H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_2D_R I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_2D_R J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_2D_R K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 2
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_2D_R L"


!! Add this node to the appropriate linked list
call add_arr_desc_2D_R (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_2D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_2D_R (arr_desc) 
type (array_descriptor_2D_R), &
                                    pointer :: arr_desc

if (associated ( Arr_List_2D_R )) then
	arr_desc%next => Arr_List_2D_R
else
	NULLIFY (arr_desc%next)
endif

Arr_List_2D_R => arr_desc

!! call dp_print_array_desc_2D_R (arr_desc)

end SUBROUTINE add_arr_desc_2D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_2D_R 
!!
SUBROUTINE dp_find_arr_desc_2D_R (arr, arr_desc, ierr)
REAL, target, intent(IN) :: arr (:,:)
type (array_descriptor_2D_R), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_2D_R

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_2D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_2D_R
!!
SUBROUTINE dp_get_arr_dist_2D_R &
                                                (arr, dist, ierr)
REAL, target, intent(IN) :: arr (:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_2D_R), pointer :: arr_desc

call dp_find_arr_desc_2D_R (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_2D_R







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_2D_R
!!
subroutine dp_alloc_array_A_2D_R &
	(arr_ptr, global_size, local_size, ierr)

REAL, POINTER :: arr_ptr (:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (2)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (2)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 2)
integer :: local_bounds (2, 2)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_2D_R &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_2D_R 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_2D_R
!!
subroutine dp_alloc_array_B_2D_R &
	(arr_ptr, global_bounds, local_bounds, ierr)

REAL, POINTER :: arr_ptr (:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 2)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 2)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(2)
integer :: b (2, 2)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 2

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_2D_R (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_2D_R 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_2D_R 
!!
subroutine dp_print_array_desc_2D_R (arr_desc)
type (array_descriptor_2D_R) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 2D R"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_2D_R


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 3D REAL global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 3D REAL global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_3D_R
!!
SUBROUTINE dp_register_array_3D_R &
                                           (arr, global_size, ierr)
REAL, target, intent(IN) :: arr (:,:,:)
integer, intent(IN) :: global_size(3)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_3D_R), &
           pointer :: arr_desc 




print *, "dp_register_array_3D_R A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_3D_R B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_3D_R C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_3D_R D"

arr_desc%arr => arr


print *, "dp_register_array_3D_R E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 3 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_3D_R F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:3) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:3) = ubound(arr)
arr_desc%distribution%global_size(1:3) = global_size


print *, "dp_register_array_3D_R G"


do i = 1, 3


print *, "dp_register_array_3D_R H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_3D_R I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_3D_R J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_3D_R K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 3
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_3D_R L"


!! Add this node to the appropriate linked list
call add_arr_desc_3D_R (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_3D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_3D_R (arr_desc) 
type (array_descriptor_3D_R), &
                                    pointer :: arr_desc

if (associated ( Arr_List_3D_R )) then
	arr_desc%next => Arr_List_3D_R
else
	NULLIFY (arr_desc%next)
endif

Arr_List_3D_R => arr_desc

!! call dp_print_array_desc_3D_R (arr_desc)

end SUBROUTINE add_arr_desc_3D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_3D_R 
!!
SUBROUTINE dp_find_arr_desc_3D_R (arr, arr_desc, ierr)
REAL, target, intent(IN) :: arr (:,:,:)
type (array_descriptor_3D_R), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_3D_R

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_3D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_3D_R
!!
SUBROUTINE dp_get_arr_dist_3D_R &
                                                (arr, dist, ierr)
REAL, target, intent(IN) :: arr (:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_3D_R), pointer :: arr_desc

call dp_find_arr_desc_3D_R (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_3D_R







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_3D_R
!!
subroutine dp_alloc_array_A_3D_R &
	(arr_ptr, global_size, local_size, ierr)

REAL, POINTER :: arr_ptr (:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (3)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (3)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 3)
integer :: local_bounds (2, 3)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_3D_R &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_3D_R 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_3D_R
!!
subroutine dp_alloc_array_B_3D_R &
	(arr_ptr, global_bounds, local_bounds, ierr)

REAL, POINTER :: arr_ptr (:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 3)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 3)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(3)
integer :: b (2, 3)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 3

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_3D_R (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_3D_R 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_3D_R 
!!
subroutine dp_print_array_desc_3D_R (arr_desc)
type (array_descriptor_3D_R) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 3D R"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_3D_R


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 4D REAL global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 4D REAL global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_4D_R
!!
SUBROUTINE dp_register_array_4D_R &
                                           (arr, global_size, ierr)
REAL, target, intent(IN) :: arr (:,:,:,:)
integer, intent(IN) :: global_size(4)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_4D_R), &
           pointer :: arr_desc 




print *, "dp_register_array_4D_R A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_4D_R B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_4D_R C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_4D_R D"

arr_desc%arr => arr


print *, "dp_register_array_4D_R E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 4 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_4D_R F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:4) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:4) = ubound(arr)
arr_desc%distribution%global_size(1:4) = global_size


print *, "dp_register_array_4D_R G"


do i = 1, 4


print *, "dp_register_array_4D_R H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_4D_R I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_4D_R J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_4D_R K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 4
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_4D_R L"


!! Add this node to the appropriate linked list
call add_arr_desc_4D_R (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_4D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_4D_R (arr_desc) 
type (array_descriptor_4D_R), &
                                    pointer :: arr_desc

if (associated ( Arr_List_4D_R )) then
	arr_desc%next => Arr_List_4D_R
else
	NULLIFY (arr_desc%next)
endif

Arr_List_4D_R => arr_desc

!! call dp_print_array_desc_4D_R (arr_desc)

end SUBROUTINE add_arr_desc_4D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_4D_R 
!!
SUBROUTINE dp_find_arr_desc_4D_R (arr, arr_desc, ierr)
REAL, target, intent(IN) :: arr (:,:,:,:)
type (array_descriptor_4D_R), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_4D_R

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_4D_R 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_4D_R
!!
SUBROUTINE dp_get_arr_dist_4D_R &
                                                (arr, dist, ierr)
REAL, target, intent(IN) :: arr (:,:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_4D_R), pointer :: arr_desc

call dp_find_arr_desc_4D_R (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_4D_R







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_4D_R
!!
subroutine dp_alloc_array_A_4D_R &
	(arr_ptr, global_size, local_size, ierr)

REAL, POINTER :: arr_ptr (:,:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (4)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (4)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 4)
integer :: local_bounds (2, 4)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_4D_R &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_4D_R 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_4D_R
!!
subroutine dp_alloc_array_B_4D_R &
	(arr_ptr, global_bounds, local_bounds, ierr)

REAL, POINTER :: arr_ptr (:,:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 4)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 4)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(4)
integer :: b (2, 4)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 4

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3), b(1,4):b(2,4)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_4D_R (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_4D_R 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_4D_R 
!!
subroutine dp_print_array_desc_4D_R (arr_desc)
type (array_descriptor_4D_R) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 4D R"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_4D_R


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 1D DOUBLE PRECISION global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 1D DOUBLE PRECISION global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_1D_D
!!
SUBROUTINE dp_register_array_1D_D &
                                           (arr, global_size, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:)
integer, intent(IN) :: global_size(1)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_1D_D), &
           pointer :: arr_desc 




print *, "dp_register_array_1D_D A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_1D_D B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_1D_D C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_1D_D D"

arr_desc%arr => arr


print *, "dp_register_array_1D_D E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 1 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_1D_D F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:1) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:1) = ubound(arr)
arr_desc%distribution%global_size(1:1) = global_size


print *, "dp_register_array_1D_D G"


do i = 1, 1


print *, "dp_register_array_1D_D H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_1D_D I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_1D_D J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_1D_D K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 1
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_1D_D L"


!! Add this node to the appropriate linked list
call add_arr_desc_1D_D (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_1D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_1D_D (arr_desc) 
type (array_descriptor_1D_D), &
                                    pointer :: arr_desc

if (associated ( Arr_List_1D_D )) then
	arr_desc%next => Arr_List_1D_D
else
	NULLIFY (arr_desc%next)
endif

Arr_List_1D_D => arr_desc

!! call dp_print_array_desc_1D_D (arr_desc)

end SUBROUTINE add_arr_desc_1D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_1D_D 
!!
SUBROUTINE dp_find_arr_desc_1D_D (arr, arr_desc, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:)
type (array_descriptor_1D_D), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_1D_D

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_1D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_1D_D
!!
SUBROUTINE dp_get_arr_dist_1D_D &
                                                (arr, dist, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_1D_D), pointer :: arr_desc

call dp_find_arr_desc_1D_D (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_1D_D







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_1D_D
!!
subroutine dp_alloc_array_A_1D_D &
	(arr_ptr, global_size, local_size, ierr)

DOUBLE PRECISION, POINTER :: arr_ptr (:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (1)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (1)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 1)
integer :: local_bounds (2, 1)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_1D_D &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_1D_D 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_1D_D
!!
subroutine dp_alloc_array_B_1D_D &
	(arr_ptr, global_bounds, local_bounds, ierr)

DOUBLE PRECISION, POINTER :: arr_ptr (:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 1)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 1)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(1)
integer :: b (2, 1)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 1

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_1D_D (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_1D_D 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_1D_D 
!!
subroutine dp_print_array_desc_1D_D (arr_desc)
type (array_descriptor_1D_D) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 1D D"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_1D_D


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 2D DOUBLE PRECISION global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 2D DOUBLE PRECISION global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_2D_D
!!
SUBROUTINE dp_register_array_2D_D &
                                           (arr, global_size, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:,:)
integer, intent(IN) :: global_size(2)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_2D_D), &
           pointer :: arr_desc 




print *, "dp_register_array_2D_D A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_2D_D B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_2D_D C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_2D_D D"

arr_desc%arr => arr


print *, "dp_register_array_2D_D E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 2 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_2D_D F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:2) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:2) = ubound(arr)
arr_desc%distribution%global_size(1:2) = global_size


print *, "dp_register_array_2D_D G"


do i = 1, 2


print *, "dp_register_array_2D_D H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_2D_D I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_2D_D J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_2D_D K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 2
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_2D_D L"


!! Add this node to the appropriate linked list
call add_arr_desc_2D_D (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_2D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_2D_D (arr_desc) 
type (array_descriptor_2D_D), &
                                    pointer :: arr_desc

if (associated ( Arr_List_2D_D )) then
	arr_desc%next => Arr_List_2D_D
else
	NULLIFY (arr_desc%next)
endif

Arr_List_2D_D => arr_desc

!! call dp_print_array_desc_2D_D (arr_desc)

end SUBROUTINE add_arr_desc_2D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_2D_D 
!!
SUBROUTINE dp_find_arr_desc_2D_D (arr, arr_desc, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:,:)
type (array_descriptor_2D_D), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_2D_D

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_2D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_2D_D
!!
SUBROUTINE dp_get_arr_dist_2D_D &
                                                (arr, dist, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_2D_D), pointer :: arr_desc

call dp_find_arr_desc_2D_D (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_2D_D







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_2D_D
!!
subroutine dp_alloc_array_A_2D_D &
	(arr_ptr, global_size, local_size, ierr)

DOUBLE PRECISION, POINTER :: arr_ptr (:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (2)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (2)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 2)
integer :: local_bounds (2, 2)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_2D_D &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_2D_D 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_2D_D
!!
subroutine dp_alloc_array_B_2D_D &
	(arr_ptr, global_bounds, local_bounds, ierr)

DOUBLE PRECISION, POINTER :: arr_ptr (:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 2)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 2)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(2)
integer :: b (2, 2)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 2

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_2D_D (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_2D_D 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_2D_D 
!!
subroutine dp_print_array_desc_2D_D (arr_desc)
type (array_descriptor_2D_D) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 2D D"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_2D_D


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 3D DOUBLE PRECISION global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 3D DOUBLE PRECISION global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_3D_D
!!
SUBROUTINE dp_register_array_3D_D &
                                           (arr, global_size, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:,:,:)
integer, intent(IN) :: global_size(3)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_3D_D), &
           pointer :: arr_desc 




print *, "dp_register_array_3D_D A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_3D_D B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_3D_D C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_3D_D D"

arr_desc%arr => arr


print *, "dp_register_array_3D_D E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 3 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_3D_D F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:3) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:3) = ubound(arr)
arr_desc%distribution%global_size(1:3) = global_size


print *, "dp_register_array_3D_D G"


do i = 1, 3


print *, "dp_register_array_3D_D H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_3D_D I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_3D_D J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_3D_D K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 3
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_3D_D L"


!! Add this node to the appropriate linked list
call add_arr_desc_3D_D (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_3D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_3D_D (arr_desc) 
type (array_descriptor_3D_D), &
                                    pointer :: arr_desc

if (associated ( Arr_List_3D_D )) then
	arr_desc%next => Arr_List_3D_D
else
	NULLIFY (arr_desc%next)
endif

Arr_List_3D_D => arr_desc

!! call dp_print_array_desc_3D_D (arr_desc)

end SUBROUTINE add_arr_desc_3D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_3D_D 
!!
SUBROUTINE dp_find_arr_desc_3D_D (arr, arr_desc, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:,:,:)
type (array_descriptor_3D_D), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_3D_D

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_3D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_3D_D
!!
SUBROUTINE dp_get_arr_dist_3D_D &
                                                (arr, dist, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_3D_D), pointer :: arr_desc

call dp_find_arr_desc_3D_D (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_3D_D







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_3D_D
!!
subroutine dp_alloc_array_A_3D_D &
	(arr_ptr, global_size, local_size, ierr)

DOUBLE PRECISION, POINTER :: arr_ptr (:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (3)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (3)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 3)
integer :: local_bounds (2, 3)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_3D_D &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_3D_D 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_3D_D
!!
subroutine dp_alloc_array_B_3D_D &
	(arr_ptr, global_bounds, local_bounds, ierr)

DOUBLE PRECISION, POINTER :: arr_ptr (:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 3)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 3)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(3)
integer :: b (2, 3)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 3

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_3D_D (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_3D_D 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_3D_D 
!!
subroutine dp_print_array_desc_3D_D (arr_desc)
type (array_descriptor_3D_D) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 3D D"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_3D_D


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 4D DOUBLE PRECISION global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 4D DOUBLE PRECISION global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_4D_D
!!
SUBROUTINE dp_register_array_4D_D &
                                           (arr, global_size, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:,:,:,:)
integer, intent(IN) :: global_size(4)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_4D_D), &
           pointer :: arr_desc 




print *, "dp_register_array_4D_D A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_4D_D B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_4D_D C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_4D_D D"

arr_desc%arr => arr


print *, "dp_register_array_4D_D E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 4 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_4D_D F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:4) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:4) = ubound(arr)
arr_desc%distribution%global_size(1:4) = global_size


print *, "dp_register_array_4D_D G"


do i = 1, 4


print *, "dp_register_array_4D_D H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_4D_D I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_4D_D J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_4D_D K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 4
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_4D_D L"


!! Add this node to the appropriate linked list
call add_arr_desc_4D_D (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_4D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_4D_D (arr_desc) 
type (array_descriptor_4D_D), &
                                    pointer :: arr_desc

if (associated ( Arr_List_4D_D )) then
	arr_desc%next => Arr_List_4D_D
else
	NULLIFY (arr_desc%next)
endif

Arr_List_4D_D => arr_desc

!! call dp_print_array_desc_4D_D (arr_desc)

end SUBROUTINE add_arr_desc_4D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_4D_D 
!!
SUBROUTINE dp_find_arr_desc_4D_D (arr, arr_desc, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:,:,:,:)
type (array_descriptor_4D_D), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_4D_D

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_4D_D 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_4D_D
!!
SUBROUTINE dp_get_arr_dist_4D_D &
                                                (arr, dist, ierr)
DOUBLE PRECISION, target, intent(IN) :: arr (:,:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_4D_D), pointer :: arr_desc

call dp_find_arr_desc_4D_D (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_4D_D







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_4D_D
!!
subroutine dp_alloc_array_A_4D_D &
	(arr_ptr, global_size, local_size, ierr)

DOUBLE PRECISION, POINTER :: arr_ptr (:,:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (4)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (4)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 4)
integer :: local_bounds (2, 4)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_4D_D &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_4D_D 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_4D_D
!!
subroutine dp_alloc_array_B_4D_D &
	(arr_ptr, global_bounds, local_bounds, ierr)

DOUBLE PRECISION, POINTER :: arr_ptr (:,:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 4)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 4)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(4)
integer :: b (2, 4)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 4

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3), b(1,4):b(2,4)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_4D_D (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_4D_D 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_4D_D 
!!
subroutine dp_print_array_desc_4D_D (arr_desc)
type (array_descriptor_4D_D) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 4D D"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_4D_D


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 1D COMPLEX global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 1D COMPLEX global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_1D_C
!!
SUBROUTINE dp_register_array_1D_C &
                                           (arr, global_size, ierr)
COMPLEX, target, intent(IN) :: arr (:)
integer, intent(IN) :: global_size(1)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_1D_C), &
           pointer :: arr_desc 




print *, "dp_register_array_1D_C A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_1D_C B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_1D_C C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_1D_C D"

arr_desc%arr => arr


print *, "dp_register_array_1D_C E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 1 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_1D_C F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:1) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:1) = ubound(arr)
arr_desc%distribution%global_size(1:1) = global_size


print *, "dp_register_array_1D_C G"


do i = 1, 1


print *, "dp_register_array_1D_C H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_1D_C I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_1D_C J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_1D_C K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 1
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_1D_C L"


!! Add this node to the appropriate linked list
call add_arr_desc_1D_C (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_1D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_1D_C (arr_desc) 
type (array_descriptor_1D_C), &
                                    pointer :: arr_desc

if (associated ( Arr_List_1D_C )) then
	arr_desc%next => Arr_List_1D_C
else
	NULLIFY (arr_desc%next)
endif

Arr_List_1D_C => arr_desc

!! call dp_print_array_desc_1D_C (arr_desc)

end SUBROUTINE add_arr_desc_1D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_1D_C 
!!
SUBROUTINE dp_find_arr_desc_1D_C (arr, arr_desc, ierr)
COMPLEX, target, intent(IN) :: arr (:)
type (array_descriptor_1D_C), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_1D_C

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_1D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_1D_C
!!
SUBROUTINE dp_get_arr_dist_1D_C &
                                                (arr, dist, ierr)
COMPLEX, target, intent(IN) :: arr (:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_1D_C), pointer :: arr_desc

call dp_find_arr_desc_1D_C (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_1D_C







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_1D_C
!!
subroutine dp_alloc_array_A_1D_C &
	(arr_ptr, global_size, local_size, ierr)

COMPLEX, POINTER :: arr_ptr (:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (1)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (1)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 1)
integer :: local_bounds (2, 1)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_1D_C &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_1D_C 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_1D_C
!!
subroutine dp_alloc_array_B_1D_C &
	(arr_ptr, global_bounds, local_bounds, ierr)

COMPLEX, POINTER :: arr_ptr (:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 1)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 1)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(1)
integer :: b (2, 1)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 1

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_1D_C (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_1D_C 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_1D_C 
!!
subroutine dp_print_array_desc_1D_C (arr_desc)
type (array_descriptor_1D_C) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 1D C"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_1D_C


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 2D COMPLEX global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 2D COMPLEX global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_2D_C
!!
SUBROUTINE dp_register_array_2D_C &
                                           (arr, global_size, ierr)
COMPLEX, target, intent(IN) :: arr (:,:)
integer, intent(IN) :: global_size(2)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_2D_C), &
           pointer :: arr_desc 




print *, "dp_register_array_2D_C A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_2D_C B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_2D_C C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_2D_C D"

arr_desc%arr => arr


print *, "dp_register_array_2D_C E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 2 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_2D_C F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:2) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:2) = ubound(arr)
arr_desc%distribution%global_size(1:2) = global_size


print *, "dp_register_array_2D_C G"


do i = 1, 2


print *, "dp_register_array_2D_C H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_2D_C I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_2D_C J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_2D_C K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 2
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_2D_C L"


!! Add this node to the appropriate linked list
call add_arr_desc_2D_C (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_2D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_2D_C (arr_desc) 
type (array_descriptor_2D_C), &
                                    pointer :: arr_desc

if (associated ( Arr_List_2D_C )) then
	arr_desc%next => Arr_List_2D_C
else
	NULLIFY (arr_desc%next)
endif

Arr_List_2D_C => arr_desc

!! call dp_print_array_desc_2D_C (arr_desc)

end SUBROUTINE add_arr_desc_2D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_2D_C 
!!
SUBROUTINE dp_find_arr_desc_2D_C (arr, arr_desc, ierr)
COMPLEX, target, intent(IN) :: arr (:,:)
type (array_descriptor_2D_C), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_2D_C

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_2D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_2D_C
!!
SUBROUTINE dp_get_arr_dist_2D_C &
                                                (arr, dist, ierr)
COMPLEX, target, intent(IN) :: arr (:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_2D_C), pointer :: arr_desc

call dp_find_arr_desc_2D_C (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_2D_C







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_2D_C
!!
subroutine dp_alloc_array_A_2D_C &
	(arr_ptr, global_size, local_size, ierr)

COMPLEX, POINTER :: arr_ptr (:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (2)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (2)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 2)
integer :: local_bounds (2, 2)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_2D_C &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_2D_C 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_2D_C
!!
subroutine dp_alloc_array_B_2D_C &
	(arr_ptr, global_bounds, local_bounds, ierr)

COMPLEX, POINTER :: arr_ptr (:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 2)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 2)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(2)
integer :: b (2, 2)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 2

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_2D_C (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_2D_C 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_2D_C 
!!
subroutine dp_print_array_desc_2D_C (arr_desc)
type (array_descriptor_2D_C) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 2D C"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_2D_C


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 3D COMPLEX global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 3D COMPLEX global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_3D_C
!!
SUBROUTINE dp_register_array_3D_C &
                                           (arr, global_size, ierr)
COMPLEX, target, intent(IN) :: arr (:,:,:)
integer, intent(IN) :: global_size(3)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_3D_C), &
           pointer :: arr_desc 




print *, "dp_register_array_3D_C A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_3D_C B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_3D_C C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_3D_C D"

arr_desc%arr => arr


print *, "dp_register_array_3D_C E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 3 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_3D_C F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:3) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:3) = ubound(arr)
arr_desc%distribution%global_size(1:3) = global_size


print *, "dp_register_array_3D_C G"


do i = 1, 3


print *, "dp_register_array_3D_C H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_3D_C I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_3D_C J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_3D_C K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 3
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_3D_C L"


!! Add this node to the appropriate linked list
call add_arr_desc_3D_C (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_3D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_3D_C (arr_desc) 
type (array_descriptor_3D_C), &
                                    pointer :: arr_desc

if (associated ( Arr_List_3D_C )) then
	arr_desc%next => Arr_List_3D_C
else
	NULLIFY (arr_desc%next)
endif

Arr_List_3D_C => arr_desc

!! call dp_print_array_desc_3D_C (arr_desc)

end SUBROUTINE add_arr_desc_3D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_3D_C 
!!
SUBROUTINE dp_find_arr_desc_3D_C (arr, arr_desc, ierr)
COMPLEX, target, intent(IN) :: arr (:,:,:)
type (array_descriptor_3D_C), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_3D_C

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_3D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_3D_C
!!
SUBROUTINE dp_get_arr_dist_3D_C &
                                                (arr, dist, ierr)
COMPLEX, target, intent(IN) :: arr (:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_3D_C), pointer :: arr_desc

call dp_find_arr_desc_3D_C (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_3D_C







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_3D_C
!!
subroutine dp_alloc_array_A_3D_C &
	(arr_ptr, global_size, local_size, ierr)

COMPLEX, POINTER :: arr_ptr (:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (3)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (3)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 3)
integer :: local_bounds (2, 3)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_3D_C &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_3D_C 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_3D_C
!!
subroutine dp_alloc_array_B_3D_C &
	(arr_ptr, global_bounds, local_bounds, ierr)

COMPLEX, POINTER :: arr_ptr (:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 3)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 3)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(3)
integer :: b (2, 3)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 3

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_3D_C (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_3D_C 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_3D_C 
!!
subroutine dp_print_array_desc_3D_C (arr_desc)
type (array_descriptor_3D_C) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 3D C"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_3D_C


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 4D COMPLEX global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 4D COMPLEX global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_4D_C
!!
SUBROUTINE dp_register_array_4D_C &
                                           (arr, global_size, ierr)
COMPLEX, target, intent(IN) :: arr (:,:,:,:)
integer, intent(IN) :: global_size(4)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_4D_C), &
           pointer :: arr_desc 




print *, "dp_register_array_4D_C A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_4D_C B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_4D_C C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_4D_C D"

arr_desc%arr => arr


print *, "dp_register_array_4D_C E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 4 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_4D_C F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:4) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:4) = ubound(arr)
arr_desc%distribution%global_size(1:4) = global_size


print *, "dp_register_array_4D_C G"


do i = 1, 4


print *, "dp_register_array_4D_C H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_4D_C I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_4D_C J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_4D_C K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 4
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_4D_C L"


!! Add this node to the appropriate linked list
call add_arr_desc_4D_C (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_4D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_4D_C (arr_desc) 
type (array_descriptor_4D_C), &
                                    pointer :: arr_desc

if (associated ( Arr_List_4D_C )) then
	arr_desc%next => Arr_List_4D_C
else
	NULLIFY (arr_desc%next)
endif

Arr_List_4D_C => arr_desc

!! call dp_print_array_desc_4D_C (arr_desc)

end SUBROUTINE add_arr_desc_4D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_4D_C 
!!
SUBROUTINE dp_find_arr_desc_4D_C (arr, arr_desc, ierr)
COMPLEX, target, intent(IN) :: arr (:,:,:,:)
type (array_descriptor_4D_C), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_4D_C

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_4D_C 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_4D_C
!!
SUBROUTINE dp_get_arr_dist_4D_C &
                                                (arr, dist, ierr)
COMPLEX, target, intent(IN) :: arr (:,:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_4D_C), pointer :: arr_desc

call dp_find_arr_desc_4D_C (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_4D_C







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_4D_C
!!
subroutine dp_alloc_array_A_4D_C &
	(arr_ptr, global_size, local_size, ierr)

COMPLEX, POINTER :: arr_ptr (:,:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (4)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (4)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 4)
integer :: local_bounds (2, 4)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_4D_C &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_4D_C 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_4D_C
!!
subroutine dp_alloc_array_B_4D_C &
	(arr_ptr, global_bounds, local_bounds, ierr)

COMPLEX, POINTER :: arr_ptr (:,:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 4)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 4)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(4)
integer :: b (2, 4)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 4

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3), b(1,4):b(2,4)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_4D_C (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_4D_C 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_4D_C 
!!
subroutine dp_print_array_desc_4D_C (arr_desc)
type (array_descriptor_4D_C) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 4D C"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_4D_C


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 1D INTEGER global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 1D INTEGER global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_1D_I
!!
SUBROUTINE dp_register_array_1D_I &
                                           (arr, global_size, ierr)
INTEGER, target, intent(IN) :: arr (:)
integer, intent(IN) :: global_size(1)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_1D_I), &
           pointer :: arr_desc 




print *, "dp_register_array_1D_I A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_1D_I B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_1D_I C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_1D_I D"

arr_desc%arr => arr


print *, "dp_register_array_1D_I E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 1 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_1D_I F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:1) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:1) = ubound(arr)
arr_desc%distribution%global_size(1:1) = global_size


print *, "dp_register_array_1D_I G"


do i = 1, 1


print *, "dp_register_array_1D_I H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_1D_I I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_1D_I J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_1D_I K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 1
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_1D_I L"


!! Add this node to the appropriate linked list
call add_arr_desc_1D_I (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_1D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_1D_I (arr_desc) 
type (array_descriptor_1D_I), &
                                    pointer :: arr_desc

if (associated ( Arr_List_1D_I )) then
	arr_desc%next => Arr_List_1D_I
else
	NULLIFY (arr_desc%next)
endif

Arr_List_1D_I => arr_desc

!! call dp_print_array_desc_1D_I (arr_desc)

end SUBROUTINE add_arr_desc_1D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_1D_I 
!!
SUBROUTINE dp_find_arr_desc_1D_I (arr, arr_desc, ierr)
INTEGER, target, intent(IN) :: arr (:)
type (array_descriptor_1D_I), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_1D_I

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_1D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_1D_I
!!
SUBROUTINE dp_get_arr_dist_1D_I &
                                                (arr, dist, ierr)
INTEGER, target, intent(IN) :: arr (:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_1D_I), pointer :: arr_desc

call dp_find_arr_desc_1D_I (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_1D_I







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_1D_I
!!
subroutine dp_alloc_array_A_1D_I &
	(arr_ptr, global_size, local_size, ierr)

INTEGER, POINTER :: arr_ptr (:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (1)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (1)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 1)
integer :: local_bounds (2, 1)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_1D_I &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_1D_I 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_1D_I
!!
subroutine dp_alloc_array_B_1D_I &
	(arr_ptr, global_bounds, local_bounds, ierr)

INTEGER, POINTER :: arr_ptr (:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 1)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 1)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(1)
integer :: b (2, 1)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 1

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_1D_I (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_1D_I 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_1D_I 
!!
subroutine dp_print_array_desc_1D_I (arr_desc)
type (array_descriptor_1D_I) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 1D I"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_1D_I


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 2D INTEGER global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 2D INTEGER global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_2D_I
!!
SUBROUTINE dp_register_array_2D_I &
                                           (arr, global_size, ierr)
INTEGER, target, intent(IN) :: arr (:,:)
integer, intent(IN) :: global_size(2)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_2D_I), &
           pointer :: arr_desc 




print *, "dp_register_array_2D_I A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_2D_I B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_2D_I C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_2D_I D"

arr_desc%arr => arr


print *, "dp_register_array_2D_I E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 2 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_2D_I F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:2) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:2) = ubound(arr)
arr_desc%distribution%global_size(1:2) = global_size


print *, "dp_register_array_2D_I G"


do i = 1, 2


print *, "dp_register_array_2D_I H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_2D_I I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_2D_I J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_2D_I K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 2
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_2D_I L"


!! Add this node to the appropriate linked list
call add_arr_desc_2D_I (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_2D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_2D_I (arr_desc) 
type (array_descriptor_2D_I), &
                                    pointer :: arr_desc

if (associated ( Arr_List_2D_I )) then
	arr_desc%next => Arr_List_2D_I
else
	NULLIFY (arr_desc%next)
endif

Arr_List_2D_I => arr_desc

!! call dp_print_array_desc_2D_I (arr_desc)

end SUBROUTINE add_arr_desc_2D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_2D_I 
!!
SUBROUTINE dp_find_arr_desc_2D_I (arr, arr_desc, ierr)
INTEGER, target, intent(IN) :: arr (:,:)
type (array_descriptor_2D_I), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_2D_I

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_2D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_2D_I
!!
SUBROUTINE dp_get_arr_dist_2D_I &
                                                (arr, dist, ierr)
INTEGER, target, intent(IN) :: arr (:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_2D_I), pointer :: arr_desc

call dp_find_arr_desc_2D_I (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_2D_I







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_2D_I
!!
subroutine dp_alloc_array_A_2D_I &
	(arr_ptr, global_size, local_size, ierr)

INTEGER, POINTER :: arr_ptr (:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (2)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (2)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 2)
integer :: local_bounds (2, 2)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_2D_I &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_2D_I 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_2D_I
!!
subroutine dp_alloc_array_B_2D_I &
	(arr_ptr, global_bounds, local_bounds, ierr)

INTEGER, POINTER :: arr_ptr (:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 2)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 2)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(2)
integer :: b (2, 2)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 2

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_2D_I (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_2D_I 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_2D_I 
!!
subroutine dp_print_array_desc_2D_I (arr_desc)
type (array_descriptor_2D_I) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 2D I"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_2D_I


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 3D INTEGER global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 3D INTEGER global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_3D_I
!!
SUBROUTINE dp_register_array_3D_I &
                                           (arr, global_size, ierr)
INTEGER, target, intent(IN) :: arr (:,:,:)
integer, intent(IN) :: global_size(3)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_3D_I), &
           pointer :: arr_desc 




print *, "dp_register_array_3D_I A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_3D_I B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_3D_I C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_3D_I D"

arr_desc%arr => arr


print *, "dp_register_array_3D_I E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 3 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_3D_I F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:3) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:3) = ubound(arr)
arr_desc%distribution%global_size(1:3) = global_size


print *, "dp_register_array_3D_I G"


do i = 1, 3


print *, "dp_register_array_3D_I H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_3D_I I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_3D_I J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_3D_I K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 3
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_3D_I L"


!! Add this node to the appropriate linked list
call add_arr_desc_3D_I (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_3D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_3D_I (arr_desc) 
type (array_descriptor_3D_I), &
                                    pointer :: arr_desc

if (associated ( Arr_List_3D_I )) then
	arr_desc%next => Arr_List_3D_I
else
	NULLIFY (arr_desc%next)
endif

Arr_List_3D_I => arr_desc

!! call dp_print_array_desc_3D_I (arr_desc)

end SUBROUTINE add_arr_desc_3D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_3D_I 
!!
SUBROUTINE dp_find_arr_desc_3D_I (arr, arr_desc, ierr)
INTEGER, target, intent(IN) :: arr (:,:,:)
type (array_descriptor_3D_I), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_3D_I

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_3D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_3D_I
!!
SUBROUTINE dp_get_arr_dist_3D_I &
                                                (arr, dist, ierr)
INTEGER, target, intent(IN) :: arr (:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_3D_I), pointer :: arr_desc

call dp_find_arr_desc_3D_I (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_3D_I







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_3D_I
!!
subroutine dp_alloc_array_A_3D_I &
	(arr_ptr, global_size, local_size, ierr)

INTEGER, POINTER :: arr_ptr (:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (3)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (3)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 3)
integer :: local_bounds (2, 3)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_3D_I &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_3D_I 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_3D_I
!!
subroutine dp_alloc_array_B_3D_I &
	(arr_ptr, global_bounds, local_bounds, ierr)

INTEGER, POINTER :: arr_ptr (:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 3)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 3)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(3)
integer :: b (2, 3)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 3

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_3D_I (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_3D_I 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_3D_I 
!!
subroutine dp_print_array_desc_3D_I (arr_desc)
type (array_descriptor_3D_I) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 3D I"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_3D_I


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 4D INTEGER global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 4D INTEGER global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_4D_I
!!
SUBROUTINE dp_register_array_4D_I &
                                           (arr, global_size, ierr)
INTEGER, target, intent(IN) :: arr (:,:,:,:)
integer, intent(IN) :: global_size(4)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_4D_I), &
           pointer :: arr_desc 




print *, "dp_register_array_4D_I A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_4D_I B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_4D_I C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_4D_I D"

arr_desc%arr => arr


print *, "dp_register_array_4D_I E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 4 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_4D_I F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:4) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:4) = ubound(arr)
arr_desc%distribution%global_size(1:4) = global_size


print *, "dp_register_array_4D_I G"


do i = 1, 4


print *, "dp_register_array_4D_I H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_4D_I I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_4D_I J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_4D_I K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 4
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_4D_I L"


!! Add this node to the appropriate linked list
call add_arr_desc_4D_I (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_4D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_4D_I (arr_desc) 
type (array_descriptor_4D_I), &
                                    pointer :: arr_desc

if (associated ( Arr_List_4D_I )) then
	arr_desc%next => Arr_List_4D_I
else
	NULLIFY (arr_desc%next)
endif

Arr_List_4D_I => arr_desc

!! call dp_print_array_desc_4D_I (arr_desc)

end SUBROUTINE add_arr_desc_4D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_4D_I 
!!
SUBROUTINE dp_find_arr_desc_4D_I (arr, arr_desc, ierr)
INTEGER, target, intent(IN) :: arr (:,:,:,:)
type (array_descriptor_4D_I), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_4D_I

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_4D_I 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_4D_I
!!
SUBROUTINE dp_get_arr_dist_4D_I &
                                                (arr, dist, ierr)
INTEGER, target, intent(IN) :: arr (:,:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_4D_I), pointer :: arr_desc

call dp_find_arr_desc_4D_I (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_4D_I







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_4D_I
!!
subroutine dp_alloc_array_A_4D_I &
	(arr_ptr, global_size, local_size, ierr)

INTEGER, POINTER :: arr_ptr (:,:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (4)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (4)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 4)
integer :: local_bounds (2, 4)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_4D_I &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_4D_I 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_4D_I
!!
subroutine dp_alloc_array_B_4D_I &
	(arr_ptr, global_bounds, local_bounds, ierr)

INTEGER, POINTER :: arr_ptr (:,:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 4)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 4)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(4)
integer :: b (2, 4)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 4

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3), b(1,4):b(2,4)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_4D_I (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_4D_I 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_4D_I 
!!
subroutine dp_print_array_desc_4D_I (arr_desc)
type (array_descriptor_4D_I) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 4D I"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_4D_I


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 1D LOGICAL global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 1D LOGICAL global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_1D_L
!!
SUBROUTINE dp_register_array_1D_L &
                                           (arr, global_size, ierr)
LOGICAL, target, intent(IN) :: arr (:)
integer, intent(IN) :: global_size(1)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_1D_L), &
           pointer :: arr_desc 




print *, "dp_register_array_1D_L A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_1D_L B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_1D_L C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (1, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_1D_L D"

arr_desc%arr => arr


print *, "dp_register_array_1D_L E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 1 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_1D_L F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:1) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:1) = ubound(arr)
arr_desc%distribution%global_size(1:1) = global_size


print *, "dp_register_array_1D_L G"


do i = 1, 1


print *, "dp_register_array_1D_L H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_1D_L I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_1D_L J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_1D_L K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 1
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_1D_L L"


!! Add this node to the appropriate linked list
call add_arr_desc_1D_L (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_1D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_1D_L (arr_desc) 
type (array_descriptor_1D_L), &
                                    pointer :: arr_desc

if (associated ( Arr_List_1D_L )) then
	arr_desc%next => Arr_List_1D_L
else
	NULLIFY (arr_desc%next)
endif

Arr_List_1D_L => arr_desc

!! call dp_print_array_desc_1D_L (arr_desc)

end SUBROUTINE add_arr_desc_1D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_1D_L 
!!
SUBROUTINE dp_find_arr_desc_1D_L (arr, arr_desc, ierr)
LOGICAL, target, intent(IN) :: arr (:)
type (array_descriptor_1D_L), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_1D_L

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_1D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_1D_L
!!
SUBROUTINE dp_get_arr_dist_1D_L &
                                                (arr, dist, ierr)
LOGICAL, target, intent(IN) :: arr (:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_1D_L), pointer :: arr_desc

call dp_find_arr_desc_1D_L (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_1D_L







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_1D_L
!!
subroutine dp_alloc_array_A_1D_L &
	(arr_ptr, global_size, local_size, ierr)

LOGICAL, POINTER :: arr_ptr (:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (1)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (1)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 1)
integer :: local_bounds (2, 1)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_1D_L &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_1D_L 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_1D_L
!!
subroutine dp_alloc_array_B_1D_L &
	(arr_ptr, global_bounds, local_bounds, ierr)

LOGICAL, POINTER :: arr_ptr (:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 1)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 1)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(1)
integer :: b (2, 1)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 1

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_1D_L (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_1D_L 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_1D_L 
!!
subroutine dp_print_array_desc_1D_L (arr_desc)
type (array_descriptor_1D_L) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 1D L"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_1D_L


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 2D LOGICAL global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 2D LOGICAL global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_2D_L
!!
SUBROUTINE dp_register_array_2D_L &
                                           (arr, global_size, ierr)
LOGICAL, target, intent(IN) :: arr (:,:)
integer, intent(IN) :: global_size(2)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_2D_L), &
           pointer :: arr_desc 




print *, "dp_register_array_2D_L A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_2D_L B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_2D_L C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (2, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_2D_L D"

arr_desc%arr => arr


print *, "dp_register_array_2D_L E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 2 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_2D_L F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:2) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:2) = ubound(arr)
arr_desc%distribution%global_size(1:2) = global_size


print *, "dp_register_array_2D_L G"


do i = 1, 2


print *, "dp_register_array_2D_L H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_2D_L I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_2D_L J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_2D_L K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 2
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_2D_L L"


!! Add this node to the appropriate linked list
call add_arr_desc_2D_L (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_2D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_2D_L (arr_desc) 
type (array_descriptor_2D_L), &
                                    pointer :: arr_desc

if (associated ( Arr_List_2D_L )) then
	arr_desc%next => Arr_List_2D_L
else
	NULLIFY (arr_desc%next)
endif

Arr_List_2D_L => arr_desc

!! call dp_print_array_desc_2D_L (arr_desc)

end SUBROUTINE add_arr_desc_2D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_2D_L 
!!
SUBROUTINE dp_find_arr_desc_2D_L (arr, arr_desc, ierr)
LOGICAL, target, intent(IN) :: arr (:,:)
type (array_descriptor_2D_L), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_2D_L

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_2D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_2D_L
!!
SUBROUTINE dp_get_arr_dist_2D_L &
                                                (arr, dist, ierr)
LOGICAL, target, intent(IN) :: arr (:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_2D_L), pointer :: arr_desc

call dp_find_arr_desc_2D_L (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_2D_L







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_2D_L
!!
subroutine dp_alloc_array_A_2D_L &
	(arr_ptr, global_size, local_size, ierr)

LOGICAL, POINTER :: arr_ptr (:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (2)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (2)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 2)
integer :: local_bounds (2, 2)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_2D_L &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_2D_L 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_2D_L
!!
subroutine dp_alloc_array_B_2D_L &
	(arr_ptr, global_bounds, local_bounds, ierr)

LOGICAL, POINTER :: arr_ptr (:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 2)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 2)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(2)
integer :: b (2, 2)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 2

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_2D_L (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_2D_L 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_2D_L 
!!
subroutine dp_print_array_desc_2D_L (arr_desc)
type (array_descriptor_2D_L) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 2D L"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_2D_L


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 3D LOGICAL global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 3D LOGICAL global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_3D_L
!!
SUBROUTINE dp_register_array_3D_L &
                                           (arr, global_size, ierr)
LOGICAL, target, intent(IN) :: arr (:,:,:)
integer, intent(IN) :: global_size(3)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_3D_L), &
           pointer :: arr_desc 




print *, "dp_register_array_3D_L A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_3D_L B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_3D_L C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (3, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_3D_L D"

arr_desc%arr => arr


print *, "dp_register_array_3D_L E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 3 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_3D_L F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:3) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:3) = ubound(arr)
arr_desc%distribution%global_size(1:3) = global_size


print *, "dp_register_array_3D_L G"


do i = 1, 3


print *, "dp_register_array_3D_L H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_3D_L I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_3D_L J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_3D_L K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 3
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_3D_L L"


!! Add this node to the appropriate linked list
call add_arr_desc_3D_L (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_3D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_3D_L (arr_desc) 
type (array_descriptor_3D_L), &
                                    pointer :: arr_desc

if (associated ( Arr_List_3D_L )) then
	arr_desc%next => Arr_List_3D_L
else
	NULLIFY (arr_desc%next)
endif

Arr_List_3D_L => arr_desc

!! call dp_print_array_desc_3D_L (arr_desc)

end SUBROUTINE add_arr_desc_3D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_3D_L 
!!
SUBROUTINE dp_find_arr_desc_3D_L (arr, arr_desc, ierr)
LOGICAL, target, intent(IN) :: arr (:,:,:)
type (array_descriptor_3D_L), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_3D_L

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_3D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_3D_L
!!
SUBROUTINE dp_get_arr_dist_3D_L &
                                                (arr, dist, ierr)
LOGICAL, target, intent(IN) :: arr (:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_3D_L), pointer :: arr_desc

call dp_find_arr_desc_3D_L (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_3D_L







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_3D_L
!!
subroutine dp_alloc_array_A_3D_L &
	(arr_ptr, global_size, local_size, ierr)

LOGICAL, POINTER :: arr_ptr (:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (3)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (3)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 3)
integer :: local_bounds (2, 3)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_3D_L &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_3D_L 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_3D_L
!!
subroutine dp_alloc_array_B_3D_L &
	(arr_ptr, global_bounds, local_bounds, ierr)

LOGICAL, POINTER :: arr_ptr (:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 3)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 3)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(3)
integer :: b (2, 3)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 3

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_3D_L (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_3D_L 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_3D_L 
!!
subroutine dp_print_array_desc_3D_L (arr_desc)
type (array_descriptor_3D_L) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 3D L"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_3D_L


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! The following routines handle 4D LOGICAL global arrays.
!!
!!
!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! This routine "registers" a 4D LOGICAL global array.
!! It allocates the array_descriptor, fills in the appropriate fields
!! and adds it to the appropriate linked list.
!! 
!!  public dp_register_array_4D_L
!!
SUBROUTINE dp_register_array_4D_L &
                                           (arr, global_size, ierr)
LOGICAL, target, intent(IN) :: arr (:,:,:,:)
integer, intent(IN) :: global_size(4)
integer, intent(OUT) :: ierr

integer max_proc_arr_size
integer i, j


integer :: blk_end


type (array_descriptor_4D_L), &
           pointer :: arr_desc 




print *, "dp_register_array_4D_L A"



!! Allocate the descriptor
allocate (arr_desc, stat=ierr)
if (ierr .ne. 0) then
	return
endif

max_proc_arr_size = MAXVAL (dp_proc_arr_size)

print *, "dp_register_array_4D_L B"

!! Allocate the array distribution
allocate (arr_desc%distribution, stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc)
	return
endif

!! Allocate the array of block starting indices
allocate (arr_desc%distribution%blk_starts_0_based &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif


print *, "dp_register_array_4D_L C"

!! Allocate the array of block sizes
allocate (arr_desc%distribution%blk_sizes   &
                   (4, 0:max_proc_arr_size-1), stat=ierr)
if (ierr .ne. 0) then
	deallocate (arr_desc%distribution%blk_starts_0_based)
	deallocate (arr_desc%distribution)
	deallocate (arr_desc)
	return
endif



print *, "dp_register_array_4D_L D"

arr_desc%arr => arr


print *, "dp_register_array_4D_L E"

!!
!! Set members of distribution to something reasonable.  In particular
!! set the sizes and bounds for dimensions beyond 4 to 1.
!!
arr_desc%distribution%evenly_divided = .TRUE.
arr_desc%distribution%global_size = 1
arr_desc%distribution%local_bounds = 1
arr_desc%distribution%local_size = 1
arr_desc%distribution%local_fractional_size = 1
arr_desc%distribution%blk_starts_0_based = 0
arr_desc%distribution%blk_sizes = 1


print *, "dp_register_array_4D_L F"

!!
!! Set correct values for this array
!!
arr_desc%distribution%local_bounds(1,1:4) = lbound(arr)
arr_desc%distribution%local_bounds(2,1:4) = ubound(arr)
arr_desc%distribution%global_size(1:4) = global_size


print *, "dp_register_array_4D_L G"


do i = 1, 4


print *, "dp_register_array_4D_L H"

	arr_desc%distribution%local_size(i) = size(arr,i)

	!! Is it evenly divided along this dimension ? 
	if (MOD (arr_desc%distribution%global_size(i),  &
                               dp_proc_arr_size(i)) .eq. 0) then

		!! Even division is the easy case.
		
		arr_desc%distribution%evenly_divided(i) = .TRUE.

		!! Fractional size is always identical to local size when
		!! evenly divided.
		arr_desc%distribution%local_fractional_size (i) =   &
		                          arr_desc%distribution%local_size(i)

	else
		!! Uneven division

		arr_desc%distribution%evenly_divided(i) = .FALSE.

		arr_desc%distribution%local_fractional_size (i) =   &
		  arr_desc%distribution%global_size(i)/REAL(dp_proc_arr_size(i))

	endif




print *, "dp_register_array_4D_L I"

	!! Now calculate block starts and sizes along this dimension
	do j = 0, dp_proc_arr_size(i)-1


		call dp_get_local_array_block &
			(i, j, 0, global_size(i)-1, &
			 arr_desc%distribution%blk_starts_0_based(i, j), &
			 blk_end, ierr)
		if (ierr .ne. DP_SUCCESS) then
			deallocate (arr_desc%distribution%blk_sizes)
			deallocate (arr_desc%distribution%blk_starts_0_based)
			deallocate (arr_desc%distribution)
			deallocate (arr_desc)
			return
		endif

		arr_desc%distribution%blk_sizes (i, j) =  1 + blk_end -  &
		                 arr_desc%distribution%blk_starts_0_based(i, j)

	enddo


print *, "dp_register_array_4D_L J"


enddo !! end of loop over number of dimensions


!! Are all of the dimensions evenly divided?
arr_desc%distribution%all_even = all (arr_desc%distribution%evenly_divided)


print *, "dp_register_array_4D_L K"

!!
!! Check that the size of the array that has been passed in is correct
!! relative to the global_size specified in the call.
!!
do i = 1, 4
	if(arr_desc%distribution%local_size(i) .ne. &
	   arr_desc%distribution%blk_sizes(i, dp_coords(i))) then

		ierr = DP_INCONSISTENT_ARRAY_SIZE
		deallocate (arr_desc%distribution%blk_starts_0_based)
		deallocate (arr_desc%distribution%blk_sizes)
		deallocate (arr_desc%distribution)
		deallocate (arr_desc)
		return
	endif
enddo


print *, "dp_register_array_4D_L L"


!! Add this node to the appropriate linked list
call add_arr_desc_4D_L (arr_desc) 


ierr = 0
return

end SUBROUTINE dp_register_array_4D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Add an array descriptor to the linked list.
!!
!!
SUBROUTINE add_arr_desc_4D_L (arr_desc) 
type (array_descriptor_4D_L), &
                                    pointer :: arr_desc

if (associated ( Arr_List_4D_L )) then
	arr_desc%next => Arr_List_4D_L
else
	NULLIFY (arr_desc%next)
endif

Arr_List_4D_L => arr_desc

!! call dp_print_array_desc_4D_L (arr_desc)

end SUBROUTINE add_arr_desc_4D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_find_arr_desc_4D_L 
!!
SUBROUTINE dp_find_arr_desc_4D_L (arr, arr_desc, ierr)
LOGICAL, target, intent(IN) :: arr (:,:,:,:)
type (array_descriptor_4D_L), &
                                    pointer :: arr_desc
integer, intent (OUT) :: ierr


arr_desc => Arr_List_4D_L

do while (associated (arr_desc))

	if (associated (arr_desc%arr, arr)) then
		ierr = 0
		return
	endif

	arr_desc => arr_desc%next
enddo

nullify (arr_desc)
ierr = -1
return

end SUBROUTINE dp_find_arr_desc_4D_L 




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Search the linked list for the entry corresponding to the 
!! indicated array.
!!
!!  public dp_get_arr_dist_4D_L
!!
SUBROUTINE dp_get_arr_dist_4D_L &
                                                (arr, dist, ierr)
LOGICAL, target, intent(IN) :: arr (:,:,:,:)
type (array_distribution), pointer :: dist
integer, intent (OUT) :: ierr


type (array_descriptor_4D_L), pointer :: arr_desc

call dp_find_arr_desc_4D_L (arr, arr_desc, ierr)

if (associated (arr_desc)) then
	dist => arr_desc%distribution
endif

nullify (dist)

ierr = -1
return

end SUBROUTINE dp_get_arr_dist_4D_L







!!
!! Allocate and register a local array based on global dimensions.
!!
!!  public dp_alloc_array_A_4D_L
!!
subroutine dp_alloc_array_A_4D_L &
	(arr_ptr, global_size, local_size, ierr)

LOGICAL, POINTER :: arr_ptr (:,:,:,:)		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_size (4)	!! The size of the 
							!! global array to 
							!! be allocated

integer :: local_size (4)			!! Size of the local
							!! array that was
							!! allocated.

integer, INTENT(OUT) :: ierr				!! Error status.  

integer :: global_bounds (2, 4)
integer :: local_bounds (2, 4)

global_bounds(1,:) = 1
global_bounds(2,:) = global_size

call dp_alloc_array_B_4D_L &
		(arr_ptr, global_bounds, local_bounds, ierr)

if (ierr .eq. 0) then
	local_size = 1 + local_bounds(2,:) - local_bounds(1,:)
endif


return

end subroutine dp_alloc_array_A_4D_L 







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Allocate and register a local array based on global bounds.
!!
!!  public dp_alloc_array_B_4D_L
!!
subroutine dp_alloc_array_B_4D_L &
	(arr_ptr, global_bounds, local_bounds, ierr)

LOGICAL, POINTER :: arr_ptr (:,:,:,:) 		!! pointer to allocated
							!! array

integer, INTENT(IN) :: global_bounds (2, 4)    !! The bounds of the
							!! global array to be
							!! allocated

integer, INTENT(OUT) :: local_bounds (2, 4)    !! Bounds of the local
							!! array that was
							!! allocated

integer, INTENT(OUT) :: ierr				!! Error status.  



integer global_size(4)
integer :: b (2, 4)
integer i


global_size = 1 + global_bounds(2, :) - global_bounds(1, :)


!! Figure out the size of the local array
do i = 1, 4

	call dp_get_local_array_block (i, dp_coords(i), &
			 global_bounds(1,i), global_bounds(2,i), &
			 local_bounds(1,i), local_bounds(2,i), ierr)

enddo !! end of loop over number of dimensions

b = local_bounds

allocate (arr_ptr (b(1,1):b(2,1), b(1,2):b(2,2), b(1,3):b(2,3), b(1,4):b(2,4)), stat = ierr)

if (ierr .ne. 0) then
	return
endif


call dp_register_array_4D_L (arr_ptr, global_size, ierr)

return

end subroutine dp_alloc_array_B_4D_L 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! 
!! 
!! 
!! 
!!  public dp_print_array_desc_4D_L 
!!
subroutine dp_print_array_desc_4D_L (arr_desc)
type (array_descriptor_4D_L) :: arr_desc 

print *, "*************************************************************"
print *, "Array descriptor : 4D L"
print *, "   array ubounds = ", ubound (arr_desc%arr)
print *, "   array lbounds = ", lbound (arr_desc%arr)
print *, "   all_even = ", arr_desc%distribution%all_even
print *, "   evenly_divided = ", arr_desc%distribution%evenly_divided
print *, "   global_size = ", arr_desc%distribution%global_size 
print *, "   local_bounds = ", arr_desc%distribution%local_bounds
print *, "   local_size = ", arr_desc%distribution%local_size
print *, "   local_fractional_size = ", &
		arr_desc%distribution%local_fractional_size
print *, "   blk_starts_0_based = ", arr_desc%distribution%blk_starts_0_based 
print *, "   blk_sizes = ", arr_desc%distribution%blk_sizes
print *, "*************************************************************"
return

end subroutine dp_print_array_desc_4D_L


!!
!! 
!!
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!




